#include <string>
#include <iostream>
#include <sstream>

using namespace std;

std::string itoa(const int& x){
  std::ostringstream o;
  if (!(o << x)) return "ERROR";
  return o.str();
}

int main(int argc, char** argv){
  string netlistName;
  int numRows;
  int numCols;
  int numTracks;
  string command;
  int seed;

  if(argc != 6){
    cout << "ERROR: Input error!" << endl;
    cout << "Usage: netlist #rows #columns #tracks seed" << endl;
    return (-1);
  }

  //Argument $1 is the netlist, $2 is the number of rows, $3 is the number of columns, $4 is the number of tracks
  //Generate the rr_graph.echo file
  netlistName = argv[1];
  numRows = atoi(argv[2]);
  numCols = atoi(argv[3]);
  numTracks = atoi(argv[4]);
  seed = atoi(argv[5]);

  //Run the tool on the netlist
  // GENERAL OPTIONS
  // N -timing_driven: both placement and routing will be timing-driven
  //
  // PLACEMENT OPTIONS
  // Y -pipeline_aware_placement: all placements done with this right now must be pipeline aware or nothing is actually done
  // X -splitDelayBlocks makes no sense since all register banks are single FF
  // ? -retime: depends on whether or not I want to do retiming
  // ? -unzipNets depends on when I want to perform register duplication
  // ? -clb_level_placement depends on the input netlist
  // ? all tuning options (tdpLambda, tdpUpdateInterval, tdpCritExponent)
  //
  // ROUTING OPTIONS
  // Y -pipeline_aware_routing
  // Y -lockRegPlacement
  // Notice that we have to do pipeline aware routing or the tool will ignore latency
  //  however, the -lockRegPlacement option means we use regular PathFinder, not Armada
  command = "../../armada " + netlistName + " " + itoa(numRows) + " " + itoa(numCols) + 
    " 2 ../exampleArchitectures/4lut_" + itoa(numRows) + "x" + itoa(numCols) + "x" + itoa(numTracks) + 
    ".rrDelay " + itoa(seed) + " -num_switches 13 456E-12 1.5E-9 0.0 0.546E-9 0.299E-9 0.299E-9  0.0 0.478E-9 1.0E9 1.0E9 1.0E9 0.478E-9 0.295E-9 -timing_driven -pipeline_aware_placement -clb_level_placement -tdpCritExponentFinal 12.0 -updateSlack -tdpLambda 0.025";
  system(command.c_str());

  return 0;
}
