// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern long int sharingCost;
//extern long int wireCost;
extern vprFPGA *FPGA;
extern double tdPlacementCritExponent;
extern bool maxCutCost;
extern bool updateSlack;
//extern bool pipeline_aware_placement;
//extern bool pipeline_aware_routing;
extern bool deltaDelayCost;

#ifdef PLACER_DEBUG
extern int numRegistersEstimatedNeeded;
#endif

static const double cross_count[50] = {   /* [0..49] */
1.0,    1.0,    1.0,    1.0828, 1.1536, 1.2206, 1.2823, 1.3385, 1.3991, 1.4493,
1.4974, 1.5455, 1.5937, 1.6418, 1.6899, 1.7304, 1.7709, 1.8114, 1.8519, 1.8924,
1.9288, 1.9652, 2.0015, 2.0379, 2.0743, 2.1061, 2.1379, 2.1698, 2.2016, 2.2334,
2.2646, 2.2958, 2.3271, 2.3583, 2.3895, 2.4187, 2.4479, 2.4772, 2.5064, 2.5356,
2.5610, 2.5864, 2.6117, 2.6371, 2.6625, 2.6887, 2.7148, 2.7410, 2.7671, 2.7933};

// Give this net a name
void Net::set_name(string n){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  name = n;
}

// Set this net's source terminal
void Net::set_source(Terminal* s){  
  if(!s)
    ERROR("net source cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  source=s;
}

// // Initial creation of vector of sinks
// void Net::create_sinkVector(){
//   if(magicNum!=17111)
//     ERROR("trying to access a deleted node...");
//   if(numSinks==0){
// //      ERROR("num sinks is 0 here...");
//     cout<<endl<<endl<<name<<" HAS NO SINKS!!";
//     return;
//   }
//   sinkVector=new vector<Terminal*>(numSinks);

//   list<Terminal*>::iterator sink_itr=sinkList.begin();
//   for(int i=0;sink_itr!=sinkList.end();sink_itr++,i++){
//     sinkVector->at(i)=(*sink_itr);
//   }
// }

// Sort vector of sinks in increasing order of separation from source...A*
void Net::sort_sinkVectorDistance(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  int i=0;int j=0;
  int irow,icol,jrow,jcol,source_row,source_col;
  source_row=source->rNode->row_low;source_col=source->rNode->col_low;
  Terminal* temp_term;
  double tempDouble;
  int tempInt;

  // some signals may not have sinks
  //if(sinkVector==0)
  // return;

  for(i=0;i<sinkVector.size();i++){
    irow=sinkVector.at(i)->rNode->row_low;
    icol=sinkVector.at(i)->rNode->col_low;

    for(j=i+1;j<sinkVector.size();j++){
      jrow=sinkVector.at(j)->rNode->row_low;
      jcol=sinkVector.at(j)->rNode->col_low;
      if((abs(source_row-jrow) + abs(source_col-jcol)) < (abs(source_row-irow) + abs(source_col-icol))){
	//swap positions in the sink vector
	temp_term=sinkVector.at(i);
	sinkVector.at(i)=sinkVector.at(j);
	sinkVector.at(j)=temp_term;

	//swap values in linkDelayVector
	tempDouble = linkDelayVector.at(i);
	linkDelayVector.at(i) = linkDelayVector.at(j);
	linkDelayVector.at(j) = tempDouble;

	//swap values in slackVector
	tempDouble = slackVector.at(i);
	slackVector.at(i) = slackVector.at(j);
	slackVector.at(j) = tempDouble;

	//swap values in latencyVector
	tempInt = latencyVector.at(i);
	latencyVector.at(i) = latencyVector.at(j);
	latencyVector.at(j) = tempInt;

	//Update i row/col info
	irow = jrow;
	icol = jcol;
      }
    }

  }
}

//Sort sinks of the net in order of increasing slack.
// Commented-out code is to implement tiebreaker - if they are the same, sort by increasing distance.
//  Should this be decreasing distance?  The furthest away sink might have the hardest to meet timing.
//  Maybe we don't really care anyways about distance.
void Net::sort_sinkVectorTiming(){
  Terminal *tempSink;
  int low, high;
  double tempDouble;
  int tempInt;
  //int sourceRow, sourceCol;
  //int lowSinkRow, lowSinkCol, highSinkRow, highSinkCol;
  
  //cout << "Net " << name << endl;
  //sourceRow = source->rNode->row_low;
  //sourceCol=source->rNode->col_low;

  for(low = 0; low < sinkVector.size(); low++){
    //lowSinkRow = sinkVector.at(low)->rNode->row_low;
    //lowSinkCol = sinkVector.at(low)->rNode->col_low;

    for(high = low + 1; high < sinkVector.size(); high++){
      //cout << "   Considering net " << name << " sink " << low << " slack = " << slackVector.at(low);
      //cout << " and sink " << high << " slack = " << slackVector.at(high)<< endl;
      //highSinkRow = sinkVector.at(low)->rNode->row_low;
      //highSinkCol = sinkVector.at(low)->rNode->col_low;

      if(slackVector.at(low) > slackVector.at(high)){
	//|| (((slackVector.at(low) == slackVector.at(high)) && 
	//  ((abs(sourceRow - lowSinkRow) + abs(sourceCol - lowSinkCol)) > (abs(sourceRow - highSinkRow) + abs(sourceCol - highSinkCol)))){
	//We have decided to swap these sinks so we have to swap all of their information at the same time
	//1) Change positions in sinkVector
	//cout << "     Swapping net " << name << " sink " << low << " slack = " << slackVector.at(low);
	//cout << " and sink " << high << " slack = " << slackVector.at(high)<< endl;
	tempSink = sinkVector.at(high);
	sinkVector.at(high) = sinkVector.at(low);
	sinkVector.at(low) = tempSink;
	//2) Swap values in linkDelayVector
	tempDouble = linkDelayVector.at(high);
	linkDelayVector.at(high) = linkDelayVector.at(low);
	linkDelayVector.at(low) = tempDouble;
	//3) Swap values slackVector
	tempDouble = slackVector.at(high);
	slackVector.at(high) = slackVector.at(low);
	slackVector.at(low) = tempDouble;
	//4) Swap values in latencyVector
	tempInt = latencyVector.at(high);
	latencyVector.at(high) = latencyVector.at(low);
	latencyVector.at(low) = tempInt;
	//Update low sink row and col
	//lowSinkRow = highSinkRow;
	//lowSinkCol = highSinkCol;
      }
    }
  }
  //for(low = 0; low < sinkVector.size(); low++){
  //  cout << "slack = " << slackVector.at(low) << endl;
  //}
}

//We have placement-locked registers, so sort all of the pipelined nets
void Net::sortLockedRegs(bool sortByTiming){

  for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
    (*iter)->sortLockedRegs(sortByTiming);
  }
}

void Net::create_slackVector(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(numSinks==0){
    //      ERROR("num sinks is 0 here...");
    cout<<endl<<endl<<name<<" HAS NO SINKS!!";
    return;
  }  
  
  //Initialize slack vector to all 0.0's
  slackVector.resize(numSinks, 0.0);
}

void Net::create_linkDelayVector(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(numSinks==0){
    //      ERROR("num sinks is 0 here...");
    cout<<endl<<endl<<name<<" HAS NO SINKS!!";
    return;
  }  
  
  //Initialize pathDelay vector to all 0.0
  linkDelayVector.resize(numSinks, 0.0);
  if(deltaDelayCost){
    oldLinkDelayVector.resize(numSinks, -100.0);
  }
}


// // Add a terminal to this net's sink list
// void Net::addto_sinkList(Terminal* s){
//   if(!s)
//     ERROR("sink cannot be NULL...");
//   if(magicNum!=17111)
//     ERROR("trying to access a deleted node...");
//   sinkList.push_back(s);
//   numSinks++;
// }

// Add a terminal to this net's sink list
void Net::addto_sinkVector(Terminal* s){
  if(!s)
    ERROR("sink cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  sinkVector.push_back(s);
  numSinks++;
}

// void Net::addto_delayList(int delay){
//   if(delay < 0 || delay > 100)
//     ERROR("trying to push a suspicious delay");

//   if(magicNum!=17111)
//     ERROR("trying to access a deleted node...");
//   delayList.push_back(delay);
// }

// Get name of this net
string Net::get_name(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Rip up the existing routing tree, and decrement sharing
void Net::rip_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"RIPPING ROUTE TREE OF "<<name;
#endif

  //rTree.postorderTraverse(rTree.get_root(), 2);
  rTree.postorderTraverse(rTree.get_root(), 93);
  rTree.set_root(0);
}

// Reset the routing nodes in the tree to not counted in sharing calculations
void Net::reset_counted(){
  rTree.postorderTraverse(rTree.get_root(),5);
}

// Rip up the branch of the current route tree that contains sink_rnode
void Net::rip_rTreeBranch(routingNode* sink_rnode){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(sink_rnode==0)
    ERROR("sink routing node cannot be NULL...");

  treeNode* sink_tree_node=rTree.findNode(rTree.get_root(),sink_rnode);
  if(sink_tree_node==0)
    ERROR("could not find sink in route tree...");

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"LOOKING AT BRANCH OF NET "<<name<<" THAT SINKS AT "<<sink_rnode->name;
#endif

  list<treeNode*>::iterator ch_itr;
  treeNode* tree_node=sink_tree_node; treeNode* temp_node;
  while(1){
    if(tree_node->get_numChildren()>1 || tree_node->get_rNode()->get_type()==SOURCE){
      tree_node->eraseChild(&ch_itr);
      /* this is the root of the subtree being deleted...break out */
      break;
    }

    temp_node=tree_node->get_parent();
    if(temp_node->get_numChildren()>1 || temp_node->get_rNode()->get_type()==SOURCE)
      ch_itr=temp_node->get_childItr(tree_node);

    /* change sharing and wire cost to refelct the deletion of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost-=1;
    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE){
      //wireCost-=1;
    }
//     wireCost-=1;

    /* decrement occupancy and delete tree node */
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()-1);

#if defined(PLACER_DEBUG)
    cout<<endl<<"\tdecrement and delete "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy();
#endif

    delete tree_node;

    tree_node=temp_node;
  }

}

// Print the current route tree of this net
void Net::print_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  rTree.print_info();
}

// Whether this net's cost contributions have already been considered during
// an annealing move
void Net::set_isCounted(bool c){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  isCounted=c;
}

// Get whether this net's cost contributions have already been considered during
// an annealing move
bool Net::get_isCounted(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return isCounted;
}

// Reconnect the sub-tree rooted at join_node to this net's current route tree.
// Increment occupancies of the routing nodes that constitute the subtree.
int Net::reconnect(treeNode* join_node){
  treeNode* child;

  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(join_node==0)
    ERROR("NULL join node here...");

  /* locate the tree node in rTree that corresponds to join_node */
  treeNode* tree_node=rTree.getNode(join_node->get_rNode());
  if(tree_node==0){
//     ERROR("could not find tree node in route tree...");
    return -1;
  }

  /* change sharing and wire cost to avoid double counting*/
  if(join_node->get_rNode()->get_occupancy()>join_node->get_rNode()->get_capacity())
    sharingCost-=1;

//   wireCost-=1;

  if(join_node->get_rNode()->get_type()==BUS_OR_WIRE){
    //wireCost-=1;
  }
  /* decrement the occupancy of join_node...since we have double counted */
  join_node->get_rNode()->set_occupancy(join_node->get_rNode()->get_occupancy() - 1);
  /* now increment the occupancies of routing nodes in the sub-tree */
  incrementOccupancies(join_node,0);

  /* reconnect the sub-tree rooted at join_node */
  join_node->create_childListItr();
  for(child=join_node->getnext_childList();child!=0;child=join_node->getnext_childList()){
    tree_node->addto_childList(child);
    child->set_parent(tree_node);
  }
  join_node->remove_childListItr();

  /* finally delete join_node */
  delete join_node;
  return 1;

}

// Increment occupancies of the routing nodes that constitute sub-tree rooted at r
void Net::incrementOccupancies(treeNode* r,int how){
  treeNode* tree_node=r;
  treeNode* child;

  if(how==0){
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy() + 1);
    /* change sharing and wire cost to reflect the addition of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost+=1;

//     wireCost+=1;

    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE){
      //wireCost+=1;
    }
  }
  else{
    if(tree_node->get_rNode()->get_type()!=SOURCE && tree_node->get_rNode()->get_type()!=SINK)
      tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy() + 1);
  }

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    incrementOccupancies(child,how);
  tree_node->remove_childListItr();

}

// Calculate the sharing and wire cost of this net's current route
void Net::calcCosts(int* sh_cost,int* wr_cost){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(*sh_cost!=0 || *wr_cost!=0)
    ERROR("uninitialized cost parameters here...");

  cout<<endl<<endl<<"------- Calculating costs for "<<name<<" -------";

  wsCostCalc(rTree.get_root(),sh_cost,wr_cost);
}

// Recursively calculate wire and sharing cost
void Net::wsCostCalc(treeNode* r,int* sh_cost,int* wr_cost){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    wsCostCalc(child,sh_cost,wr_cost);
  tree_node->remove_childListItr();

  *wr_cost+=1;
  cout<<endl<<"\tincremented wire cost due to "<<tree_node->get_rNode()->name<<
    " to "<<*wr_cost;

  routingNode* rnode=tree_node->get_rNode();
//   if(rnode->get_occupancy()>rnode->get_capacity()){
//     *sh_cost+=1;
//     cout<<" and sharing cost to "<<*sh_cost;
//   }

  if(rnode->get_occupancy()>rnode->get_capacity() && rnode->get_counted()==false){
//     *sh_cost+=(rnode->get_occupancy() - rnode->get_capacity());
    *sh_cost+=1;
    cout<<" and sharing cost to "<<*sh_cost;
    rnode->set_counted(true);
  }

}

// Get the total number of nodes in the current route tree
int Net::get_rNodeCount(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  int node_count=0;
  countNodes(rTree.get_root(),&node_count);
  return node_count;
}

// Count number of nodes in the current route tree
void Net::countNodes(treeNode* r,int* count_ptr){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    countNodes(child,count_ptr);
  tree_node->remove_childListItr();

//   *count_ptr+=1;
  if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
    *count_ptr+=1;

}

// Make the source routing node the root of  rTree
void Net::init_rTreeRoot(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  treeNode* root_node=new treeNode;
  root_node->set_rNode(source->get_rNode());
  rTree.set_root(root_node);
}

// Get this net's source terminal
Terminal* Net::get_source(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return source;
}

// Get num sinks of this net
int Net::get_numSinks(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return numSinks;
}

// Print details about this net
void Net::print_Info(bool pipeline_aware){
  int i;
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"===================================";
  cout<<endl<<"              PRINT NET              ";
  cout<<endl<<"===================================";

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR NET "<<name;
  cout<<endl<<"SOURCE LOGIC BLOCK IS "<<source->get_lBlock()->get_name();
  cout << "Required Time = " << source->requiredTime;
  if(!(pipelined && pipeline_aware)){
    cout<<endl<<"SINK LOGIC BLOCKS ARE (latency/delay/slack) [node number]-->";
    vector<Terminal*>::iterator term_itr = sinkVector.begin();
    for(i = 0;term_itr!=sinkVector.end();term_itr++, i++){
      cout << endl << "  "<<(*term_itr)->get_lBlock()->get_name();
      cout << " (" << latencyVector[i] << "/" << linkDelayVector[i] << "/" << slackVector[i] << ")";
      //cout << " [" << (*term_itr)->get_rNode()->get_name() << "]";
    }
    cout << endl << "MinRow = " << minRow;
    cout << endl << "MaxRow = " << maxRow;
    cout << endl << "MinCol = " << minCol;
    cout << endl << "MaxCol = " << maxCol;
    cout << endl << "Wire cost = " << wireCost;
    cout << endl << "Timing cost = " << timingCost;
  }
  else{
    cout<<endl<<"SINK LOGIC BLOCKS ARE (latency/minimum slack) [node number]-->";
    vector<Terminal*>::iterator term_itr = sinkVector.begin();
    for(i = 0;term_itr!=sinkVector.end();term_itr++, i++){
      cout << endl << "  "<<(*term_itr)->get_lBlock()->get_name();
      cout << " (" << latencyVector[i] << "/" << slackVector[i] << ")";
      cout << " [" << (*term_itr)->get_rNode()->get_name() << "]";

      cout << "Required Time = " << (*term_itr)->requiredTime;
      cout << " Departure time = " << (*term_itr)->departureTime;
    }
  }

  print_rTree();

  if(pipelined && pipeline_aware){
    for(list<Terminal *>::iterator iter = pipelineRegisterList.begin(); iter != pipelineRegisterList.end(); iter++){
      cout << "Pipeline Register " << (*iter)->name << " is a sink of net " << (*iter)->sourcePipeliningNet->name <<
	" and is the source of net " << (*iter)->destPipeliningNet->name << endl;
    }

    for(list<PipeliningNet *>::iterator pNetIter = pipelineNetList.begin(); pNetIter != pipelineNetList.end(); pNetIter++){
      (*pNetIter)->print_Info();
    }
  }
  cout<<endl; 
}

// Get a pointer to this net's route tree
routeTree* Net::get_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  return &rTree;
}

// Return tree node that holds the routing node pointed to by rnode
treeNode* Net::find_treeNode(routingNode* rnode){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(rnode==0)
    ERROR("cant find NULL routing node in tree...");
  treeNode* tree_node=rTree.findNode(rTree.get_root(),rnode);
  if(tree_node==0)
    ERROR("uh,oh! could not find routing node in tree...");

//   cout<<endl<<"FOUND tree node corresponding to "<<tree_node->get_rNode()->get_name();

  return tree_node;
}

//Calculate the new path delay for this net
//void Net::updatePathDelay(){
//}

//Zero out the link delay and slack vector values
void Net::clearLinkDelayAndSlackVectors(){
  for(vector <double>::iterator iter = linkDelayVector.begin(); iter != linkDelayVector.end(); iter++){
    *iter = 0.0;
  }
  for(vector <double>::iterator iter = slackVector.begin(); iter != slackVector.end(); iter++){
    *iter = 0.0;
  }
}

void Net::addto_latencyVector(int latency){
  if(latency < 0 || latency > 100)
    ERROR("trying to push a suspicious latency");

  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  latencyVector.push_back(latency);
}

//Removes all traces of the existing net and pipelined nets
void Net::ripUpPipelineInformation(){
  list<PipeliningNet *>::iterator pNetIter;
  list<Terminal *>::iterator dNodeIter;
  vector<Terminal*>::iterator sinkIter;

  //First, go through all of the pipelining nets and decrement occupany and sharing
  //Then remove the net's themselves.
  //After that, remove the pipelining registers that we used.
#if defined(ROUTER_DEBUG)
  cout << "RIPPING UP PIPELINED NET" << endl;
  cout << "\tBefore ripup" << endl;
  print_Info();
#endif

  //Make sure that the routing tree is empty
  if(!rTree.root == 0){
    ERROR("Pipelined net has non-empty unpipelined routing tree!");
  }

  for(pNetIter = pipelineNetList.begin(); pNetIter != pipelineNetList.end(); pNetIter++){
    //We have to decrement the occupancy/sharing on the source of the entire net, but for all
    //subsequent nets we only decrement the occupany/sharing of main body and sinks.
    (*pNetIter)->ripUpRTree((*pNetIter)->latency == 0);
    delete (*pNetIter);
  }
  pipelineNetList.clear();

  for(dNodeIter = pipelineRegisterList.begin(); dNodeIter != pipelineRegisterList.end(); dNodeIter++){
    delete (*dNodeIter);
  }
  pipelineRegisterList.clear();
  
#if defined(ROUTER_DEBUG)
  cout << "\tAfter ripup" << endl;
  print_Info();
#endif

  //Go and erase and reference from sources and sinks
  if(source->sourcePipeliningNet){
    ERROR("Somehow source of net has a source pipelining net!");
  }
  source->destPipeliningNet = 0;
  for(sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
    (*sinkIter)->sourcePipeliningNet = 0;
    if((*sinkIter)->destPipeliningNet){
      ERROR("Somehow sink of net has a destination pipelining net!");
    }
  }
}

//Recalculate the occupancy of this net using the "newOccupancy" variable
void Net::reCalculateOccupancy(bool pipeline_aware){
  if(!pipelined || !pipeline_aware){
    if(rTree.root)
      rTree.reCalculateOccupancy(rTree.root);
  }
  else{
    for(list<PipeliningNet *>::iterator pipeIter = pipelineNetList.begin(); pipeIter != pipelineNetList.end(); pipeIter++){
      (*pipeIter)->reCalculateOccupancy();
    }
  }
}

//Make sure that we have a real route to all sinks
bool Net::followSinks(bool pipeline_aware){
  queue<treeNode*> nodeQ;
  queue<PipeliningNet*> netQ;
  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator latencyIter;
  list<treeNode*>::iterator childIter;
  vector<Terminal*>::iterator pipeSinkIter;
  bool found;

  if(!pipelined || !pipeline_aware){
    for(sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
      //Start at the source of the routing tree
      found = false;
      nodeQ.push(rTree.root);
      while(!nodeQ.empty()){
	//If the top item in the queue is the sink, break out and go to the next sink
	if(nodeQ.front()->rNode == (*sinkIter)->rNode){
	  found = true;
	  while(!nodeQ.empty()){
	    nodeQ.pop();
	  }
	}
	else{
	  //If not, add it's children, then pop it off
	  for(childIter = nodeQ.front()->childList.begin(); childIter != nodeQ.front()->childList.end(); childIter++){
	    nodeQ.push(*childIter);
	  }
	  nodeQ.pop();
	}
      }
      if(!found){
	cout << "Could not find sink " << (*sinkIter)->rNode->name << endl;
	return false;
      }
    }
    return true;
  }
  else{
    //Make sure that all of the main net's sinks are accounted for by one of the pipelined net's sinks
    for(sinkIter = sinkVector.begin(), latencyIter = latencyVector.begin(); sinkIter != sinkVector.end(), latencyIter != latencyVector.end(); sinkIter++, latencyIter++){
      found = false;
      
      //Start at the source of the net
      netQ.push(source->destPipeliningNet);
      nodeQ.push(netQ.front()->rTree.root);
      
      while(!netQ.empty() || !nodeQ.empty()){
	//If the top of the queue is the sink, at the correct latency, break out and go to the next sink
	if(nodeQ.front()->rNode == (*sinkIter)->rNode && netQ.front()->latency == (*latencyIter)){
	  found = true;
	  while(!nodeQ.empty()){
	    nodeQ.pop();
	    netQ.pop();
	  }
	}
	else{
	  //Otherwise, add it's children, then pop it off
	  for(childIter = nodeQ.front()->childList.begin(); childIter != nodeQ.front()->childList.end(); childIter++){
	    nodeQ.push(*childIter);
	    netQ.push(netQ.front());
	  }

	  //If this is a sink terminal for this net, it might also have a follow on pipelining net
	  for(pipeSinkIter = netQ.front()->sinkVector.begin(); pipeSinkIter != netQ.front()->sinkVector.end(); pipeSinkIter++){
	    if(nodeQ.front()->rNode == (*pipeSinkIter)->rNode){
	      if((*pipeSinkIter)->destPipeliningNet){
		if((*pipeSinkIter)->destPipeliningNet->source != *pipeSinkIter){
		  cout << "Problem! Destination pipelining net does not have same source as previous net's sink!" << endl;
		  return false;
		}
		nodeQ.push((*pipeSinkIter)->destPipeliningNet->rTree.root);
		netQ.push((*pipeSinkIter)->destPipeliningNet);
	      }
	    }
	  }
	  nodeQ.pop();
	  netQ.pop();
	}
      }
      if(!found){
	cout << "Could not find sink " << (*sinkIter)->rNode->name << endl;
	return false;
      }
    }
    return true;
  }
}

//After calculating the slack of each of the individual pipelining nets within this net we can calculate the 
//overall slack for each of the main net's sinks.  We define that to be the minimum slack of any link
//between the sink and the source.
void Net::calculatePipelinedSlacks(){
  PipeliningNet *currPNet;
  Terminal* currSink;
  double minSlack;
  vector<Terminal*>::iterator sinkIter;
  vector<Terminal*>::iterator pNetSinkIter;
  int i, j;

  for(sinkIter = sinkVector.begin(), i = 0; sinkIter != sinkVector.end(), i < slackVector.size(); sinkIter++, i++){
    currPNet = (*sinkIter)->sourcePipeliningNet;
    currSink = (*sinkIter);
    minSlack = 1.0E9;
    
    while(currPNet){
      //Find the current sink in the current pipelining net
      for(pNetSinkIter = currPNet->sinkVector.begin(), j = 0; pNetSinkIter != currPNet->sinkVector.end(), j < currPNet->slackVector.size(); pNetSinkIter++, j++){
	if(*pNetSinkIter == currSink){
	  //cout << "Sink " << (*pNetSinkIter)->name << " has slack " << currPNet->slackVector.at(j) << endl;
	  if(minSlack > currPNet->slackVector.at(j)){
	    //cout << "Setting minSlack to " << currPNet->slackVector.at(j) << endl;
	    minSlack = currPNet->slackVector.at(j);
	  }
	  break;
	}
      }
      //Check to see if we found this sink
      if(pNetSinkIter == currPNet->sinkVector.end()){
	ERROR("Couldn't find sink in pipelining net!");
      }
      
      //Set the current sink to be the source of the current net and the current net to be the source net of the current sink
      currSink = currPNet->source;
      currPNet = currSink->sourcePipeliningNet;
    }
    
    //Set the minimum slack to be the slack of the entire net's sink
    slackVector.at(i) = minSlack;
    //cout << "Set entire net's sink " << (*sinkIter)->name << " slack to " << slackVector.at(i) << endl;
  }
}

void Net::printCriticalPath(bool pipeline_aware){
  list<PipeliningNet *>::iterator pNetIter;
  vector <double>::iterator dIter;
  int i;
  if(pipelined && pipeline_aware){
    for(pNetIter = pipelineNetList.begin(); pNetIter != pipelineNetList.end(); pNetIter++){
      (*pNetIter)->printCriticalPath();
    }
  }
  else{
    for(dIter = slackVector.begin(), i = 0; dIter != slackVector.end(); dIter++, i++){
      if(*dIter < 1E-20){
	cout << "Net " << name << " link between source " << source->name << " (node " << source->rNode->name << ") and sink " << (sinkVector.at(i))->name;
	cout << " (node " << (sinkVector.at(i))->rNode->name << ") has slack " << *dIter << endl;
	rTree.printPath(sinkVector.at(i));
      }
    }
  }
}

// Go through all sinks for this net and set up all of the pipelined nets assuming we are
// fully zipped (maximum register sharing) and that we can accumulate up to 3 delays per delay block.
// This will only get called on pipelined nets (nets with at least one sink with a non-zero delay).
// Notice that the meaning of pipelined nets differs from that used during the routing phase.
// Here, we only create a pipelined net between delay blocks when we have to - instead of each net representing
// a single delay opportunity, this might be multiple delays.
// We assume that sinks are already sorted in non-decreasing order of latency.
// Notice, this function is a bit RaPiD specific.  If we are doing this on an architecture
// that has mixed register resources, we might have to decide on a case-by-case basis as to how many
// delays we want to attempt to pick up.

// void Net::createZipped3PipelinedNets(){
//   int currMaxLatency = 0;
//   PipeliningNet* currPNet;
//   PipeliningNet* nextPNet;
//   Terminal* delayRegTerm;
//   char tempStr [MAX_STR_LEN];
//   int targetLatency;
//   int acquiredDelay;

//   //Make a pipelined net for the source
//   strcpy(tempStr, name);
//   sprintf(tempStr, "%sPipelineNet%d", tempStr, pipelineNetList.size());
//   currPNet = new PipeliningNet(tempStr, source, 0);
//   pipelineNetList.push_back(currPNet);

//   // Go through each sink in the netlist and figure out the least number of delay blocks we'll need
//   // to meet the latency requirement for this sink
//   for(int i = 0; i < sinkVector.size(); i++){
//     targetLatency = latencyVector.at(i);

//     //Create pipelining netlists until we get enough latency
//     while(targetLatency != currMaxLatency){
//       // Create a new pipelining netlist
//       strcpy(tempStr, name);
//       sprintf(tempStr, "%sPipelineNet%d", tempStr, pipelineNetList.size());

//       // The latency will be either the current latency + 3 or the current latency + difference to target
//       // whichever is smaller
//       acquiredDelay = min(targetLatency - currMaxLatency, 3);
//       nextPNet = new PipeliningNet(tempStr, currMaxLatency + acquiredDelay);
//       pipelineNetList.push_back(nextPNet);

//       // Create the terminal that will represent the delay block
//       strcpy(tempStr, name);
//       sprintf(tempStr, "%sReg%d", tempStr, pipelineRegisterList.size());
//       delayRegTerm = new Terminal(this, nextPNet, currPNet, tempStr, acquiredDelay);
//       pipelineRegisterList.push_back(delayRegTerm);

//       // Add this new terminal to the sink list of the current pipelining net
//       currPNet->sinkVector.push_back(delayRegTerm);
//       currPNet->linkDelayVector.push_back(0);
//       currPNet->slackVector.push_back(0);
      
//       // Make this new terminal the source of the next pipelining net
//       nextPNet->source = delayRegTerm;
      
//       currMaxLatency += acquiredDelay;
//       currPNet = nextPNet;
//     }

//     //Now that we've matched target and current latency, add the sink to the end pipelining net
//     currPNet->sinkVector.push_back(sinkVector.at(i));
//     currPNet->linkDelayVector.push_back(0);
//     currPNet->slackVector.push_back(0);
//   }
//   print_Info();
// }

// Go through all sinks for this net and set up all of the pipelined nets assuming we are
// fully zipped (maximum register sharing).
// This will only get called on pipelined nets (nets with at least one sink with a non-zero delay).
// Notice that the meaning of pipelined nets is exactly the same as the one used during the routing phase.
// We assume that sinks are already sorted in non-decreasing order of latency.
// We keep to one linear string of pipelined nets and attach the sinks where they belong.
void Net::createZippedPipelinedNets(){
  int currMaxLatency = 0;
  PipeliningNet* currPNet;
  PipeliningNet* nextPNet;
  Terminal* delayRegTerm;
  //char tempStr [MAX_STR_LEN];
  string tempStr;
  int targetLatency;

  //Make a pipelined net for the source
  //strcpy(tempStr, name);
  //sprintf(tempStr, "%sPipelineNet%d", tempStr, pipelineNetList.size());
  tempStr = name + "PipelineNet" + itoa(pipelineNetList.size());
  currPNet = new PipeliningNet(tempStr, source, 0, this);
  pipelineNetList.push_back(currPNet);

  //Make the source terminal point to this new pipelined net
  currPNet->source->destPipeliningNet = currPNet;

  // Go through each sink in the netlist and figure out the least number of delay blocks we'll need
  // to meet the latency requirement for this sink
  for(int i = 0; i < sinkVector.size(); i++){
    targetLatency = latencyVector.at(i);

    //Create pipelining netlists until we get enough latency
    while(targetLatency != currMaxLatency){
      // Create a new pipelining netlist
      //strcpy(tempStr, name);
      //sprintf(tempStr, "%sPipelineNet%d", tempStr, pipelineNetList.size());
      tempStr = name + "PipelineNet" + itoa(pipelineNetList.size());

      // We guarantee that we are looking for a sink that needs a latency larger than what we have so far.
      // Thus, create a net with a higher latency
      nextPNet = new PipeliningNet(tempStr, currMaxLatency + 1, this);
      pipelineNetList.push_back(nextPNet);

      // Create the terminal that will represent the delay block
      //strcpy(tempStr, name);
      //sprintf(tempStr, "%sReg%d", tempStr, pipelineRegisterList.size());
      tempStr = name + "Reg" + itoa(pipelineRegisterList.size());
      delayRegTerm = new Terminal(this, currPNet, nextPNet, tempStr);
      pipelineRegisterList.push_back(delayRegTerm);

      // Add this new terminal to the sink list of the current pipelining net
      currPNet->sinkVector.push_back(delayRegTerm);
      currPNet->linkDelayVector.push_back(0);
      if(deltaDelayCost){
	currPNet->oldLinkDelayVector.push_back(0);
      }
      currPNet->slackVector.push_back(0);
      
      // Make this new terminal the source of the next pipelining net
      nextPNet->source = delayRegTerm;
      
      currMaxLatency += 1;
      currPNet = nextPNet;
    }

    //Now that we've matched target and current latency, add the sink to the end pipelining net
    currPNet->sinkVector.push_back(sinkVector.at(i));
    sinkVector.at(i)->sourcePipeliningNet = currPNet;
    currPNet->linkDelayVector.push_back(0);
    if(deltaDelayCost){
      currPNet->oldLinkDelayVector.push_back(0);
    }
    currPNet->slackVector.push_back(0);
  }

  //Set the pipeline register ID so that when we create more registers 
  // while unzipping and zipping, we will have consistent numbering
  registerID = pipelineRegisterList.size();

  //print_Info();
}


// Remove the pipeline net information we generated during placement before
// we do routing
void Net::cleanUpPipelinePlacementInformation(){  
  for(list<PipeliningNet *>::iterator pNetIter = pipelineNetList.begin(); pNetIter != pipelineNetList.end(); pNetIter++){
    delete (*pNetIter);
  }
  pipelineNetList.clear();

  for(list<Terminal *>::iterator dNodeIter = pipelineRegisterList.begin(); dNodeIter != pipelineRegisterList.end(); dNodeIter++){
    delete (*dNodeIter);
  }
  pipelineRegisterList.clear();

  //Go and erase and reference from sources and sinks
  if(source->sourcePipeliningNet){
    ERROR("Somehow source of net has a source pipelining net!");
  }
  source->destPipeliningNet = 0;
  for(vector<Terminal*>::iterator sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
    (*sinkIter)->sourcePipeliningNet = 0;
    if((*sinkIter)->destPipeliningNet){
      ERROR("Somehow sink of net has a destination pipelining net!");
    }
  }
}

//Update the net timing info based on the placement
void Net::updatePlacementTiming(){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;
  
  int i;


  if(pipelined){
    //Go through all of the pipelined nets and update their timing info
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      (*iter)->updatePlacementTiming();
    }
  }
  else{
    sourceType = source->getHostTimingMoveType();
    sourceBank = source->getHostBankNum();
    sourceRow = source->getLogicBlockRow();
    sourceCol = source->getLogicBlockCol();

    // Cycle through the sinks
    i = 0;
    for(vector<Terminal *>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
      sinkType = (*iter)->getHostTimingMoveType();
      sinkBank = (*iter)->getHostBankNum();
      sinkRow = (*iter)->getLogicBlockRow();
      sinkCol = (*iter)->getLogicBlockCol();

      //Fill in the linkDelayVector based on the timing table
      linkDelayVector.at(i) = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);

      if(deltaDelayCost){
	oldLinkDelayVector.at(i) = linkDelayVector.at(i);
      }
      //cout << "Link delay between source " << source->name << " and sink " << (*iter)->name;
      //cout << " estimated to be " << linkDelayVector.at(i) << endl;
    }
  }
}

// Calculate the estimated wirelength of this net - including any pipelined nets that might be included
void Net::calculatePlacementCost(bool timing_driven, double criticalPathDelay){
  int currRow;
  int currCol;
  int i;
  double criticality;
  double crossingFactor;

  wireCost = 0.0;
  timingCost = 0.0;

  if(pipelined){
    //Go through all of the pipelined nets and figure out their individual costs
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      (*iter)->calculatePlacementCost(timing_driven, criticalPathDelay);
      wireCost += (*iter)->wireCost;
      timingCost += (*iter)->timingCost;

      //printf("Pipelining Net %s has Wire Cost = %f\n", (*iter)->name, (*iter)->wireCost);
      //printf("Pipelining Net %s has Timing Cost = %e\n", (*iter)->name, (*iter)->timingCost);
    }
  }
  else{
    // If the net is not a pipelined net, simply figure out the min and max locations of all source and sink terminals
    // The source terminal must be mapped to a logic block or a delay block
    currRow = source->getLogicBlockRow();
    currCol = source->getLogicBlockCol();
      
    maxRow = currRow;
    minRow = currRow;
    maxCol = currCol;
    minCol = currCol;

    // Cycle through the sinks to see if any change the bounding box
    //  and also calculate timing cost
    i = 0;
    for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
      currRow = (*iter)->getLogicBlockRow();
      currCol = (*iter)->getLogicBlockCol();
      if(currRow < minRow){
	minRow = currRow;
      }
      else if (currRow > maxRow){
	maxRow = currRow;
      }
      if(currCol < minCol){
	minCol = currCol;
      }
      else if(currCol > maxCol){
	maxCol = currCol;
      }
      
      if(timing_driven){	
	//If we are doing timing-driven placement, the cost of this net should be calculated as:
	//   criticality of (source, sink) = 1 - (slack of (source, sink) / critical path delay)
	//   timing cost = delay of (source, sink) * criticality of (source, sink) ^ criticality exponent
	criticality = 1.0 - (slackVector.at(i) / criticalPathDelay);
	criticality = doubmax(criticality, 0.0);
	timingCost += linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent);
      }
    }

    if(maxCutCost){
      // Increment the proper cut cost vector locations based on max and min Col locations
      // Should we have and adjustment factor for the number of sinks?
      for(i = minCol; i < maxCol; i++){
	FPGA->cutCostVector.at(i) += 1.0;
      }
    }

    //if(wireCost != (double)(maxCol - minCol + maxRow - minRow)){
    //  cout << "net " << name << " has wire cost discrepancy!" <<endl;
    //}
    //wireCost = (double)(maxCol - minCol + maxRow - minRow);

    if(sinkVector.size() > 49){
      crossingFactor = 2.7933 + 0.02616 * (sinkVector.size() - 49);
    }
    else{
      crossingFactor = cross_count[sinkVector.size()];
    }

    //This is to implement the same cost function as in VPR - place.c:2623
    //Unfortunately, I don't want to go through the trouble of calculating their
    // chanx_place_cost_fac function, so I'll just use 0.01 so that I can compare
    // with their 4lut_sanitized architecture.  It shouldn't make a difference
    //Since we assume that all architectures have uniform channels.
    if((maxCol - minCol) == 0 && (maxRow - minRow) == 0){
      wireCost = 0;
    }
    else{
      wireCost = ((double)(maxCol - minCol + 1)) * crossingFactor * 0.01;
      wireCost += ((double)(maxRow - minRow + 1)) * crossingFactor * 0.01;
    }
  }
}

void Net::updateSinkDelayAndSlack(Terminal *sink){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;

  double newDelay;
  int i;

  if(pipelined){
    //If this sink is attached to a pipelined net, we don't have to worry about it here
    // the corresponding pipelined net updateSinkDelay will be called
  }
  else{
    //cout << "Updating non-pipelined Sink Delay and Slack for net " << name << endl;
    sourceType = source->getHostTimingMoveType();
    sourceBank = source->getHostBankNum();
    sourceRow = source->getLogicBlockRow();
    sourceCol = source->getLogicBlockCol();

    //cout << "\tSource is at "<< sourceCol << ", " << sourceRow << endl;

    //Find this terminal in the sink vector
    i = 0;
    for(vector<Terminal *>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
      if((*iter) == sink){
	sinkType = (*iter)->getHostTimingMoveType();
	sinkBank = (*iter)->getHostBankNum();
	sinkRow = (*iter)->getLogicBlockRow();
	sinkCol = (*iter)->getLogicBlockCol();

	//cout << "\tSink " << (*iter)->name << " is at "<< sinkCol << ", " << sinkRow << endl;

	//Figure out the new delay for this source/sink pair
	newDelay = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);

	//If this new delay is faster or slower, update the slack
	if(updateSlack){
	  slackVector.at(i) = slackVector.at(i) + (linkDelayVector.at(i) - newDelay);
	}
	
	//Update the delay
	if(deltaDelayCost){
	  oldLinkDelayVector.at(i) = linkDelayVector.at(i);
	}
	linkDelayVector.at(i) = newDelay;
	//cout << "Updated link delay between source " << source->name << " and sink " << (*iter)->name;
	//cout << " estimated to be " << linkDelayVector.at(i) << " with slack " << slackVector.at(i) << endl;

	return;
      }
    }
    ERROR("We didn't find this sink!");
  }
}

void Net::updateSourceDelaysAndSlacks(){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;

  double newDelay;
  int i;

  if(pipelined){
    //If this source is attached to a pipelined net, we don't have to worry about it here
    // the corresponding pipelined net updateSourceDelay will be called
  }
  else{
    //cout << "Updating non-pipelined source Delay and Slack for net " << name << endl;
    sourceType = source->getHostTimingMoveType();
    sourceBank = source->getHostBankNum();
    sourceRow = source->getLogicBlockRow();
    sourceCol = source->getLogicBlockCol();

    //cout << "\tSource is at "<< sourceCol << ", " << sourceRow << endl;


    //Update the delay for all of the sinks
    i = 0;
    for(vector<Terminal *>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
      sinkType = (*iter)->getHostTimingMoveType();
      sinkBank = (*iter)->getHostBankNum();
      sinkRow = (*iter)->getLogicBlockRow();
      sinkCol = (*iter)->getLogicBlockCol();

      //cout << "\tSink " << (*iter)->name << " is at "<< sinkCol << ", " << sinkRow << endl;

      //Figure out the new delay for this source/sink pair
      newDelay = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);
      
      //If this new delay is faster or slower, update the slack
      if(updateSlack){
	slackVector.at(i) = slackVector.at(i) + (linkDelayVector.at(i) - newDelay);
      }
      
      //Update the delay
      if(deltaDelayCost){
	oldLinkDelayVector.at(i) = linkDelayVector.at(i);
      }
      linkDelayVector.at(i) = newDelay;
      //cout << "Updated link delay between source " << source->name << " and sink " << (*iter)->name;
      //cout << " estimated to be " << linkDelayVector.at(i) << " with slack " << slackVector.at(i) << endl;
      
    }
  }
}

void Net::reCalculatePlacementCost(bool timing_driven, double criticalPathDelay){
  int currRow;
  int currCol;
  int minRowTemp;
  int maxRowTemp;
  int minColTemp;
  int maxColTemp;
  int i;
  double criticality;
  double crossingFactor;
  
  if(pipelined){
//     newWireCost = 0.0;
//     newTimingCost = 0.0;

//     //Go through all of the pipelined nets and figure out their delta costs
//     for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
//       (*iter)->reCalculatePlacementCost(timing_driven, criticalPathDelay);
//       newWireCost += (*iter)->newWireCost;

//       if(timing_driven){
// 	newTimingCost += (*iter)->newTimingCost;
//       }
//       //printf("Pipelining Net %s has new wire Cost = %f\n", (*iter)->name, (*iter)->newWireCost);
//       //printf("Pipelining Net %s has new timing Cost = %e\n", (*iter)->name, (*iter)->newTimingCost);
//     }
    ERROR("Why is this getting called on a pipelined net?");
  }

  else{ // Figure out the min and max locations of all source and sink terminals
    currRow = source->getLogicBlockRow();
    currCol = source->getLogicBlockCol();
    
    maxRowTemp = currRow;
    minRowTemp = currRow;
    maxColTemp = currCol;
    minColTemp = currCol;
    
    // Cycle through the sinks to see if any change the bounding box
    i = 0;
    newWireCost = 0.0;
    newTimingCost = 0.0;
    if(deltaDelayCost){
      newTimingCostPrime = 0.0;
    }

    for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
      currRow = (*iter)->getLogicBlockRow();
      currCol = (*iter)->getLogicBlockCol();

      if(currRow < minRowTemp){
	minRowTemp = currRow;
      }
      else if(currRow > maxRowTemp){
	maxRowTemp = currRow;
      }
      if(currCol < minColTemp){
	minColTemp = currCol;
      }
      else if (currCol > maxColTemp){
	maxColTemp = currCol;
      }

      if(timing_driven){
	//If we are doing timing-driven placement, the cost of this net should be calculated as:
	//   criticality of (source, sink) = 1 - (slack of (source, sink) / critical path delay)
	//   timing cost = delay of (source, sink) * criticality of (source, sink) ^ criticality exponent
	//cout << "Slack = " << slackVector.at(i) << " criticality = " << 1.0 - (slackVector.at(i) / criticalPathDelay) << endl;
	criticality = 1.0 - (slackVector.at(i) / criticalPathDelay);
	criticality = doubmax(criticality, 0.0);
	//cout << "LinkDelay = " << linkDelayVector.at(i) << " LinkTimingCost = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) <<endl;
	
	newTimingCost += linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent);

	if(deltaDelayCost){
	  if(oldLinkDelayVector.at(i) < 0.0){
	    ERROR("Old link delay is not set!");
	  }
	  newTimingCostPrime = (linkDelayVector.at(i) - oldLinkDelayVector.at(i)) * pow(criticality, tdPlacementCritExponent);
	}

	//cout << "TimingCost for sink " << (*iter)->name << " = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) << endl;
      }
    }

    if(maxCutCost){
      //We are safe from updating this net twice (just in case the blocks we moved were part of the same net
      // because the minCol and maxCol will only be changed once.
      //The new minCol might be larger than the old minCol (we decreased the span) - we should decrement those locations
      for(i = minCol; i < minColTemp; i++){
	FPGA->cutCostVector.at(i) += -1.0;
      }
      //The new minCol might be smaller than the old minCol (we increased the span) - we should increment those locations
      for(i = minColTemp; i < minCol; i++){
	FPGA->cutCostVector.at(i) += 1.0;
      }
      //The new maxCol might be smaller than the old maxCol (we decreased the span) - we should decrement those locations
      for(i = maxColTemp; i < maxCol; i++){
	FPGA->cutCostVector.at(i) += -1.0;
      }
      //The new maxCol might be larger than the old maxCol (we increased the span) - we should increment those locations
      for(i = maxCol; i < maxColTemp; i++){
	FPGA->cutCostVector.at(i) += 1.0;
      }
    }
    
    //Set the values in the main net variables
    minRow = minRowTemp;
    maxRow = maxRowTemp;  
    minCol = minColTemp;
    maxCol = maxColTemp;
    
    //newWireCost = (double)(maxCol - minCol + maxRow - minRow);

    if(sinkVector.size() > 49){
      crossingFactor = 2.7933 + 0.02616 * (sinkVector.size() - 49);
    }
    else{
      crossingFactor = cross_count[sinkVector.size()];
    }

    //This is to implement the same cost function as in VPR - place.c:2623
    //Unfortunately, I don't want to go through the trouble of calculating their
    // chanx_place_cost_fac function, so I'll just use 0.01 so that I can compare
    // with their 4lut_sanitized architecture.  It shouldn't make a difference
    //Since we assume that all architectures have uniform channels.
    if((maxRow - minRow) == 0 && (maxCol - minCol) == 0){
      newWireCost = 0;
    }
    else{
      newWireCost = ((double)(maxRow - minRow + 1)) * crossingFactor * 0.01;
      newWireCost += ((double)(maxCol - minCol + 1)) * crossingFactor * 0.01;
     
    }
    //cout << "New Timing cost for net " << name << " = " << newTimingCost << endl;
    //cout << "New Wire Cost for net " << name << " = " << newWireCost << endl;
  }
}

double Net::getDeltaWireCost(){
  double dWireCost = 0.0;

  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      dWireCost += (*iter)->getDeltaWireCost();
    }
    return (dWireCost);
  }
  else{ 
    return (newWireCost - wireCost);
  }
}

void Net::overWriteOldWireCost(){
  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      (*iter)->overWriteOldWireCost();
    }
  }
  else{ 
    wireCost = newWireCost;
  }
}

double Net::getDeltaTimingCost(){
  double dTimingCost = 0.0;

  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      dTimingCost += (*iter)->getDeltaTimingCost();
    }
    return (dTimingCost);
  }
  else{
    //cout << "Net " << name << " newTiming cost = " << newTimingCost << " old timing Cost = " << timingCost <<endl;
    return (newTimingCost - timingCost);
  }
}

double Net::getDeltaTimingCostPrime(){
  double dTimingCostPrime = 0.0;

  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      dTimingCostPrime += (*iter)->getDeltaTimingCostPrime();
    }
    return (dTimingCostPrime);
  }
  else{
    return (newTimingCostPrime);
  }
}

void Net::overWriteOldTimingCost(){
  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      (*iter)->overWriteOldTimingCost();
    }
  }
  else{
    timingCost = newTimingCost;
  }
}

void Net::clearTimingCostPrime(){
  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      (*iter)->clearTimingCostPrime();
    }
  }
  else{
    newTimingCostPrime = 0.0;
  }
}

void Net::zeroSlackVectors(){
  if(pipelined){
    for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
      (*iter)->zeroSlackVectors();
    }
  }
  else{
    for(vector <double>::iterator iter = slackVector.begin(); iter != slackVector.end(); iter++){
      (*iter) = 0.0;
    }
  }
}

void Net::removePipelineRegFromPRList(Terminal *pipeReg){
  list<Terminal *>::iterator rlIter;
    
  for(rlIter = pipelineRegisterList.begin(); rlIter != 
	pipelineRegisterList.end(); rlIter++){
    if(*rlIter == pipeReg){
      break;
    }
  }

  if(rlIter == pipelineRegisterList.end()){
    ERROR("Register not found!");
  }
  else{
    pipelineRegisterList.erase(rlIter);
  }
}

void Net::removePipelineNetFromPNList(PipeliningNet *pNet){
  list<PipeliningNet *>::iterator pIter;
  
  for(pIter = pipelineNetList.begin(); pIter != pipelineNetList.end(); pIter++){
    if(*pIter == pNet){
      break;
    }
  }
  if(pIter == pipelineNetList.end()){
    ERROR("Net not found!");
  }
  else{
    pipelineNetList.erase(pIter);
  }
}

//Go through all of the sinks of this net and reduce their latency
void Net::subtractLatencyFromAllSinks(int numRemoved){
  for(vector<int>::iterator iter = latencyVector.begin(); 
      iter != latencyVector.end(); iter++){
    (*iter) -= numRemoved;
  }
}

//Go through the vector of sinks and find this sink.  Then reduce the 
// latency on it.
void Net::subtractLatencyFromThisSink(Terminal *sink, int numRemoved){
  vector<Terminal*>::iterator iter;
  int i;

  for(iter = sinkVector.begin(), i = 0; iter != sinkVector.end(); iter++, i++){
    if((*iter) == sink){
      break;
    }
  }
  
  if(iter == sinkVector.end()){
    ERROR("Sink not found!");
  }
  else{
    latencyVector.at(i) -= numRemoved;
  }
}

//We would like to convert this net into a pipelined net and return a pointer
// to the new net
PipeliningNet* Net::makePipelined(Netlist *netlist){
  //char tempStr [MAX_STR_LEN];
  string tempStr;
  PipeliningNet *pNet;
  vector<Terminal*>::iterator sinkIter;
  vector <double>::iterator ldIter;
  vector <double>::iterator sIter;

  //We will no longer be using the conventional net, so we have to take out
  // the contribution of the conventional net's wire and timing cost
  netlist->wireCostSurplus += wireCost;
  netlist->timingCostSurplus += timingCost;

  pipelined = true;

  //Make a new pipelined net
  //strcpy(tempStr, name);
  //sprintf(tempStr, "%sPipelineNet%d", tempStr, registerID);
  tempStr = name + "PipelineNet" + itoa(registerID);
  //cout << "Converting net " << name << " to  pipelined net" << tempStr << endl;
  pNet =  new PipeliningNet(tempStr, source, 0, this);
  source->destPipeliningNet = pNet;
  pipelineNetList.push_back(pNet);

  //Add all of the sinks to this pipelining net
  for(sinkIter = sinkVector.begin(), ldIter = linkDelayVector.begin(), sIter = slackVector.begin();
      sinkIter != sinkVector.end(), ldIter != linkDelayVector.end(), sIter != slackVector.end();
      sinkIter++, ldIter++, sIter++){
    pNet->pushSinkPlusData(*sinkIter, *ldIter, *sIter);
    (*sinkIter)->sourcePipeliningNet = pNet;
  }
  return pNet;
}

//Find the mention of this sink in the latency vector and increase the
// value by this much.
void Net::increaseSinkLatency(Terminal *sink, int additionalLatency){
  vector<Terminal*>::iterator iter;
  int select;
  
  for(iter = sinkVector.begin(), select = 0; iter != sinkVector.end(); iter++, select++){
    if(*iter == sink){
      break;
    }
  }

  if(iter == sinkVector.end()){
    ERROR("Sink not found!");
  }
  else{
    latencyVector.at(select) += additionalLatency;
  }
}

//Incrase the latency of all the sinks
void Net::increaseAllSinkLatency(int additionalLatency){
  for(  vector<int>::iterator iter = latencyVector.begin(); iter != latencyVector.end(); iter++){
    (*iter) += additionalLatency;
  }
}

//Find this sink and return the latency
int Net::returnLatencyOfSink(Terminal *sink){
  vector<Terminal*>::iterator iter;
  int select;
  
  for(iter = sinkVector.begin(), select = 0; iter != sinkVector.end(); iter++, select++){
    if(*iter == sink){
      break;
    }
  }

  if(iter == sinkVector.end()){
    ERROR("Sink not found!");
  }
  else{
    return latencyVector.at(select);
  }
}

void Net::setLatencyOfSink(Terminal *sink, int lat){
  vector<Terminal*>::iterator iter;
  int select;
  
  for(iter = sinkVector.begin(), select = 0; iter != sinkVector.end(); iter++, select++){
    if(*iter == sink){
      break;
    }
  }

  if(iter == sinkVector.end()){
    ERROR("Sink not found!");
  }
  else{
    latencyVector.at(select) = lat;
  }
}

//Go through all of the sinks and update their latency
void Net::updateLatencyAllSinks(){
  int select;
  vector<Terminal*>::iterator iter;
  
  for(iter = sinkVector.begin(), select = 0; iter != sinkVector.end(); iter++, select++){
    if((*iter)->sourcePipeliningNet){
      latencyVector.at(select) = (*iter)->sourcePipeliningNet->latency;
    }
    else{
      ERROR("Sink has no source pipelining net!");
    }
  }
}

//Go through all of the nets and make sure that all pipelining nets have the correct parent net
void Net::checkNet(){
  for(list<PipeliningNet *>::iterator iter = pipelineNetList.begin(); iter != pipelineNetList.end(); iter++){
    if((*iter)->parentNet != this){
      ERROR("Pipelining Net does not have same parent as host net!");
    }
    (*iter)->checkSourcesAndSinks();
  }
  for(list<Terminal *>::iterator iter = pipelineRegisterList.begin(); iter != pipelineRegisterList.end(); iter++){
    (*iter)->checkSourceAndDestNet();
  }

  //Check to make sure that all of the registers in this net are not nerfed.
  for(list<Terminal *>::iterator iter = pipelineRegisterList.begin(); iter != pipelineRegisterList.end(); iter++){
    if((*iter)->retimeNerfed){
      ERROR("Nerfed register here!");
    }
  }
}

//Go through this netlist's pipelining registers and print their placement in order of increasing latency.
//That way, when we want to suck back in a placement we can put the registers back in banks in the correct order.
void Net::printVPRPlacement(FILE *fp){
  for(int i = 0; i < netMaxLatency; i++){
    for(list<Terminal *>::iterator prIter = pipelineRegisterList.begin(); prIter != pipelineRegisterList.end(); prIter++){
      if((*prIter)->sourcePipeliningNet->latency == i){
	fprintf(fp,"\nPLACEMENT: %s\t\t%d\t%d\t%d",(*prIter)->get_name().c_str(), (*prIter)->rb->getHostCLB()->col, 
		(*prIter)->rb->getHostCLB()->row, (*prIter)->rb->getHostBankNum());
      }
    }
  }
}

bool Net::foundNewSink(routingNode *node){
  for(vector<Terminal*>::iterator sinkVectorIter = sinkVector.begin(); sinkVectorIter != sinkVector.end(); sinkVectorIter++){
    if(node == (*sinkVectorIter)->get_rNode()){
      return true;
    }
  }
  return false;
}

//Add all of the pipelining registers in this net to the list
void Net::addPipelineRegistersToList(list <Terminal*> *PRList){
  for(list<Terminal *>::iterator iter = pipelineRegisterList.begin(); iter != pipelineRegisterList.end(); iter++){
    PRList->push_back(*iter);
  }
}

//Print out the Manhattan distance between the source and sinks of this net
void Net::printDistanceToAllSinks(){
  int sourceRow = source->getLogicBlockRow();
  int sourceCol = source->getLogicBlockCol();

  for(vector<Terminal*>::iterator sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
    cout << "Distance to sink = " << abs(sourceCol - (*sinkIter)->getLogicBlockCol()) + abs(sourceRow - (*sinkIter)->getLogicBlockRow()) << endl;
  }

}

//Find the Manhattan distance to the closest sink
int Net::minDistanceToAllSinks(){
  int minDistance = BIGINTEGER;

  int sourceRow = source->getLogicBlockRow();
  int sourceCol = source->getLogicBlockCol();

  for(vector<Terminal*>::iterator sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
    if(abs(sourceCol - (*sinkIter)->getLogicBlockCol()) + abs(sourceRow - (*sinkIter)->getLogicBlockRow()) < minDistance){
      minDistance = abs(sourceCol - (*sinkIter)->getLogicBlockCol()) + abs(sourceRow - (*sinkIter)->getLogicBlockRow());
    }
  }
  return minDistance;
}

//Pipeline Aware Stuff
// Make the source routing node the root of rTree
// Then increment the occupancy of the source node
void PipeliningNet::init_rTreeRoot(){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");
  treeNode* root_node=new treeNode;
  root_node->set_rNode(source->get_rNode());
  rTree.set_root(root_node);

  //Set the back cost of the source correctly here
  //I should have something more than just it's own intrinsic delay
  root_node->linkDelay = root_node->rNode->intrinsicDelay;

  //Increment occupancy and put it into the tree at latency 0
  //source->rNode->set_occupancy(source->rNode->occupancy + 1);
  //source->rNode->set_inPRTree(0);
}

// Return tree node that holds the routing node pointed to by rnode
treeNode* PipeliningNet::find_treeNode(routingNode* rnode){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");
  if(rnode==0)
    ERROR("cant find NULL routing node in tree...");
  treeNode* tree_node=rTree.findNode(rTree.get_root(),rnode);
  if(tree_node==0)
    ERROR("uh,oh! could not find routing node in tree...");

//   cout<<endl<<"FOUND tree node corresponding to "<<tree_node->get_rNode()->get_name();

  return tree_node;
}

// Print the current route tree of this net
void PipeliningNet::print_rTree(){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");
  rTree.print_info();
}

// Get the total number of nodes in the current route tree
int PipeliningNet::get_rNodeCount(){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");
  int node_count=0;
  countNodes(rTree.get_root(),&node_count);
  return node_count;
}

// Count number of nodes in the current route tree
void PipeliningNet::countNodes(treeNode* r,int* count_ptr){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    countNodes(child,count_ptr);
  tree_node->remove_childListItr();

  if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
    *count_ptr+=1;
}

//Copy the contents of the routing tree into tempRTree
void PipeliningNet::copyRTree(){
  //Remove the old contents of tempRTree
  tempRTree.clear();
  //cout << "Cleared tempRTree" << endl;
  //tempRTree.print_info();
 
  //Copy rTree into tempRTree
  tempRTree.populateWith(rTree.get_root());
  //cout << "Copied tempRTree" << endl;
  //tempRTree.print_info();
}

// Print details about this net
void PipeliningNet::print_Info(){
  int i;
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");

  cout<<endl<<"\t===================================";
  cout<<endl<<"\t             PRINT PIPELINED NET              ";
  cout<<endl<<"\t===================================";

  cout<<endl<<endl<<"\tBEGIN PRINT INFO FOR NET "<<name;

  cout <<endl << "Wire cost = " << wireCost;
  cout <<endl << "Timing cost = " << timingCost;

  cout<<endl<<"\tLatency = " << latency;
  //  if(source->get_lBlock()){

  //}
  //else{
  cout<<endl<<"\tSOURCE Node IS "<<source->get_name();
  cout << "Required Time = " << source->requiredTime;
  cout << " Departure time = " << source->departureTime;
    //}
  
  cout<<endl<<"\tSINKS ARE (delay/slack)-->";
  vector<Terminal*>::iterator term_itr = sinkVector.begin();
  for(i = 0;term_itr!=sinkVector.end();term_itr++, i++){
    //if((*term_itr)->get_lBlock()){
    //  cout<<endl<<"\t\tLogic Block "<<(*term_itr)->get_lBlock()->get_name();
    //}
    //else{
    //  cout<<endl<<"\t\tDelay Node #"<<(*term_itr)->get_rNode()->get_name();
    //}
    cout<<endl<<"\t\t"<<(*term_itr)->get_name();
    cout<<" (" << linkDelayVector[i] << "/" << slackVector[i] << ")";

    cout << "Required Time = " << (*term_itr)->requiredTime;
    cout << " Departure time = " << (*term_itr)->departureTime;
  }
  
  cout << endl << "\tPipelined Net MinRow = " << minRow;
  cout << endl << "\tPipelined Net MaxRow = " << maxRow;
  cout << endl << "\tPipelined Net MinCol = " << minCol;
  cout << endl << "\tPipelined Net MaxCol = " << maxCol;

  cout <<endl<< "\tRoute Tree" << endl;
  rTree.print_info();
  //cout<<endl; 
}

//Rip up the routing tree and decrement the occupancy/sharing
//We may or may not decrement the occupany on the source of the net
void  PipeliningNet::ripUpRTree(bool countSource){
  if(countSource){
    rTree.postorderTraverse(rTree.root, 86);
  }
  else{
    rTree.postorderTraverse(rTree.root, 87);
  }
}

void PipeliningNet::reCalculateOccupancy(){
  //If we are the 0-latency net, increment the occupany of our source, if not, start with the children
  if(latency == 0){
    if(rTree.root){
      rTree.reCalculateOccupancy(rTree.root);
    }
  }
  else{
    if(rTree.root){
      for(list<treeNode*>::iterator childIter = rTree.root->childList.begin(); childIter != rTree.root->childList.end(); childIter++){
	rTree.reCalculateOccupancy(*childIter);
      }
    }
  }
}

//Update the net timing info based on the placement
void PipeliningNet::updatePlacementTiming(){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;
  int i;

  //A pipelining net might have a source that is a logic block
  // or it might be a delay node that has no logic block associated with it
  // in this case, it will try looking at the delay block associated with the terminal
  sourceType = source->getHostTimingMoveType();
  sourceBank = source->getHostBankNum();
  sourceRow = source->getLogicBlockRow();
  sourceCol = source->getLogicBlockCol();
    
  // Cycle through the sinks
  i = 0;
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    sinkType = (*iter)->getHostTimingMoveType();
    sinkBank = (*iter)->getHostBankNum();
    sinkRow = (*iter)->getLogicBlockRow();
    sinkCol = (*iter)->getLogicBlockCol();

    //Fill in the linkDelayVector based on the timing table
    linkDelayVector.at(i) = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);
    //cout << "Link delay between source " << source->name << " and sink " << (*iter)->name;
    //cout << " estimated to be " << linkDelayVector.at(i) << endl;
    if(deltaDelayCost){
      oldLinkDelayVector.at(i) = linkDelayVector.at(i);
    }
  }
}

void PipeliningNet::printCriticalPath(){
  vector <double>::iterator iter;
  int i;

  for(iter = slackVector.begin(), i = 0; iter != slackVector.end(); iter++, i++){
    if(*iter < 1E-20){
	cout << "Net " << name << " link between source " << source->name << " (node " << source->rNode->name << ") and sink " << (sinkVector.at(i))->name;
	cout << " (node " << (sinkVector.at(i))->rNode->name << ") has slack " << *iter << endl;

      rTree.printPath(sinkVector.at(i));
    }
  }
}

void PipeliningNet::updateSinkDelayAndSlack(Terminal *sink){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;
  double newDelay;

  int i;

  sourceType = source->getHostTimingMoveType();
  sourceBank = source->getHostBankNum();
  sourceRow = source->getLogicBlockRow();
  sourceCol = source->getLogicBlockCol();

  //cout << "Updating pipelined net Sink Delay and Slack for net " << name << endl;
  //cout << "\tSource is at "<< sourceCol << ", " << sourceRow << endl;

  //Find this terminal in the sink vector
  i = 0;
  for(vector<Terminal *>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    if((*iter) == sink){
      sinkType = (*iter)->getHostTimingMoveType();
      sinkBank = (*iter)->getHostBankNum();
      sinkRow = (*iter)->getLogicBlockRow();
      sinkCol = (*iter)->getLogicBlockCol();

      //cout << "\tSink " << (*iter)->name << " is at "<< sinkCol << ", " << sinkRow << endl;

      //Figure out the new delay for this source/sink pair
      newDelay = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);

      //If this new delay is faster or slower, update the slack
      if(updateSlack){
	slackVector.at(i) = slackVector.at(i) + (linkDelayVector.at(i) - newDelay);
      }
      
      //Update the delay
      if(deltaDelayCost){
	oldLinkDelayVector.at(i) = linkDelayVector.at(i);
      }
      linkDelayVector.at(i) = newDelay;
      //cout << "\t\tUpdated link delay between source " << source->name << " and sink " << (*iter)->name;
      //cout << " estimated to be " << linkDelayVector.at(i) << " with slack " << slackVector.at(i) << endl;
      
      return;
    }
  }
  ERROR("We didn't find this sink!");
}

void PipeliningNet::updateSourceDelaysAndSlacks(){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;
  double newDelay;
  int i;

  sourceType = source->getHostTimingMoveType();
  sourceBank = source->getHostBankNum();
  sourceRow = source->getLogicBlockRow();
  sourceCol = source->getLogicBlockCol();
  
  //cout << "Updating pipelined net Source Delay and Slack for net " << name << endl;
  //cout << "\tSource is at "<< sourceCol << ", " << sourceRow << endl;

  //Update the delay for all of the sinks
  i = 0;
  for(vector<Terminal *>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    sinkType = (*iter)->getHostTimingMoveType();
    sinkBank = (*iter)->getHostBankNum();
    sinkRow = (*iter)->getLogicBlockRow();
    sinkCol = (*iter)->getLogicBlockCol();

    //cout << "\tSink " << (*iter)->name << " is at "<< sinkCol << ", " << sinkRow << endl;

    //Figure out the new delay for this source/sink pair
    newDelay = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);
    
    //If this new delay is faster or slower, update the slack
    if(updateSlack){
      slackVector.at(i) = slackVector.at(i) + (linkDelayVector.at(i) - newDelay);
    }

    //Update the delay
    if(deltaDelayCost){
      oldLinkDelayVector.at(i) = linkDelayVector.at(i);
    }
    linkDelayVector.at(i) = newDelay;
    //cout << "\t\tUpdated link delay between source " << source->name << " and sink " << (*iter)->name;
    //cout << " estimated to be " << linkDelayVector.at(i) << " with slack " << slackVector.at(i) << endl;
  }
}

void PipeliningNet::updateSourceDelaysAndSlacksForwards(){
  int sourceRow;
  int sourceCol;
  e_timingMoveBlockType sourceType;
  int sourceBank;
  int sinkRow;
  int sinkCol;
  e_timingMoveBlockType sinkType;
  int sinkBank;
  double newDelay;
  int i;

  //cout << "Updating all sink delay and slack info for net " << name << endl;

  sourceType = source->getHostTimingMoveType();
  sourceBank = source->getHostBankNum();
  sourceRow = source->getLogicBlockRow();
  sourceCol = source->getLogicBlockCol();
  
  //cout << "Updating pipelined net Source Delay and Slack for net " << name << endl;
  //cout << "\tSource is at "<< sourceCol << ", " << sourceRow << endl;

  //Update the delay for all of the sinks
  i = 0;
  for(vector<Terminal *>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    sinkType = (*iter)->getHostTimingMoveType();
    sinkBank = (*iter)->getHostBankNum();
    sinkRow = (*iter)->getLogicBlockRow();
    sinkCol = (*iter)->getLogicBlockCol();

    //cout << "\tSink " << (*iter)->name << " is at "<< sinkCol << ", " << sinkRow << endl;

    //Figure out the new delay for this source/sink pair
    newDelay = FPGA->getTimingTable(sourceType, sourceBank, sinkType, sinkBank, sinkRow - sourceRow, sinkCol - sourceCol);
    
    //If this new delay is faster or slower, update the slack
    slackVector.at(i) = slackVector.at(i) + (linkDelayVector.at(i) - newDelay);

    //Update the delay
    if(deltaDelayCost){
      oldLinkDelayVector.at(i) = linkDelayVector.at(i);
    }
    linkDelayVector.at(i) = newDelay;
    //cout << "\t\tUpdated link delay between source " << source->name << " and sink " << (*iter)->name;
    //cout << " estimated to be " << linkDelayVector.at(i) << " with slack " << slackVector.at(i) << endl;
    
    //If this sink is nerfed, call this on the downstream net too
    if((*iter)->retimeNerfed){
      (*iter)->destPipeliningNet->updateSourceDelaysAndSlacksForwards();
    }
  }
}

void PipeliningNet::calculatePlacementCost(bool timing_driven, double criticalPathDelay){
  int currRow;
  int currCol;
  int i;
  double criticality;
  double crossingFactor;

  // Figure out the min and max locations of all source and sink terminals
  currRow = source->getLogicBlockRow();
  currCol = source->getLogicBlockCol();
  
  maxRow = currRow;
  minRow = currRow;
  maxCol = currCol;
  minCol = currCol;
  
  // Cycle through the sinks to see if any change the bounding box
  i = 0;
  //wireCost = 0.0;
  timingCost = 0.0;
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    currRow = (*iter)->getLogicBlockRow();
    currCol = (*iter)->getLogicBlockCol();

    if(currRow < minRow){
      minRow = currRow;
    }
    else if(currRow > maxRow){
	maxRow = currRow;
    }
    if(currCol < minCol){
      minCol = currCol;
    }
    else if (currCol > maxCol){
      maxCol = currCol;
    }
    
    if(timing_driven){
      //If we are doing timing-driven placement, the cost of this net should be calculated as:
      //   criticality of (source, sink) = 1 - (slack of (source, sink) / critical path delay)
      //   timing cost = delay of (source, sink) * criticality of (source, sink) ^ criticality exponent
      //cout << "Slack = " << slackVector.at(i) << " criticality = " << 1.0 - (slackVector.at(i) / criticalPathDelay) << endl;
      criticality = 1.0 - (slackVector.at(i) / criticalPathDelay);
      criticality = doubmax(criticality, 0.0);
      //cout << "LinkDelay = " << linkDelayVector.at(i) << " LinkTimingCost = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) <<endl;

      timingCost += linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent);
      //cout << "Running timingCost = " << timingCost << endl;
    }
  }

  if(maxCutCost){
    // Increment the proper cut cost vector locations based on max and min Col locations
    // Should we have and adjustment factor for the number of sinks?
    for(i = minCol; i < maxCol; i++){
      FPGA->cutCostVector.at(i) += 1.0;
    }
  }

  //if(wireCost != (double)(maxCol - minCol + maxRow - minRow)){
  //  cout << "net " << name << " has wire cost discrepancy!" << endl;
  //}
  //wireCost = (double)(maxCol - minCol + maxRow - minRow);

  if(sinkVector.size() > 49){
    crossingFactor = 2.7933 + 0.02616 * (sinkVector.size() - 49);
  }
  else{
    crossingFactor = cross_count[sinkVector.size()];
  }
  
  //This is to implement the same cost function as in VPR - place.c:2623
  //Unfortunately, I don't want to go through the trouble of calculating their
  // chanx_place_cost_fac function, so I'll just use 0.01 so that I can compare
  // with their 4lut_sanitized architecture.  It shouldn't make a difference
  //Since we assume that all architectures have uniform channels.
  if((maxCol - minCol) == 0 && (maxRow - minRow) == 0){
    wireCost = 0;
  }
  else{
    wireCost = ((double)(maxRow - minRow + 1)) * crossingFactor * 0.01;
    wireCost += ((double)(maxCol - minCol + 1)) * crossingFactor * 0.01;
  }
}

void PipeliningNet::reCalculatePlacementCost(bool timing_driven, double criticalPathDelay){
  int currRow;
  int currCol;
  int minRowTemp;
  int maxRowTemp;
  int minColTemp;
  int maxColTemp;
  int i;
  double criticality;
  double crossingFactor;

  // Figure out the min and max locations of all source and sink terminals
  currRow = source->getLogicBlockRow();
  currCol = source->getLogicBlockCol();
  
  maxRowTemp = currRow;
  minRowTemp = currRow;
  maxColTemp = currCol;
  minColTemp = currCol;
  
  // Cycle through the sinks to see if any change the bounding box
  i = 0;
  newWireCost = 0.0;
  newTimingCost = 0.0;

  // cout << "Recalculating placement cost of net " << name << endl;
  
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    currRow = (*iter)->getLogicBlockRow();
    currCol = (*iter)->getLogicBlockCol();
    if(currRow < minRowTemp){
      minRowTemp = currRow;
    }
    else if(currRow > maxRowTemp){
      maxRowTemp = currRow;
    }
    if(currCol < minColTemp){
      minColTemp = currCol;
    }
    else if (currCol > maxColTemp){
      maxColTemp = currCol;
    }
    
    if(timing_driven){
      //If we are doing timing-driven placement, the cost of this net should be calculated as:
      //   criticality of (source, sink) = 1 - (slack of (source, sink) / critical path delay)
      //   timing cost = delay of (source, sink) * criticality of (source, sink) ^ criticality exponent
      //cout << "\tSlack = " << slackVector.at(i) << " criticality = " << 1.0 - (slackVector.at(i) / criticalPathDelay) << endl;
      criticality = 1.0 - (slackVector.at(i) / criticalPathDelay);
      criticality = doubmax(criticality, 0.0);
      //cout << "\tLinkDelay = " << linkDelayVector.at(i) << " LinkTimingCost = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) <<endl;
      
      newTimingCost += linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent);
      //cout << "\tTimingCost for sink " << (*iter)->name << " = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) << endl;

      if(deltaDelayCost){
	newTimingCostPrime = (linkDelayVector.at(i) - oldLinkDelayVector.at(i)) * pow(criticality, tdPlacementCritExponent);
      }
    }
  }

  if(maxCutCost){
    //We are safe from updating this net twice (just in case the blocks we moved were part of the same net
    // because the minCol and maxCol will only be changed once.
    //The new minCol might be larger than the old minCol (we decreased the span) - we should decrement those locations
    for(i = minCol; i < minColTemp; i++){
      FPGA->cutCostVector.at(i) += -1.0;
    }
    //The new minCol might be smaller than the old minCol (we increased the span) - we should increment those locations
    for(i = minColTemp; i < minCol; i++){
      FPGA->cutCostVector.at(i) += 1.0;
    }
    //The new maxCol might be smaller than the old maxCol (we decreased the span) - we should decrement those locations
    for(i = maxColTemp; i < maxCol; i++){
      FPGA->cutCostVector.at(i) -= 1.0;
    }
    //The new maxCol might be larger than the old maxCol (we increased the span) - we should increment those locations
    for(i = maxCol; i < maxColTemp; i++){
      FPGA->cutCostVector.at(i) += 1.0;
    }
  }

  //Set the values in the main net variables
  minRow = minRowTemp;
  maxRow = maxRowTemp;  
  minCol = minColTemp;
  maxCol = maxColTemp;

  //newWireCost = (double)(maxCol - minCol + maxRow - minRow);
  if(sinkVector.size() > 49){
    crossingFactor = 2.7933 + 0.02616 * (sinkVector.size() - 49);
  }
  else{
    crossingFactor = cross_count[sinkVector.size()];
  }
  
  //This is to implement the same cost function as in VPR - place.c:2623
  //Unfortunately, I don't want to go through the trouble of calculating their
  // chanx_place_cost_fac function, so I'll just use 0.01 so that I can compare
  // with their 4lut_sanitized architecture.  It shouldn't make a difference
  //Since we assume that all architectures have uniform channels.
  if((maxCol - minCol) == 0 && (maxRow - minRow) == 0){
    newWireCost = 0;
  }
  else{
    newWireCost = ((double)(maxRow - minRow + 1)) * crossingFactor * 0.01;
    newWireCost += ((double)(maxCol - minCol + 1)) * crossingFactor * 0.01;
  }

  //cout << "\tNew Timing cost for pipelined net " << name << " = " << newTimingCost << endl;
  //cout << "\tNew Wire Cost for pipelined net " << name << " = " << newWireCost << endl;
}

void PipeliningNet::reCalculatePlacementCostForwards(bool timing_driven, double criticalPathDelay){
  int currRow;
  int currCol;
  int minRowTemp;
  int maxRowTemp;
  int minColTemp;
  int maxColTemp;
  int i;
  double criticality;
  double crossingFactor;

  // Figure out the min and max locations of all source and sink terminals
  currRow = source->getLogicBlockRow();
  currCol = source->getLogicBlockCol();
  
  maxRowTemp = currRow;
  minRowTemp = currRow;
  maxColTemp = currCol;
  minColTemp = currCol;
  
  // Cycle through the sinks to see if any change the bounding box
  i = 0;
  newWireCost = 0.0;
  newTimingCost = 0.0;

  // cout << "Recalculating placement cost of net " << name << endl;
  
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++, i++){
    currRow = (*iter)->getLogicBlockRow();
    currCol = (*iter)->getLogicBlockCol();

    if(currRow < minRowTemp){
      minRowTemp = currRow;
    }
    else if(currRow > maxRowTemp){
      maxRowTemp = currRow;
    }
    if(currCol < minColTemp){
      minColTemp = currCol;
    }
    else if (currCol > maxColTemp){
      maxColTemp = currCol;
    }

    if(timing_driven){
      //If we are doing timing-driven placement, the cost of this net should be calculated as:
      //   criticality of (source, sink) = 1 - (slack of (source, sink) / critical path delay)
      //   timing cost = delay of (source, sink) * criticality of (source, sink) ^ criticality exponent
      //cout << "\tSlack = " << slackVector.at(i) << " criticality = " << 1.0 - (slackVector.at(i) / criticalPathDelay) << endl;
      criticality = 1.0 - (slackVector.at(i) / criticalPathDelay);
      criticality = doubmax(criticality, 0.0);
      //cout << "\tLinkDelay = " << linkDelayVector.at(i) << " LinkTimingCost = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) <<endl;
      
      newTimingCost += linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent);
      //cout << "\tTimingCost for sink " << (*iter)->name << " = " << linkDelayVector.at(i) * pow(criticality, tdPlacementCritExponent) << endl;
    }

    if((*iter)->retimeNerfed){
      (*iter)->destPipeliningNet->reCalculatePlacementCostForwards(timing_driven, criticalPathDelay);
    }
  }

  if(maxCutCost){
    //We are safe from updating this net twice (just in case the blocks we moved were part of the same net
    // because the minCol and maxCol will only be changed once.
    //The new minCol might be larger than the old minCol (we decreased the span) - we should decrement those locations
    for(i = minCol; i < minColTemp; i++){
      FPGA->cutCostVector.at(i) += -1.0;
    }
    //The new minCol might be smaller than the old minCol (we increased the span) - we should increment those locations
    for(i = minColTemp; i < minCol; i++){
      FPGA->cutCostVector.at(i) += 1.0;
    }
    //The new maxCol might be smaller than the old maxCol (we decreased the span) - we should decrement those locations
    for(i = maxColTemp; i < maxCol; i++){
      FPGA->cutCostVector.at(i) -= 1.0;
    }
    //The new maxCol might be larger than the old maxCol (we increased the span) - we should increment those locations
    for(i = maxCol; i < maxColTemp; i++){
      FPGA->cutCostVector.at(i) += 1.0;
    }
  }

  //Set the values in the main net variables
  minRow = minRowTemp;
  maxRow = maxRowTemp;  
  minCol = minColTemp;
  maxCol = maxColTemp;

  //newWireCost = (double)(maxCol - minCol + maxRow - minRow);

  if(sinkVector.size() > 49){
    crossingFactor = 2.7933 + 0.02616 * (sinkVector.size() - 49);
  }
  else{
    crossingFactor = cross_count[sinkVector.size()];
  }
  
  //This is to implement the same cost function as in VPR - place.c:2623
  //Unfortunately, I don't want to go through the trouble of calculating their
  // chanx_place_cost_fac function, so I'll just use 0.01 so that I can compare
  // with their 4lut_sanitized architecture.  It shouldn't make a difference
  //Since we assume that all architectures have uniform channels.
  if((maxCol - minCol) == 0 && (maxRow - minRow) == 0){
    newWireCost = 0;
  }
  else{
    newWireCost = ((double)(maxRow - minRow + 1)) * crossingFactor * 0.01;
    newWireCost += ((double)(maxCol - minCol + 1)) * crossingFactor * 0.01;
  }

  //cout << "\tNew Timing cost for pipelined net " << name << " = " << newTimingCost << endl;
  //cout << "\tNew Wire Cost for pipelined net " << name << " = " << newWireCost << endl;
}


double PipeliningNet::getDeltaWireCost(){
  return (newWireCost - wireCost);
}

void PipeliningNet::overWriteOldWireCost(){
  wireCost = newWireCost;
}

double PipeliningNet::getDeltaTimingCost(){
  //cout << "Net " << name << " newTiming cost = " << newTimingCost << " old timing Cost = " << timingCost <<endl;
  return (newTimingCost - timingCost);
}

double PipeliningNet::getDeltaTimingCostPrime(){
  return (newTimingCostPrime);
}

void PipeliningNet::overWriteOldTimingCost(){
  //cout << "Net " << name << " overwriting old timing cost " << timingCost << " with new timing cost " << newTimingCost << endl;
  timingCost = newTimingCost;
}

void PipeliningNet::clearTimingCostPrime(){
  newTimingCostPrime = 0.0;
}

void PipeliningNet::zeroSlackVectors(){
  for(vector <double>::iterator iter = slackVector.begin(); iter != slackVector.end(); iter++){
    (*iter) = 0.0;
  }
}

//Go through all of the sinks of this pipelining net at sort the sinks
void PipeliningNet::sortLockedRegs(bool sortByTiming){
  if(sortByTiming){
    sort_sinkVectorTiming();
  }
  else{
    sort_sinkVectorDistance();
  }
}

void PipeliningNet::sort_sinkVectorDistance(){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");

  int i=0;int j=0;
  int irow,icol,jrow,jcol,source_row,source_col;
  source_row=source->rNode->row_low;source_col=source->rNode->col_low;
  Terminal* temp_term;
  double tempDouble;
  int tempInt;

  for(i=0;i<sinkVector.size();i++){
    irow=sinkVector.at(i)->rNode->row_low;
    icol=sinkVector.at(i)->rNode->col_low;

    for(j=i+1;j<sinkVector.size();j++){
      jrow=sinkVector.at(j)->rNode->row_low;
      jcol=sinkVector.at(j)->rNode->col_low;
      if((abs(source_row-jrow) + abs(source_col-jcol)) < (abs(source_row-irow) + abs(source_col-icol))){
	//swap positions in the sink vector
	temp_term=sinkVector.at(i);
	sinkVector.at(i)=sinkVector.at(j);
	sinkVector.at(j)=temp_term;

	//swap values in linkDelayVector
	tempDouble = linkDelayVector.at(i);
	linkDelayVector.at(i) = linkDelayVector.at(j);
	linkDelayVector.at(j) = tempDouble;

	//swap values in slackVector
	tempDouble = slackVector.at(i);
	slackVector.at(i) = slackVector.at(j);
	slackVector.at(j) = tempDouble;

	//Update i row/col info
	irow = jrow;
	icol = jcol;
      }
    }
  }
  //for(i = 0; i < sinkVector.size(); i++){
  //  cout << "sink = " << sinkVector.at(i)->name << " slack = " << slackVector.at(i) << endl;
  //}
}

//Sort sinks of the net in order of increasing slack.
// Commented-out code is to implement tiebreaker - if they are the same, sort by increasing distance.
//  Should this be decreasing distance?  The furthest away sink might have the hardest to meet timing.
//  Maybe we don't really care anyways about distance.
void PipeliningNet::sort_sinkVectorTiming(){
  Terminal *tempSink;
  int low, high;
  double tempDouble;
  int tempInt;
  //int sourceRow, sourceCol;
  //int lowSinkRow, lowSinkCol, highSinkRow, highSinkCol;
  
  //cout << "Net " << name << endl;
  //sourceRow = source->rNode->row_low;
  //sourceCol=source->rNode->col_low;

  for(low = 0; low < sinkVector.size(); low++){
    //lowSinkRow = sinkVector.at(low)->rNode->row_low;
    //lowSinkCol = sinkVector.at(low)->rNode->col_low;

    for(high = low + 1; high < sinkVector.size(); high++){
      //cout << "   Considering net " << name << " sink " << low << " slack = " << slackVector.at(low);
      //cout << " and sink " << high << " slack = " << slackVector.at(high)<< endl;
      //highSinkRow = sinkVector.at(low)->rNode->row_low;
      //highSinkCol = sinkVector.at(low)->rNode->col_low;

      if(slackVector.at(low) > slackVector.at(high)){
	//|| (((slackVector.at(low) == slackVector.at(high)) && 
	//  ((abs(sourceRow - lowSinkRow) + abs(sourceCol - lowSinkCol)) > (abs(sourceRow - highSinkRow) + abs(sourceCol - highSinkCol)))){
	//We have decided to swap these sinks so we have to swap all of their information at the same time
	//1) Change positions in sinkVector
	//cout << "     Swapping net " << name << " sink " << low << " slack = " << slackVector.at(low);
	//cout << " and sink " << high << " slack = " << slackVector.at(high)<< endl;
	tempSink = sinkVector.at(high);
	sinkVector.at(high) = sinkVector.at(low);
	sinkVector.at(low) = tempSink;
	//2) Swap values in linkDelayVector
	tempDouble = linkDelayVector.at(high);
	linkDelayVector.at(high) = linkDelayVector.at(low);
	linkDelayVector.at(low) = tempDouble;
	//3) Swap values slackVector
	tempDouble = slackVector.at(high);
	slackVector.at(high) = slackVector.at(low);
	slackVector.at(low) = tempDouble;
	//Update low sink row and col
	//lowSinkRow = highSinkRow;
	//lowSinkCol = highSinkCol;
      }
    }
  }
  //for(low = 0; low < sinkVector.size(); low++){
  //  cout << "sink = " << sinkVector.at(low)->name << " slack = " << slackVector.at(low) << endl;
  //}
}

void PipeliningNet::rip_rTree(){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"RIPPING ROUTE TREE OF "<<name;
#endif

  rTree.postorderTraverse(rTree.get_root(),93);
  rTree.set_root(0);
}

// Get a pointer to this net's route tree
routeTree* PipeliningNet::get_rTree(){
  if(magicNum!=32983)
    ERROR("trying to access a deleted node...");

  return &rTree;
}

//Add a sink/link delay/slack amount triple to this pnet
void PipeliningNet::pushSinkPlusData(Terminal *sink, double linkDelay, double slack){
  sinkVector.push_back(sink);
  linkDelayVector.push_back(linkDelay);
  slackVector.push_back(slack);
}

//Remove a sink/link delay/slack amount triple from this pnet
void PipeliningNet::removeSinkPlusData(Terminal *sink){
  vector<Terminal*>::iterator tvIter;
  int select;

  for(tvIter = sinkVector.begin(), select = 0; 
      tvIter != sinkVector.end(); tvIter++, select++){
    if(*tvIter == sink){
      break;
    }
  }

  if(tvIter == sinkVector.end()){
    ERROR("Sink not found!");
  }
  else{
    //Remove the sink
    sinkVector.erase(tvIter);
    //Remove the associated link delay
    linkDelayVector.erase(linkDelayVector.begin() + select);
    //Remove the associated slack
    slackVector.erase(slackVector.begin() + select);
  }
}

//Remove a sink/link delay/slack amount triple from this pnet
void PipeliningNet::removeSinkPlusData(int select){
  //Make sure that we have that many sinks
  if(sinkVector.size() < select){
    ERROR("Attempting to remove a sink that does not exist");
  }

  //Remove the ith sink
  sinkVector.erase(sinkVector.begin() + select);
  //Remove the associated link delay
  linkDelayVector.erase(linkDelayVector.begin() + select);
  //Remove the associated slack
  slackVector.erase(slackVector.begin() + select);
}

//We are trying to find sinks that are at the latency numMove.
//If this Pnet is not at this latency, recursively call this on the
// destination pnets of the sinks of this net.
//If this Pnet is at this latency, detach the sinks from this net, attach them
// to the destination pnet provided, then proceed to update the latency of
// the destination pnets of the sinks of this net going forward.
void PipeliningNet::detachAndReattachHereForward(PipeliningNet *recvPipeliningNet, Netlist *netlist, int numMove){
  //Remove this net from the parent net's list of pipelining nets
  parentNet->removePipelineNetFromPNList(this);
  
  if(latency < numMove){
    for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
      //Indicate that this register has been removed
      (*iter)->retimeNerfed = true;
      if((*iter)->destPipeliningNet){
	//Remove this interveining register from the register bank that it is currently mapped to.
	//However, since we want to be able to undo this move, notify the register bank of the change
	// but do not modify the information in the register terminal itself.
	(*iter)->getRegisterBank()->removeStartReference(*iter);
	//cout << "Removing register " << (*iter)->name <<endl;

	//Remove this interviening register from the parent conventional net and
	// the host netlist list of registers.
	//Notice we are not actually deleting the register itself yet.
	parentNet->removePipelineRegFromPRList(*iter);
	netlist->removePipelineRegisterFromPlacementPRVector(*iter);

	//This is a net that we want to get rid of, so recursively call this function on all of the downstream nets.
	(*iter)->destPipeliningNet->detachAndReattachHereForward(recvPipeliningNet, netlist, numMove);
      }
      else{
	ERROR("This sink has no downstream net!");
      }
    }
  }
  else if(latency == numMove){
    //This is a net that we are interested in.  We would like to take all of the sinks of this net
    // and detach them from this net and attach them to the host net provided.
    for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
      //The slack that we have on this link is the sink's required time (should be the CPD since
      // it's a register), minus the source's departure time minus the link delay.
      //cout << "Attaching sink " << (*iter)->name << " to net " << recvPipeliningNet->name << endl;

      recvPipeliningNet->pushSinkPlusData(*iter, 0.0, (*iter)->requiredTime - recvPipeliningNet->source->departureTime);
      (*iter)->sourcePipeliningNet = recvPipeliningNet;

      //Now that we know this sink is being used, so if it is a register, update the routing node assignment for
      // all of the remaining registers in this bank.
      if((*iter)->getRegisterBank()){
	(*iter)->getRegisterBank()->updateRoutingNodes();
      }

      //Finally, update the latency of the downstream pipelining nets
      if((*iter)->destPipeliningNet){
	(*iter)->destPipeliningNet->reduceLatency(numMove);
      }
    }
    
    //Update the latency of this net - notice, we may have technically made this a non-pipelined
    // net, but that's not a huge deal.  We might push registers bank onto this net anyways later on.
    latency = 0;
  }
  else if(latency > numMove){
    ERROR("We are beyond the desired latency!");
  }
}

//Reduce the latency of this net and all downstream nets.
void PipeliningNet::reduceLatency(int amount){
  latency -= amount;
  
  //Recursively call this on all downstream pipelined nets.
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->destPipeliningNet){
      (*iter)->destPipeliningNet->reduceLatency(amount);
    }
  }
}

//Update the latency of this net and all downstream nets
void PipeliningNet::updateAllLatencies(int lat){
  latency = lat;
  //Recursively call this on all downstream pipelined nets.
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->destPipeliningNet){
      (*iter)->destPipeliningNet->updateAllLatencies(lat + 1);
    }
  }
}

//We would like to take each of the sinks of this net and push in numRegs
// number of registers into each of their register banks.
void PipeliningNet::insertRegistersExistingSinkBanks(int numRegs, Netlist *netlist, double criticalPathDelay){
  PipeliningNet *nextPNet;
  //char tempStr [MAX_STR_LEN];
  string tempStr;
  Terminal *regTerm;
  Terminal *currSink;
  registerBank *regBank;
  list <Terminal *> sinkList;
  
  if(numRegs > 0){ 
    for(vector<Terminal*>::iterator iter = sinkVector.begin(); 
	iter != sinkVector.end(); iter++){
      //Push all the sinks from this net into another list
      sinkList.push_back(*iter);
    }
    //Clear all of the sinks from this net
    sinkVector.clear();
    linkDelayVector.clear();
    slackVector.clear();

    for(list<Terminal*>::iterator iter = sinkList.begin();
	iter != sinkList.end(); iter++){
      regBank = (*iter)->rb;
      currSink = (*iter);
      
      for(int i = 0; i < numRegs; i++){
	//Build these pipelining nets and register from the back forward.
	//1) Create a new pipelining net
	//strcpy(tempStr, parentNet->name);
	//sprintf(tempStr, "%sPipelineNet%d", tempStr, parentNet->registerID + 1);
	tempStr = parentNet->name + "PipelineNet" + itoa(parentNet->registerID + 1);
	//cout << "Adding pipelined net" << tempStr << endl;
	nextPNet = new PipeliningNet(tempStr, 0, parentNet);
	parentNet->pipelineNetList.push_back(nextPNet);
	
	//2) Make a new register, mark it as nerfed and make it the source of this new net
	//strcpy(tempStr, parentNet->name);
	//sprintf(tempStr, "%sReg%d", tempStr, parentNet->registerID);
	tempStr = parentNet->name + "Reg" + itoa(parentNet->registerID);
	//cout << "Adding new register " << tempStr << " to bank " << regBank->getHostDelayBlock()->col
	//     << " , " << regBank->getHostDelayBlock()->row << " bank " << regBank->getHostBankNum() << endl;
	parentNet->registerID ++;
	regTerm = new Terminal(parentNet, this, nextPNet, tempStr);
	regTerm->retimeNerfed = true;
	parentNet->pipelineRegisterList.push_back(regTerm);
	netlist->placementPipelineRegisterVector.push_back(regTerm);
	regTerm->departureTime = 0.0;
	regTerm->requiredTime = criticalPathDelay;
	nextPNet->source = regTerm;
	
	//3) Add the prior sinks to the sink list of this new net
	//The slack should be the sink required time minus the source departure time
	//  minus the delay of the wire between the two.
	nextPNet->pushSinkPlusData(currSink, 0.0, currSink->requiredTime - nextPNet->source->departureTime);
	currSink->sourcePipeliningNet = nextPNet;
	
	//4) Place the newly created register into the existing half-occupied register bank
	regBank->placeTerminalFront(regTerm);
	
	currSink = regTerm;
      }

      //Attach the last created register to the sink list of this net
      this->pushSinkPlusData(currSink, 0.0, currSink->requiredTime - this->source->departureTime);
      
      //Fix the routing node assignment of all the registers in this bank
      regBank->updateRoutingNodes();
    }
  }
}

//We would like to grab new register banks and insert this many registers between the source
// of this net and it's sinks.
bool PipeliningNet::insertRegistersNewBanks(int numRegs, Netlist *netlist, double criticalPathDelay){
  PipeliningNet *nextPNet;
  //char tempStr [MAX_STR_LEN];
  string tempStr;
  Terminal *regTerm;
  Terminal *currSink;
  registerBank *regBank;
  
  if(numRegs > 0){ 
    regBank = FPGA->getRandomFreeRegBank();
    if(!regBank){
      //We just ran out of empty register banks in the system.
      //We haven't actually done anything here yet (could have inserted registers into existing banks)
      // so we don't really have to worry about finishing stuff up yet.
      return false;
    }

    //Build these pipelining nets and register from the back forward.
    //1) Create a new pipelining net
    //strcpy(tempStr, parentNet->name);
    //sprintf(tempStr, "%sPipelineNet%d", tempStr, parentNet->registerID + 1);
    tempStr = parentNet->name + "PipelineNet" + itoa(parentNet->registerID + 1);
    //cout << "Adding pipelined net" << tempStr << endl;
    nextPNet = new PipeliningNet(tempStr, 0, parentNet);
    parentNet->pipelineNetList.push_back(nextPNet);
    
    //2) Make a new register, mark it as nerfed and make it the source of this new net and place it
    //strcpy(tempStr, parentNet->name);
    //sprintf(tempStr, "%sReg%d", tempStr, parentNet->registerID);
    tempStr = parentNet->name + "Reg" + itoa(parentNet->registerID);
    //cout << "Adding new register " << tempStr << " to bank " << regBank->getHostDelayBlock()->col
    // << " , " << regBank->getHostDelayBlock()->row << " bank " << regBank->getHostBankNum() << endl;
    parentNet->registerID ++;
    regTerm = new Terminal(parentNet, this, nextPNet, tempStr);
    regTerm->retimeNerfed = true;
    parentNet->pipelineRegisterList.push_back(regTerm);
    netlist->placementPipelineRegisterVector.push_back(regTerm);
    regTerm->departureTime = 0.0;
    regTerm->requiredTime = criticalPathDelay;
    nextPNet->source = regTerm;

#ifdef PLACER_DEBUG
    numRegistersEstimatedNeeded --;
#endif


    //4) Place the newly created register into the random register bank we found
    regBank->placeTerminal(regTerm);
    if(regBank->bankFull()){
      regBank = NULL;
    }

    //Detach all of the sinks from this net and add them to this new net
    for(vector<Terminal*>::iterator iter = sinkVector.begin(); 
	iter != sinkVector.end(); iter++){
      nextPNet->pushSinkPlusData(*iter, 0.0, (*iter)->requiredTime - nextPNet->source->departureTime);
      (*iter)->sourcePipeliningNet = nextPNet;
    }
    sinkVector.clear();
    linkDelayVector.clear();
    slackVector.clear();

    currSink = regTerm;

    for(int i = 1; i < numRegs; i++){
      if(!regBank){
	regBank = FPGA->getRandomFreeRegBank();
	if(!regBank){
	  //We just ran out of empty register banks in the system.
	  //We have created an unknown number of new nerfed registers and new pipelining nets
	  // in a string.  However, we have not connected the last created register to this net
	  // yet. We have to do that so we don't lose it.

	  //Attach the last created register to the sink list of this net
	  this->pushSinkPlusData(currSink, 0.0, 0.0);
    
	  return false;
	}
      }

      //1) Create a new pipelining net
      //strcpy(tempStr, parentNet->name);
      //sprintf(tempStr, "%sPipelineNet%d", tempStr, parentNet->registerID + 1);
      tempStr = parentNet->name + "PipelineNet" + itoa(parentNet->registerID + 1);
      //cout << "Adding pipelined net" << tempStr << endl;
      nextPNet = new PipeliningNet(tempStr, 0, parentNet);
      parentNet->pipelineNetList.push_back(nextPNet);
      
      //2) Make a new register, mark it as nerfed and make it the source of this new net
      //strcpy(tempStr, parentNet->name);
      //sprintf(tempStr, "%sReg%d", tempStr, parentNet->registerID);
      tempStr = parentNet->name + "Reg" + itoa(parentNet->registerID);
      //cout << "Adding new register " << tempStr << " to bank " << regBank->getHostDelayBlock()->col
      //   << " , " << regBank->getHostDelayBlock()->row << " bank " << regBank->getHostBankNum() << endl;
      parentNet->registerID ++;
      regTerm = new Terminal(parentNet, this, nextPNet, tempStr);
      regTerm->retimeNerfed = true;
      parentNet->pipelineRegisterList.push_back(regTerm);
      netlist->placementPipelineRegisterVector.push_back(regTerm);
      regTerm->departureTime = 0.0;
      regTerm->requiredTime = criticalPathDelay;
      nextPNet->source = regTerm;

#ifdef PLACER_DEBUG
      numRegistersEstimatedNeeded --;
#endif

	
      //3) Add the prior sink to the sink list of this new net
      //The slack should be the sink required time minus the source departure time
      //  minus the delay of the wire between the two.
      nextPNet->pushSinkPlusData(currSink, 0.0, currSink->requiredTime - nextPNet->source->departureTime);
      currSink->sourcePipeliningNet = nextPNet;
      
      //4) Place the newly created register into the existing half-occupied register bank
      regBank->placeTerminalFront(regTerm);
      
      currSink = regTerm;
      
      if(regBank->bankFull()){
	regBank->updateRoutingNodes();
	regBank = NULL;
      }
    }

    //Attach the last created register to the sink list of this net
    this->pushSinkPlusData(currSink, 0.0, currSink->requiredTime - this->source->departureTime);
    
    //Fix the routing node assignment of the last register bank if it wasn't full
    if(regBank){
      regBank->updateRoutingNodes();
    }
  }
  return true;
}

//Add this net back into the main net's list the try to add back in sinks
void PipeliningNet::addBackNetAndRegistersForwards(Netlist *netlist, int numMove){
  //Add this net back into the main list of netlists
  parentNet->pipelineNetList.push_back(this);

  //We need to reset all of this net's values so we get a correct delta later
  resetValues();

  //cout << "Adding net " << name << " back" << endl;

  //Now put the old registers back into the proper netlist lists and
  // back into the proper register banks.
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    //If this this is nerfed, that means that we have to add it back into circulation
    if((*iter)->retimeNerfed){
      //Add this register back into the main lists of registers
      //cout << "Adding register " << (*iter)->name << " back" << endl;
      parentNet->pipelineRegisterList.push_back(*iter);
      netlist->placementPipelineRegisterVector.push_back(*iter);

      //Insert this register back into the register bank it was in.
      //We might be inserting registers into the middle of the register bank,
      // so we have to say what we are putting them after.
      //Also, notice that unlike other "placeTerminal" functions, this function updates 
      // the routing node assignments as we put registers in.  This is because we
      // don't know when we are done so it must be left correctly all the time.
      if((*iter)->rb){
	//cout << "Putting register " << (*iter)->name << " back into register bank "
	//     << (*iter)->rb->getHostDelayBlock()->col << " , " << (*iter)->rb->getHostDelayBlock()->row 
	//     << " bank " << (*iter)->rb->getHostBankNum() << endl;
	(*iter)->rb->insertRegisterForwards(*iter, (*iter)->sourcePipeliningNet->source);
      }
      
      //Recursively call this function on the net that this register drives
      (*iter)->destPipeliningNet->addBackNetAndRegistersForwards(netlist, numMove);
    }
    else{
      //If this sink wasn't nerfed, then that means we have reached the end of the 
      // removed nets.
      //Reattach this sink to this net
      //cout << "Adding sink " << (*iter)->name << " back to net " << name << endl;
      (*iter)->sourcePipeliningNet = this;

      //Update the downstream latencies.
      latency = numMove;
      if((*iter)->destPipeliningNet){
	(*iter)->destPipeliningNet->updateAllLatencies(latency + 1);
      }
    }
  }
}

//Add this net back into the main net's list the try to add back in source
void PipeliningNet::addBackNetAndRegistersBackwards(Terminal *sink, Netlist *netlist){
  //Now put the old registers back into the proper netlist lists and
  // back into the proper register banks.
  //If the source of this net this is nerfed, that means that we might have to add it back into circulation
  if(source->retimeNerfed){
    //Just because this register has been nerfed doesn't mean that the sink
    // we provided is actually connected.
    //We might have one net that drives two different input pins on the same logic block.
    //In that case, this sink might have gotten disconnected.
    //We should reconnect it, but we shouldn't do all of the other things.
    if(!hasSink(sink)){
      sink->sourcePipeliningNet = source->destPipeliningNet;
      source->destPipeliningNet->pushSinkPlusData(sink, 0.0, sink->requiredTime - source->departureTime);
    }
    else{
      //We need to reset all of this net's values so we get a correct delta later
      resetValues();
      
      //Add this net back into the main list of netlists
      parentNet->pipelineNetList.push_back(this);
      //cout << "Adding net " << name << " back" << endl;
      
      //Add this register back into the main lists of registers
      //cout << "Adding register " << source->name << " back" << endl;
      parentNet->pipelineRegisterList.push_back(source);
      netlist->placementPipelineRegisterVector.push_back(source);
      
      //Insert this register back into the register bank it was in.
      //We might be inserting registers into the middle of the register bank,
      // so we have to say what we are putting them after.
      //Also, notice that unlike other "placeTerminal" functions, this function updates 
      // the routing node assignments as we put registers in.  This is because we
      // don't know when we are done so it must be left correctly all the time.
      if((source)->rb){
	//cout << "Putting register " << source->name << " back into register bank "
	//   << (source)->rb->getHostDelayBlock()->col << " , " << (source)->rb->getHostDelayBlock()->row 
	//   << " bank " << (source)->rb->getHostBankNum() << endl;
	//This relies on the fact that this net must have exactly one sink
	(source)->rb->insertRegisterBackwards(source, sinkVector.front());
	
	//Recursively call this function on the net that drives this register
      source->sourcePipeliningNet->addBackNetAndRegistersBackwards(source, netlist);
      }
    }
  }
  else{
    //If this source has not been nerfed, we have definately reached the point where
    // we want to attach this terminal to this net
    //The slack should be the sink required time minus the source departure time
    sink->sourcePipeliningNet = source->destPipeliningNet;
    source->destPipeliningNet->pushSinkPlusData(sink, 0.0, sink->requiredTime - source->departureTime);
  }
}

//Decrement the cut cost vector associated with the span of this pipelining net and downstream pnets
void PipeliningNet::decrementNerfedSpanForwards(){
  for(int i = minCol; i < maxCol; i++){
    FPGA->cutCostVector.at(i) += -1.0;
  }

  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->retimeNerfed){
      (*iter)->destPipeliningNet->decrementNerfedSpanForwards();
    }
  }
}

//Decrement the cut cost vector associated with the span of this pipelining net and upstream pnets
void PipeliningNet::decrementNerfedSpanBackwards(Terminal *sink){
  if(source->retimeNerfed){
    //If the source of this net is nerfed, there should only be one thing
    // attached to this net.
    if(sinkVector.size() != 1){
      ERROR("Sink vector != 1");
    }
    
    //If the sink we have provided is the one that is currently connected,
    // then we can try to kill stuff upstream, along with deleting this net and the source
    if(sinkVector.front() == sink){
      source->sourcePipeliningNet->decrementNerfedSpanBackwards(source);

      for(int i = minCol; i < maxCol; i++){
	FPGA->cutCostVector.at(i) += -1.0;
      }
    }
  }
}

//Return the wire cost of this net and nerfed downstream nets
double PipeliningNet::returnNerfedWireCostForwards(){
  double wc = wireCost;
  //cout << "Net " << name << " wire cost = " << wireCost <<endl;

  //Go through all of the sinks of this pipelining net and if they are nerfed,
  // recursively call this function on the downstream net
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->retimeNerfed){
      wc += (*iter)->destPipeliningNet->returnNerfedWireCostForwards();
    }
  }
  return wc;
}

//Return the wire cost of this net and nerfed upstream nets.
//Notice that we since we might have orphaned a net midway due to fan-out,
// we also evaluate the impact of the last, un-nerfed net.
double PipeliningNet::returnNerfedWireCostBackwards(bool timing_driven, double criticalPathDelay, Terminal *sink){
  double wc = 0.0; 
  //See if the source of this pipelined net is nerfed.  If it is, then recursively
  // call this function on the upstream net
  if(source->retimeNerfed){
    //If the source of this net is nerfed, there should only be one thing
    // attached to this net.
    if(sinkVector.size() != 1){
      ERROR("Sink vector != 1");
    }
    
    //If the sink we have provided is the one that is currently connected,
    // then we can try to kill stuff upstream, along with deleting this net and the source
    if(sinkVector.front() == sink){
      wc = wireCost;
      //cout << "Net " << name << " wire cost = " << wireCost <<endl;
      wc += source->sourcePipeliningNet->returnNerfedWireCostBackwards(timing_driven, criticalPathDelay, source);
    }
  }
  else{
    //If the source of this pipelined net is not nerfed, we have either reached the
    // split point for a register that had to be maintained due to fan-out
    // or we have reached the point where we re-attached the logic block input.
    //Either way, we can safely re-assess the delta wirecost of this net.
    reCalculatePlacementCost(timing_driven, criticalPathDelay);
    //cout << "Delta wire cost for net " << name << " = " << getDeltaWireCost() << endl;
    wc = -1 * getDeltaWireCost();
    overWriteOldWireCost();
  }

  return wc;
}

//Return the timing cost of this net and nerfed downstream nets
double PipeliningNet::returnNerfedTimingCostForwards(){
  double tc = timingCost;

  //cout << "Net " << name << " timing cost = " << timingCost <<endl;
  //Go through all of the sinks of this pipelinine net and if they are nerfed,
  // recursively call this function on the downstream net
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->retimeNerfed){
      tc += (*iter)->destPipeliningNet->returnNerfedTimingCostForwards();
    }
  }
  return tc;
}

//Return the timing cost of this net and nerfed upstream nets
//This function does not actually perform the "recalculatePlacement" for this net
// so must be called after the returnNerfedWireCostBackwards function
double PipeliningNet::returnNerfedTimingCostBackwards(Terminal *sink){
  double tc = 0.0; 
  //See if the source of this pipelined net is nerfed.  If it is, then recursively
  // call this function on the upstream net
  if(source->retimeNerfed){
    if(sinkVector.size() != 1){
      ERROR("Sink vector != 1");
    }
    
    //If the sink we have provided is the one that is currently connected,
    // then we can try to kill stuff upstream, along with deleting this net and the source
    if(sinkVector.front() == sink){
      tc = timingCost;
      //cout << "Net " << name << " timing cost = " << timingCost <<endl;
      tc += source->sourcePipeliningNet->returnNerfedTimingCostBackwards(source);
    }
  }
  else{
    //If the source of this pipelined net is not nerfed, we have either reached the
    // split point for a register that had to be maintained due to fan-out
    // or we have reached the point where we re-attached the logic block input.
    //Either way, we can safely re-assess the delta timing cost of this net.
    //cout << "Delta timing cost for net " << name << " = " << getDeltaTimingCost() << endl;
    tc = -1 * getDeltaTimingCost();
    overWriteOldTimingCost();
  }

  return tc;
}

//Delete nerfed sink registers and nerfed downstream nets
void PipeliningNet::deleteNerfedNetsAndRegistersForwards(){

  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->retimeNerfed){
      (*iter)->destPipeliningNet->deleteNerfedNetsAndRegistersForwards();
      //cout << "Freeing net " << (*iter)->destPipeliningNet->name << endl;
      delete (*iter)->destPipeliningNet;
      //cout << "Freeing register " << (*iter)->name << endl;
      delete (*iter);
    }
  }
}

//Delete nerfed sink registers and nerfed upstream nets
void PipeliningNet::deleteNerfedNetsAndRegistersBackwards(Terminal *sink){
  if(source->retimeNerfed){
    //If the source of this net is nerfed, there should only be one thing
    // attached to this net.
    if(sinkVector.size() != 1){
      ERROR("Sink vector != 1");
    }
    
    //If the sink we have provided is the one that is currently connected,
    // then we can try to kill stuff upstream, along with deleting this net and the source
    if(sinkVector.front() == sink){
      source->sourcePipeliningNet->deleteNerfedNetsAndRegistersBackwards(source);
      //cout << "Freeing register " << source->name << endl;
      delete source;
      //cout << "Freeing net " << this->name << endl;
      delete this;
    }
    //If not, let the other guy worry about it.
  }
}

//Reset all of the values for this net.
void PipeliningNet::resetValues(){
  minRow = -1;
  maxRow = -1;
  minCol = -1;
  maxCol = -1;
  wireCost = 0.0;
  timingCost = 0.0;
}

//Go through the sources and sinks of this pipelining net
// to make sure that the nets they connect to have the same parent net
void PipeliningNet::checkSourcesAndSinks(){
  if(source->sourcePipeliningNet){
    if(source->sourcePipeliningNet->parentNet != parentNet){
      ERROR("All things attached to pipelining net should share a common parent net!");
    }
  }
  for(vector<Terminal*>::iterator iter = sinkVector.begin(); iter != sinkVector.end(); iter++){
    if((*iter)->destPipeliningNet){
      if((*iter)->destPipeliningNet->parentNet != parentNet){
	ERROR("All things attached to pipelining net should share a common parent net!");
      }
    }
  }
}
 
//Does this pipelining net actually have this sink attached to it?
bool PipeliningNet::hasSink(Terminal *sink){
  vector<Terminal*>::iterator tvIter;

  for(tvIter = sinkVector.begin(); tvIter != sinkVector.end(); tvIter++){
    if(*tvIter == sink){
      break;
    }
  }
  
  if(tvIter == sinkVector.end()){
    return false;
  }
  else{
    return true;
  }
}

//See if the node is within the routing bounding box
bool PipeliningNet::checkNodeRoutingBoundingBox(routingNode *node){
  //Notice that node numbers are taken by the wires to the right and down of an array location
  if(node->row_high < minRow - routeBoundingBox - 1 || 
     node->col_high < minCol - routeBoundingBox - 1 || 
     node->row_low > maxRow + routeBoundingBox ||
     node->col_low > maxCol + routeBoundingBox){
    return false;
  }
  return true;
}

bool PipeliningNet::foundNewSink(routingNode *node){
  for(vector<Terminal*>::iterator sinkVectorIter = sinkVector.begin(); sinkVectorIter != sinkVector.end(); sinkVectorIter++){
    if(node == (*sinkVectorIter)->rNode){
      return true;
    }
  }
  return false;    
}

void PipeliningNet::printDistanceToAllSinks(){
  int sourceRow = source->getLogicBlockRow();
  int sourceCol = source->getLogicBlockCol();

  for(vector<Terminal*>::iterator sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
    cout << "Distance to sink = " << abs(sourceCol - (*sinkIter)->getLogicBlockCol()) + abs(sourceRow - (*sinkIter)->getLogicBlockRow()) << endl;
  }

}

//Find the Manhattan distance to the closest sink
int PipeliningNet::minDistanceToAllSinks(){
  int minDistance = BIGINTEGER;

  int sourceRow = source->getLogicBlockRow();
  int sourceCol = source->getLogicBlockCol();

  for(vector<Terminal*>::iterator sinkIter = sinkVector.begin(); sinkIter != sinkVector.end(); sinkIter++){
    if(abs(sourceCol - (*sinkIter)->getLogicBlockCol()) + abs(sourceRow - (*sinkIter)->getLogicBlockRow()) < minDistance){
      minDistance = abs(sourceCol - (*sinkIter)->getLogicBlockCol()) + abs(sourceRow - (*sinkIter)->getLogicBlockRow());
    }
  }
  return minDistance;
}
