// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

extern vprFPGA* FPGA;

#ifdef PLACER_DEBUG
extern int numRegistersEstimatedNeeded;
#endif

// Give this terminal a name
void Terminal::set_name(string n){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  name = n;
}

// Set logic block associated with this terminal
void Terminal::set_lBlock(logicBlock* l){
  if(!l)
    ERROR("logic block cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  lBlock=l;
  
}

// Map this terminal to a routing node
void Terminal::set_rNode(routingNode* r){
  if(!r)
    ERROR("routing node cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  rNode=r;
}

// Net that this terminal drives or is driven by
void Terminal::set_net(Net* n){
  if(!n)
    ERROR("net cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  net=n;
}

// Set the type of this terminal
void Terminal::set_terminalType(enum e_terminalType t){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  terminalType=t;
}

// Whether this is a sink terminal of net
void Terminal::set_isSink(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  isSink=u;
}

// Whether this is the source terminal of net
void Terminal::set_isSource(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  isSource=u;
}

// Get name of this terminal
string Terminal::get_name(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Get logic block associated with this terminal
logicBlock* Terminal::get_lBlock(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return lBlock;
}

// Get net that this terminal drives or is driven by
Net* Terminal::get_net(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return net;
}

// Get net that this terminal drives or is driven by
routingNode* Terminal::get_rNode(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return rNode;
}

// Whether this is a sink terminal
bool Terminal::get_isSink(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return isSink;
}

// Whether this is a source terminal
bool Terminal::get_isSource(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return isSource;
}

// Print details about this terminal
void Terminal::print_Info(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"\tBEGIN PRINT INFO FOR TERMINAL "<<name<<" OF TYPE "<<terminalType;

  if(isSink){
    cout<<endl<<"\tSINK TERMINAL OF NET "<<net->get_name();
    cout<<endl<<"\tINPUT TERMINAL OF "<<lBlock->get_name();
  }
  if(isSource){
    cout<<endl<<"\tSOURCE TERMINAL OF NET "<<net->get_name();
    cout<<endl<<"\tOUTPUT TERMINAL OF "<<lBlock->get_name();
  }

}

void Terminal::clearRNode(){
  rNode = 0;
}

// This terminal belongs to a logic block/IO pin or pipeline register
//  figure out which it is.
enum e_genericBlockType Terminal::getHostBlockType(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  
  if(lBlock){
    return(lBlock->get_blockType());
  }
  else if(rb){
    return REGISTER_BANK;
  }
  else{
    ERROR("Terminal is not associated with any logic block or register bank!");
  }
}

enum e_timingMoveBlockType Terminal::getHostTimingMoveType(){
  if(getHostBlockType() == GBT_NO_TYPE){
    ERROR("Terminal's host is not associated with any type of block!");
  }
  else if(getHostBlockType() == INPUTPAD || getHostBlockType() == OUTPUTPAD){
    return IOPIN;
  }
  else if(getHostBlockType() == REGISTER_BANK){
    return REGISTERBANK;
  }
  else{
    return LB;
  }
}

// This terminal might have a bank number if it is a pipeline register
int Terminal::getHostBankNum(){
    if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  
  if(lBlock){
    return 0;
  }
  else{
    //If this terminal does not have a logic block associated with it, it must be
    // a pipelining register
    if(!rb){
      ERROR("This block does not have a register bank associated with it!");
    }
    else{
      return rb->getHostBankNum();
    }
  }
}

//This terminal could be the sink of a net
// If it is, update the timing of this net for this sink
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
void Terminal::updateSinkDelayAndSlack(){
  if(sourcePipeliningNet){
    sourcePipeliningNet->updateSinkDelayAndSlack(this);
  }
  else if(net){
    net->updateSinkDelayAndSlack(this);
  }
}

//This terminal is the sink of a pipeline net.
//Update the nerfed nets upstream of this block too.
void Terminal::updateSinkDelayAndSlackBackwards(){
  if(sourcePipeliningNet){
    //cout << "Updating delay and slack info for net " << sourcePipeliningNet->name << endl;
    sourcePipeliningNet->updateSinkDelayAndSlack(this);
    if(sourcePipeliningNet->source->retimeNerfed){
      sourcePipeliningNet->source->updateSinkDelayAndSlackBackwards();
    }
  }
  else{
    ERROR("This should not be used on a non-pipelined net!");
  }
}

//This terminal could be the source of a net
// If it is, update the timing of this net
//There is never an occasion in which on terminal will ever be the source of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
void Terminal::updateSourceDelaysAndSlacks(){
  if(destPipeliningNet){
    destPipeliningNet->updateSourceDelaysAndSlacks();
  }

  else if(net){
    net->updateSourceDelaysAndSlacks();
  }
}

//This terminal could be the source of a pipelined net.
//Update the nerfed nets downstream too.
void Terminal::updateSourceDelaysAndSlacksForwards(){
  if(destPipeliningNet){
    destPipeliningNet->updateSourceDelaysAndSlacksForwards();
  }
  else{
    ERROR("This should not be used on a non-pipelined net!");
  }
}


//Go through the nets associated with this terminal and update their wire and timing costs
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
void Terminal::updateSinkWireAndTimingCost(bool timing_driven, double criticalPathDelay){
  if(sourcePipeliningNet){
    sourcePipeliningNet->reCalculatePlacementCost(timing_driven, criticalPathDelay);
  }
  else if(net){
    net->reCalculatePlacementCost(timing_driven, criticalPathDelay);
  }
}

//Go through the nets associated with this terminal and update their wire and timing costs
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
//Also, there are brand new upstream nets so recursively call this on upstream nets
void Terminal::updateSinkWireAndTimingCostBackwards(bool timing_driven, double criticalPathDelay){
  if(sourcePipeliningNet){
    //cout << "Updating wire and timing info for net " << sourcePipeliningNet->name << endl;
    sourcePipeliningNet->reCalculatePlacementCost(timing_driven, criticalPathDelay);
    if(sourcePipeliningNet->source->retimeNerfed){
      sourcePipeliningNet->source->updateSinkWireAndTimingCostBackwards(timing_driven, criticalPathDelay);
    }
  }
  else{
    ERROR("This should only be called on pipelined nets!");
  }
}

//Go through the nets associated with this terminal and update their wire and timing costs
//There is never an occasion in which on terminal will ever be the source of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
void Terminal::updateSourceWireAndTimingCosts(bool timing_driven, double criticalPathDelay){
  if(destPipeliningNet){
    destPipeliningNet->reCalculatePlacementCost(timing_driven, criticalPathDelay);
  }
  else if(net){
    net->reCalculatePlacementCost(timing_driven, criticalPathDelay);
  }
}

//Go through the nets associated with this terminal and update their wire and timing costs
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net.
void Terminal::updateSourceWireAndTimingCostsForwards(bool timing_driven, double criticalPathDelay){
  if(destPipeliningNet){
    destPipeliningNet->reCalculatePlacementCostForwards(timing_driven, criticalPathDelay);
  }
  else{
    ERROR("This should only be called on pipelined nets!");
  }
}

//Go through the nets associated with this terminal and report their delta wire cost
// then copy the new value to the old value
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
double Terminal::sinkDeltaWireCost(){
  double wireCost = 0.0;

  if(sourcePipeliningNet){
    wireCost += sourcePipeliningNet->getDeltaWireCost();
    sourcePipeliningNet->overWriteOldWireCost();
  }  
  else if(net){
    wireCost += net->getDeltaWireCost();
    net->overWriteOldWireCost();
  }
  return wireCost;  
}

//Go through the nets associated with this terminal and report their delta wire cost
// then copy the new value to the old value.  Since there are newly created nets upstream
// of this terminal, recursively call this on nerfed registers.
//This function will only be used on pipelined nets
double Terminal::sinkDeltaWireCostBackwards(){
  double wireCost = 0.0;

  if(sourcePipeliningNet){
    wireCost += sourcePipeliningNet->getDeltaWireCost();
    //cout << "Net " << sourcePipeliningNet->name << " has a delta wire cost of " << sourcePipeliningNet->getDeltaWireCost() <<endl;
    sourcePipeliningNet->overWriteOldWireCost();
    if(sourcePipeliningNet->source->retimeNerfed){
      wireCost += sourcePipeliningNet->source->sinkDeltaWireCostBackwards();
    }
  }  
  else{
    ERROR("This should not be called on non-pipelined nets!");
  }
  return wireCost;  
}

//Go through the nets associated with this terminal and report their delta wire cost
// then copy the new value to the old value
//There is never an occasion in which on terminal will ever be the source of both a regular net and
// a pipelining net
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
double Terminal::sourceDeltaWireCost(){
  double wireCost = 0.0;

  if(destPipeliningNet){
    wireCost += destPipeliningNet->getDeltaWireCost();
    destPipeliningNet->overWriteOldWireCost();
  }
  else if(net){
    wireCost += net->getDeltaWireCost();
    net->overWriteOldWireCost();
  }
  return wireCost;
}

//Go through the nets associated with this terminal and report their delta wire cost
// then copy the new value to the old value.  Since there are newly created nets downstream
// of this terminal, recursively call this on nerfed registers.
//This function will only be used on pipelined nets
double Terminal::sourceDeltaWireCostForwards(){
  double wireCost = 0.0;

  if(destPipeliningNet){
    wireCost += destPipeliningNet->getDeltaWireCost();
    //cout << "Net " << destPipeliningNet->name << " has a delta wire cost of " << destPipeliningNet->getDeltaWireCost() <<endl;
    destPipeliningNet->overWriteOldWireCost();

    for(vector<Terminal*>::iterator iter = destPipeliningNet->sinkVector.begin(); 
	iter != destPipeliningNet->sinkVector.end(); iter++){
      if((*iter)->retimeNerfed){
	wireCost += (*iter)->sourceDeltaWireCostForwards();
      }
    }
  }

  else{
    ERROR("This should not be called on non-pipelined nets!");
  }
  return wireCost;
}

//Go through the nets associated with this terminal and report their delta timing cost
// then copy the new value to the old value.
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
double Terminal::sinkDeltaTimingCost(){
  double timingCost = 0.0;

  if(sourcePipeliningNet){
    timingCost += sourcePipeliningNet->getDeltaTimingCost();
    sourcePipeliningNet->overWriteOldTimingCost();
  }  
  else if(net){
    timingCost += net->getDeltaTimingCost();
    net->overWriteOldTimingCost();
  }
  return timingCost;
}

//Go through the nets associated with this terminal and report their delta timing cost prime
// then zero out the value.
//There is never an occasion in which on terminal will ever be the sink of both a regular net and
// a pipelining net.
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
double Terminal::sinkDeltaTimingCostPrime(){
  double timingCostPrime = 0.0;

  if(sourcePipeliningNet){
    timingCostPrime += sourcePipeliningNet->getDeltaTimingCostPrime();
    sourcePipeliningNet->clearTimingCostPrime();
  }  
  else if(net){
    timingCostPrime += net->getDeltaTimingCostPrime();
    net->clearTimingCostPrime();
  }
  return timingCostPrime;
}


//Go through the nets associated with this terminal and report their delta timing cost
// then copy the new value to the old value.  Since there are newly created nets upstream
// of this terminal, recursively call this on nerfed registers.
//This function will only be used on pipelined nets
double Terminal::sinkDeltaTimingCostBackwards(){
  double timingCost = 0.0;

  if(sourcePipeliningNet){
    //cout << "Timing cost for net " << sourcePipeliningNet->name << " = " << sourcePipeliningNet->getDeltaTimingCost() <<endl;
    timingCost += sourcePipeliningNet->getDeltaTimingCost();
    sourcePipeliningNet->overWriteOldTimingCost();
    if(sourcePipeliningNet->source->retimeNerfed){
      timingCost += sourcePipeliningNet->source->sinkDeltaTimingCostBackwards();
    }
  }  
  else{
    ERROR("This should not be used on a non-pipelined net!");
  }
  return timingCost;
}

//Go through the nets associated with this terminal and report their delta wire cost
// then copy the new value to the old value
//There is never an occasion in which on terminal will ever be the source of both a regular net and
// a pipelining net
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
double Terminal::sourceDeltaTimingCost(){
  double timingCost = 0.0;

  if(destPipeliningNet){
    timingCost += destPipeliningNet->getDeltaTimingCost();
    destPipeliningNet->overWriteOldTimingCost();
  }  
  else if(net){
    timingCost += net->getDeltaTimingCost();
    net->overWriteOldTimingCost();
  }  
  return timingCost;
}

//Go through the nets associated with this terminal and report their delta wire cost prime
// then clear the value
//There is never an occasion in which on terminal will ever be the source of both a regular net and
// a pipelining net
//However, all pipelined registers know about their parent regular net so we have to check for the
// pipelined net first.
double Terminal::sourceDeltaTimingCostPrime(){
  double timingCostPrime = 0.0;

  if(destPipeliningNet){
    timingCostPrime += destPipeliningNet->getDeltaTimingCostPrime();
    destPipeliningNet->clearTimingCostPrime();
  }  
  else if(net){
    timingCostPrime += net->getDeltaTimingCostPrime();
    net->clearTimingCostPrime();
  }  
  return timingCostPrime;
}

//Go through the nets associated with this terminal and report their delta timing cost
// then copy the new value to the old value.  Since there are newly created nets downstream
// of this terminal, recursively call this on nerfed registers.
//This function will only be used on pipelined nets
double Terminal::sourceDeltaTimingCostForwards(){
  double timingCost = 0.0;

  if(destPipeliningNet){
    timingCost += destPipeliningNet->getDeltaTimingCost();
    destPipeliningNet->overWriteOldTimingCost();

    for(vector<Terminal*>::iterator iter = destPipeliningNet->sinkVector.begin(); 
	iter != destPipeliningNet->sinkVector.end(); iter++){
      if((*iter)->retimeNerfed){
	timingCost += (*iter)->sourceDeltaTimingCostForwards();
      }
    }
  }  
  else{
    ERROR("This should not be used on a non-pipelined net!");
  }
  return timingCost;
}

//If this terminal is a source, it might fan out to more than one sink.
//Return true if this terminal has a fanout greater than one
bool Terminal::hasFanOut(){
  if(destPipeliningNet){
    return(destPipeliningNet->sinkVector.size() > 1);
  }
  else{
    ERROR("Should not be called on terminals that are not part of pipelined nets");
  }
}

//Figure out how many register banks would be freed up if we were to remove this number of
// registers from the net driven by this output terminal.
int Terminal::numRegsFreedForward(int numDelays){
  int registersFreed = 0;

  // This terminal drives a pipelining net, thus, we can assume that
  // the net it drives has sinks that are register banks.
  for(vector<Terminal*>::iterator iter = destPipeliningNet->sinkVector.begin(); iter != destPipeliningNet->sinkVector.end(); iter++){
    //Grab the register bank of this sink and see how many registers that are being used
    //If there are exactly numDelays number of registers being used or fewer, we will completely vacate this register bank
    if((*iter)->rb->getNumMappedRegs() <= numDelays){
      registersFreed += (*iter)->rb->getNumPossibleDelay();
    }
    //If there are fewer than numDelay number of registers being used, we should
    // grab more from next bank down.  Recursively call this function on the last register in the bank
    if((*iter)->rb->getNumMappedRegs() < numDelays){
      registersFreed += (*iter)->rb->getLastSource()->numRegsFreedForward(numDelays - (*iter)->rb->getNumMappedRegs());
    }
  }
  return (registersFreed);
}

//Figure out how many registers we would require if we were to add this number of registers
// to the net that drives this input terminal
int Terminal::numRegsAddedBackward(int numDelays){
  int numNotRemoving;
  //This terminal may or may not be driven by a pipelined net.
  if(!net->pipelined){
    //If this net is not a pipelined net then we will certainly need extra registers
    return (numDelays);
  }
  else{
    // This terminal is driven by a pipelining net, but it may be at latency 0
    if(!sourcePipeliningNet->source->rb){
      //In this case, even if there are other signals in the netlist that
      // would be able to satisfy the latency requirement, we will push this on
      // unzipped, then let the system take care of zipping these signals back together later.
      return (numDelays);
    }
    else{
      // If this terminal is in a register bank, check the fan-out.
      // If this net has a fan-out greater than one we can't push additional
      // registers into this register bank without screwing up the others.
      // Again, this might create unzipped nets, but we will let the system take care of these later on.
      //One more note - if this logic block directly feeds itself through some number of registers,
      // we could be removing all of the regsters from the bank we are trying to add back into.
      //For example, we allow 3 register per bankm a logic block feeds back to itself after 4 delays
      // and we are trying to retime 4 registers backwards.  We think we will be able to share the
      // register bank at that the 4th register is currently occupying - however, by the time we actually
      // add in new registers we will have removed all trace of the old registers.  We will think that
      // the sink is at latency 0 and we will grab a fresh register.
      //Thus, we will think we are freeing 2 register banks (and if we do this incorrectly) and we
      // only need one new register bank since 2 of the new registers can go into the existing bank.
      //How do we discover this scenario?
      //Number 1, this only occurs if we are looping back on ourselves.
      //Number 2, this only occurs if we are trying to eliminate enough registers on the loop that
      // we get back to the last register bank.
      if(sourcePipeliningNet->parentNet->source->lBlock == lBlock){
	if(sourcePipeliningNet->sinkVector.size() > 1){
	  //If this net has fan-out, we can't add any registers into this bank.
	  return (numDelays);
	}
	else{
	  numNotRemoving = sourcePipeliningNet->parentNet->returnLatencyOfSink(this) - numDelays;
	  if(numNotRemoving == 0){
	    // We are eliminating all of the registers on this loop.
	    // No register bank will remain to be added to.
	    return (numDelays);
	  }
	  else if(numNotRemoving < sourcePipeliningNet->source->rb->getNumMappedRegs()){
	    //We are eliminating all but a couple of registers at the end.  It is
	    // enough we reached the last register bank, but not enough to eliminate
	    // the last bank entirely.  Thus, we will have this number of registers
	    // remaining in the register bank when we want to add things back.
	    //The number of registers we will be able to fit in the last remaining
	    // register bank, then will be the capacity of the bank minus the stragglers
	    //Could this be a negative number?
	    //Do we have to do max of (0,  (numDelays - (sourcePipeliningNet->source->rb->getNumPossibleDelay() - numNotRemoving)))?
	    if(numDelays - (sourcePipeliningNet->source->rb->getNumPossibleDelay() - numNotRemoving) < 0){
	      ERROR("Need max statement!");
	    }
	    return(numDelays - (sourcePipeliningNet->source->rb->getNumPossibleDelay() - numNotRemoving));
	  }
	  else{
	    //We are not eliminating enough registers to get to the last bank.
	    return (numDelays - min(numDelays, sourcePipeliningNet->source->rb->getNumFreeRegs()));
	  }
	}
      }
      else{
	if(sourcePipeliningNet->sinkVector.size() > 1){
	  return (numDelays);
	}
	else{
	  // Figure out how much free space is left in this bank.
	  return (numDelays - min(numDelays, sourcePipeliningNet->source->rb->getNumFreeRegs()));
	}
      }
    }
  }
}

//Figure out how many register banks would be freed up if we were to remove this number of
// registers from the net that drives this terminal.
//While we are doing this, figure out if the number of registers we want to
// retime forwards is even available as an output.  That is,
// we will only allow this function to retime to a currently
// available signals if there is upstream fanout.  If we have a latency 9 
///signal in groups of 3 registers, we will only be able to retime backwards 
// 3, 6 or 9 registers if we have seen any fanout in between.
//If this fails, then we will return a very small negative number
//NOTICE - This function does *not* return an accurate number of freed registers
// if a single net drives two separate inputs of the same logic block.
//In this case, we will assume we will free almost 2x as many registers
// as we really are able to.  
//This is OK, though, since a completely accurate count is not necessary
// for things to work OK.  If we do not free enough registers, we will
// simply fail to do the forward retiming.  This failure happens even if
// we *do* accurately estimate the number of freed registers since we 
// cannot account for the grouping of registers into bank.
int Terminal::numRegsFreedBackward(int numDelays){
  int registersFreed = 0;
  //Are there any other sinks attached to the same net as this terminal is?  If so, then
  // we won't be able to remove these registers since we don't want to disturb the other signal.
  if(sourcePipeliningNet->sinkVector.size() == 1){
    //This input terminal is driven by a specific register bank
    //This input terminal is driven by a pipelining net.  Since we already
    // know that there is numDelays number of registers we can pull off
    // of this net, we can assum that the net that drives this terminal
    // is sourced in a register bank.
    registerBank *rb = sourcePipeliningNet->source->rb;

    //Grab the register bank of the source and see how many registers are held there.
    //If there are exactly numDelays number of registers being used or fewer, we will completely vacate this register bank
    if(rb->getNumMappedRegs() <= numDelays){
      registersFreed += rb->getNumPossibleDelay();
    }
    
    //If there are fewer than numDelays number of registers being used, we should grab more
    // from the next bank up.  Recursively call this funtion on the first register in the bank
    if(rb->getNumMappedRegs() < numDelays){
      registersFreed += rb->getFirstSink()->numRegsFreedBackward(numDelays - rb->getNumMappedRegs());
    }
  }
  else{
    //Once we have seen any fanout along this route, we can only retime this number
    // of registers if that output is the already available - we can't grab a signal
    // internal to a register bank.
    if(!this->signalAvailable(numDelays)){
      registersFreed = -1000000;
    }
  }

  return (registersFreed);
}

//Is this signal the direct output of a register bank?
bool Terminal::signalAvailable(int numDelays){
  if(numDelays < 0){
    return false;
  }
  else if(numDelays == 0){
    return true;
  }
  else{
    return (sourcePipeliningNet->source->rb->getFirstSink()->signalAvailable(numDelays - sourcePipeliningNet->source->rb->getNumMappedRegs()));
  }
}

//Figure out how many registers we would require if we were to add this number of registers
// to the net that this output terminal drives.  We will assume full sharing.
int Terminal::numRegsAddedForward(int numDelays){
  int maxSharedRegs;
  int latency;
  bool fanOut;
  bool loopsBack;
  int numRegistersNotRemoving;

  //The easiest thing to do is just say that we will automatically put
  // these new registers into their own register bank(s) without
  // worrying about what the existing output net has.
  //return numDelays;
  
  //However, we could do some more complicated things.
  //This terminal may or may not drive a pipelined net yet
  if(!net->pipelined){
    //If this net is not a pipelined net then we will certainly need extra registers
    return (numDelays);
  }
  else{
    //If it does drive a pipelined net, we should figure out what the minimum number of
    // registers we can push into all of the existing register blocks are.
    //For example, we want to have three additional latencies after this block and there
    // is space in all of the register bank sinks of the existing net to fit
    // one more register.  Thus, we can push one register into each of the existing
    // register banks then have all of the sinks share the output of two common
    // registers.
    maxSharedRegs = BIGINTEGER;
    for(vector<Terminal*>::iterator iter = destPipeliningNet->sinkVector.begin(); 
	iter != destPipeliningNet->sinkVector.end(); iter++){
      loopsBack = false;
      fanOut = false;

      (*iter)->loopsBackTo(lBlock, &latency, &fanOut, &loopsBack);

      if(!(*iter)->rb){
	//This sink is not mapped to a register block - this means that it must be a
	// real sink - not a register.  Thus, we can't add any registers into this sink.
	maxSharedRegs = 0;
      }
      //If this logic block directly feeds itself through some number of registers,
      // we could be removing all of the regsters from the bank we are trying to add back into.
      //For example, we allow 3 register per bank, a logic block feeds back to itself after 4 delays
      // and we are trying to retime 4 registers forwards.  We think we will be able to share the
      // register bank at that the 4th register is currently occupying - however, by the time we actually
      // add in new registers we will have removed all trace of the old registers.  We will think that
      // the sink is at latency 0 and we will grab a fresh register.
      //Thus, we will think we are freeing 2 register banks (and if we do this incorrectly) and we
      // only need one new register bank since 2 of the new registers can go into the existing bank.
      //How do we discover this scenario?
      //Number 1, this only occurs if we are looping back on ourselves through this sink.
      //Number 2, this only occurs if there is no fan-out downstream of this register bank
      //Number 3, this only occurs if we are trying to eliminate enough registers on the loop that
      // we get back to the first register bank.
      else if(loopsBack){
	if(fanOut){
	  //We will not be removing any registers from this register bank due to downstream fan-out
	  if(numDelays < latency){
	    //We are going to try and add registers to this register bank
	    if((*iter)->rb->getNumFreeRegs() < maxSharedRegs){
	      maxSharedRegs = (*iter)->rb->getNumFreeRegs();
	    }
	  }
	  else if(numDelays == latency){
	    //We will delete all of the registers from this bank so we won't try to add any
	    maxSharedRegs = 0;
	  }
	  else{
	    ERROR("NumDelay > latency of sink!");
	  }
	}
	else{
	  //There is no fan-out, so we might delete all, some or none of the registers in this bank
	  if(numDelays < latency){
	    numRegistersNotRemoving = latency - numDelays;

	    //We might delete some or none of the registers in this bank
	    if(numRegistersNotRemoving >= (*iter)->rb->getNumMappedRegs()){
	      //We won't try to delete any registers from this register bank
	      if((*iter)->rb->getNumFreeRegs() < maxSharedRegs){
		maxSharedRegs = (*iter)->rb->getNumFreeRegs();
	      }
	    }
	    else{
	      //We will be deleting some, but not all of the registers in this bank
	      if(numRegistersNotRemoving == 0){
		ERROR("Shouldn't have gotten here!");
	      }
	      else{
		if((*iter)->rb->getNumPossibleDelay() - numRegistersNotRemoving < maxSharedRegs){
		  maxSharedRegs = (*iter)->rb->getNumPossibleDelay() - numRegistersNotRemoving;
		}
	      }
	    }
	  }
	  else if(numDelays == latency){
	    //We will delete all of the registers from this bank so we won't try to add any
	    maxSharedRegs = 0;
	  }
	  else{
	    ERROR("NumDelay > latency of sink!");
	  }
	}
      }

      //Grab the register bank of this sink and see how many registers are available
      else if((*iter)->rb->getNumFreeRegs() < maxSharedRegs){
	maxSharedRegs = (*iter)->rb->getNumFreeRegs();
      }
    }

    maxSharedRegs = min(maxSharedRegs, numDelays);
    return (numDelays - maxSharedRegs);
  }
}

//This terminal is the output port of a logic block.  We would like to
// remove numMove registers from each fanout of this port by disconnecting
// the pipelined net currently attached to terminal, creating a new one,
// appropriately reconnecting the fanout of this port, then returning
//  a pointer to the old net so that we can undo the move if necessary.
PipeliningNet* Terminal::detachAndReattachHereForward(Netlist *netlist, int numMove){
  PipeliningNet *oldNet = destPipeliningNet;
  Net *parentNet = destPipeliningNet->parentNet;
  string tempStr;

  //1) Create a new pipelining net for this terminal at latency 0 then add it to the
  // main net's list.
  tempStr = parentNet->name + "PipelineNet" + itoa(parentNet->registerID + 1);
  //sprintf(tempStr, "%sPipelineNet%d", tempStr, parentNet->registerID + 1);
  parentNet->registerID ++;
  destPipeliningNet = new PipeliningNet(tempStr, this, 0, parentNet);
  parentNet->pipelineNetList.push_back(destPipeliningNet);

  //2) Go through the old net, follow all of the sinks' nets until they are 
  // at latency numMove, then attach them to this terminal.
  oldNet->detachAndReattachHereForward(destPipeliningNet, netlist, numMove);

  //3) Update the latency on all of the sinks of this net.
  parentNet->subtractLatencyFromAllSinks(numMove);

  return oldNet;
}

//This terminal is the input port of a logic block.  We would like to
// remove numMove registers from the fan-in of this port by disconnecting
// the pipelined net currently attached to terminal, attaching it to one
// further upstream, getting rid of any unnecessary registers and
// pipelined nets in the process, then return a pointer to the old net so
// that we can undo the move if necessary.
//This pointer may be null if we do not get to remove any registers.
//  a pointer to the old net so that we can undo the move if necessary.
PipeliningNet* Terminal::detachAndReattachHereBackward(Netlist *netlist, int numMove){
  PipeliningNet *endNet;

  //1) We have to know where to re-attach this terminal to if we want to undo the move
  endNet = sourcePipeliningNet;

  //2) Starting at this sink, go upstream and start nerfing registers until
  // we either reach numMove registers or else there we encounter fan-out.
  //Then attach this sink to the appropriate upstream net.
  detachAndReattachHereBackward(this, netlist, sourcePipeliningNet->latency - numMove, true);

  //3) Reduce the latency on this sink.
  sourcePipeliningNet->parentNet->subtractLatencyFromThisSink(this, numMove);

  return endNet;
}

//We are trying to find the upstream pipelining net that matches
// the latency we are looking for.  We might be nerfing registers along the way,
// but maybe not.
//When we find a pipelining net at the latency we want, snip off the leg that
// we have been following and attach the provided sink.
void Terminal::detachAndReattachHereBackward(Terminal *sink, Netlist *netlist, int latency, bool nerfRegisters){
  if(sourcePipeliningNet){
    //Check to see if we have reached the desired latency
    if(sourcePipeliningNet->latency > latency){
      //Check to see if we are nerfing registers
      if(nerfRegisters){
	//If we are nerfing registers, check to see if the fanout of the source 
	// pipelining net is greater than 1
	if(sourcePipeliningNet->sinkVector.size() > 1){
	  //If the source pipelining net has fan out, we cannot nerf any more registers.
	  //Remove this sink from the source pipelining net's list of sinks
	  sourcePipeliningNet->removeSinkPlusData(this);

	  sourcePipeliningNet->source->detachAndReattachHereBackward(sink, netlist, latency, false);
	}
	else{
	  //If the source pipeling net does not have fan out, nerf this register then recursively
	  // call this function on the upstream source
	  //cout << "Nerfing register " << sourcePipeliningNet->source->name << endl;
	  sourcePipeliningNet->source->retimeNerfed = true;
	  (sourcePipeliningNet->source)->getRegisterBank()->removeEndReference(sourcePipeliningNet->source);
	  //Remove this interviening register from the parent conventional net and
	  // the host netlist list of registers.
	  //Notice we are not actually deleting the register itself yet.
	  sourcePipeliningNet->parentNet->removePipelineRegFromPRList(sourcePipeliningNet->source);
	  netlist->removePipelineRegisterFromPlacementPRVector(sourcePipeliningNet->source);

	  //Remove this net from the parent net's list of pipelining nets
	   sourcePipeliningNet->parentNet->removePipelineNetFromPNList(sourcePipeliningNet);

	  sourcePipeliningNet->source->detachAndReattachHereBackward(sink, netlist, latency, true);
	}
      }
      else{
	//If we are not nerfing registers, just call this on the upstream source
	sourcePipeliningNet->source->detachAndReattachHereBackward(sink, netlist, latency, false);
      }
    }
     
    else if(sourcePipeliningNet->latency == latency){
      //If we were still nerfing registers, remove the existing sink from the net and place
      // the new one in it's place
      if(nerfRegisters){
	//Remove this terminal from the sinks of this net and add the
	// provided sink
	//cout << "Removing sink " << name << " from pnet " << sourcePipeliningNet->name << endl;
	sourcePipeliningNet->removeSinkPlusData(this);
      }
      //The slack should be the sink required time minus the source departure time
      //  minus the delay of the wire between the two.
      //cout << "Adding sink " << sink->name << " to pnet " << sourcePipeliningNet->name << endl;
      sourcePipeliningNet->pushSinkPlusData(sink, 0.0, sink->requiredTime - sourcePipeliningNet->source->departureTime);
      sink->sourcePipeliningNet = sourcePipeliningNet;
    }
    else if(sourcePipeliningNet->latency < latency){
      ERROR("We are beyong the desired latency!");
    }
  }
  else{
    ERROR("This source has no upstream net!");
  }
}

//This terminal is the input port of a logic block.  We would like to 
// add numMove registers onto this signal.
bool Terminal::addNRegistersBackward(Netlist *netlist, int numMove, double criticalPathDelay){
  int newRegisters = 0;
  string tempStr;
  PipeliningNet *currPNet;
  PipeliningNet *nextPNet;
  Terminal *regTerm;
  registerBank *regBank;
  
  //This net is not driven by a pipelined net
  if(!net->pipelined){
    //We have to convert this regular net into a pipelined net.
    currPNet = net->makePipelined(netlist);
    newRegisters = numMove;
    
    //Remove this terminal from the newly created net.
    currPNet->removeSinkPlusData(this);
  }
  else{
    currPNet = sourcePipeliningNet;
    newRegisters = numMove;

    //Remove this terminal from the existing pnet.
    currPNet->removeSinkPlusData(this);
      
    // This terminal is driven by a pipelining net, but it may be at latency 0.
    // That is, the source is not a register.
    if(!currPNet->source->rb){
    }
    else{
      //This terminal is a register, but there might be fanout that prevents us from pushing any
      // more register into this bank.
      //We have this as # of sinks of 0 because we just removed a sink above!
      if(currPNet->sinkVector.size() > 0){
      }
      else{
	//Push as many registers as we can into this bank
	regBank = currPNet->source->rb;
	while(regBank->getNumFreeRegs() > 0 && newRegisters > 0){
	  //1) Create a new pipeline net at a latency one greater than the current one
	  //   and add it to the main net's list
	  //strcpy(tempStr, net->name);
	  //sprintf(tempStr, "%sPipelineNet%d", tempStr, net->registerID + 1);
	  tempStr = net->name + "PipelineNet" + itoa(net->registerID + 1);
	  //cout << "Adding pipelined net" << tempStr << endl;
	  nextPNet = new PipeliningNet(tempStr, currPNet->latency + 1, net);
	  net->pipelineNetList.push_back(nextPNet);

	  //2) Make a new register, but mark it as nerfed
	  //strcpy(tempStr, net->name);
	  //sprintf(tempStr, "%sReg%d", tempStr, net->registerID);
	  tempStr = net->name + "Reg" + itoa(net->registerID);
	  //cout << "Adding new register " << tempStr << " to bank " << regBank->getHostDelayBlock()->col
	  //     << " , " << regBank->getHostDelayBlock()->row << " bank " << regBank->getHostBankNum() << endl;
	  net->registerID ++;
	  regTerm = new Terminal(net, currPNet, nextPNet, tempStr);
	  regTerm->retimeNerfed = true;
	  net->pipelineRegisterList.push_back(regTerm);
	  netlist->placementPipelineRegisterVector.push_back(regTerm);
	  regTerm->departureTime = 0.0;
	  regTerm->requiredTime = criticalPathDelay;

	  //3) Make this register a sink of the current net and the source of the next net
	  //The slack should be the sink required time minus the source departure time
	  //  minus the delay of the wire between the two.
	  //That is, CPD - source DT - 0.
	  currPNet->pushSinkPlusData(regTerm, 0.0, regTerm->requiredTime - currPNet->source->departureTime);
	  nextPNet->source = regTerm;
	  
	  //4) Place the newly created register into the existing half-occupied register bank
	  regBank->placeTerminal(regTerm);
	  regTerm->setRegisterBank(regBank);

	  newRegisters--;
	  currPNet = nextPNet;
	}
      }
    }
  }

  //Now, start creating any registers we need in a randomly selected register bank
  regBank = NULL;
  while(newRegisters > 0){
    if(!regBank){
      regBank = FPGA->getRandomFreeRegBank();
      if(!regBank){
	//We have just run out of register banks.
	//Add this terminal to whatever our current pNet is (so we don't lose it)
	// and return false;
	currPNet->pushSinkPlusData(this, 0.0, 0.0);
	this->sourcePipeliningNet = currPNet;
	return false;
      }
    }
    //1) Create a new pipeline net at a latency one greater than the current one
    //   and add it to the main net's list
    //strcpy(tempStr, net->name);
    //sprintf(tempStr, "%sPipelineNet%d", tempStr, net->registerID + 1);
    tempStr = net->name + "PipelineNet" + itoa(net->registerID + 1);
    //cout << "Adding pipelined net" << tempStr << endl;
    nextPNet = new PipeliningNet(tempStr, currPNet->latency + 1, net);
    net->pipelineNetList.push_back(nextPNet);
    
    //2) Make a new register, but mark it as nerfed
    //strcpy(tempStr, net->name);
    //sprintf(tempStr, "%sReg%d", tempStr, net->registerID);
    tempStr = net->name + "Reg" + itoa(net->registerID);
    //cout << "Adding new register " << tempStr << " to bank " << regBank->getHostDelayBlock()->col
    // << " , " << regBank->getHostDelayBlock()->row << " bank " << regBank->getHostBankNum() << endl;
    net->registerID ++;
    regTerm = new Terminal(net, currPNet, nextPNet, tempStr);
    regTerm->retimeNerfed = true;
    net->pipelineRegisterList.push_back(regTerm);
    netlist->placementPipelineRegisterVector.push_back(regTerm);
    regTerm->departureTime = 0.0;
    regTerm->requiredTime = criticalPathDelay;
    
#ifdef PLACER_DEBUG
    numRegistersEstimatedNeeded --;
#endif


    //3) Make this register a sink of the curre nte and the source of the next net
    //The slack should be the sink required time minus the source departure time
    //  minus the delay of the wire between the two.
    //That is, CPD - source DT - 0.
    currPNet->pushSinkPlusData(regTerm, 0.0, regTerm->requiredTime - currPNet->source->departureTime);
    nextPNet->source = regTerm;
    
    //4) Place the newly created register into the existing half-occupied register bank
    regBank->placeTerminal(regTerm);
    regTerm->setRegisterBank(regBank);

    //5) Check to see if the register bank is full
    if(regBank->bankFull()){
      regBank = NULL;
    }
    
    newRegisters--;
    currPNet = nextPNet;    
  }

  //Attach this terminal to the current pNet;
  currPNet->pushSinkPlusData(this, 0.0, this->requiredTime - currPNet->source->departureTime);
  this->sourcePipeliningNet = currPNet;

  //Update the latency of this sink
  net->increaseSinkLatency(this, numMove);

  return true;
}

//This terminal is the output port of a logic block.  We would like to 
// add numMove registers onto this signal.
bool Terminal::addNRegistersForward(Netlist *netlist, int numMove, double criticalPathDelay){
  int newRegisters;
  int maxSharedRegs = BIGINTEGER;
  PipeliningNet *currPNet = destPipeliningNet;
  
  //This net is not driven by a pipelined net
  if(!net->pipelined){
    //We have to convert this regular net into a pipelined net.
    currPNet = net->makePipelined(netlist);
  }

  //Go through each of the sinks and figure out what the minimum number of empty slots
  // each of the sink register banks have.
  maxSharedRegs = BIGINTEGER;
  for(vector<Terminal*>::iterator iter = currPNet->sinkVector.begin(); 
      iter != currPNet->sinkVector.end(); iter++){
    if(!(*iter)->rb){
      //This sink is not mapped to a register block - this means that it must be a
      // real sink - not a register.  Thus, we can't add any registers into this sink.
      maxSharedRegs = 0;
    }
    //Grab the register bank of this sink and see how many registers are available
    else if((*iter)->rb->getNumFreeRegs() < maxSharedRegs){
      maxSharedRegs = (*iter)->rb->getNumFreeRegs();
    }
  }
  maxSharedRegs = min(maxSharedRegs, numMove);  
  
  //Go ahead and insert this number of register into each of the sink register banks
  currPNet->insertRegistersExistingSinkBanks(maxSharedRegs, netlist, criticalPathDelay);
  
  //Now insert the rest of the registers we need as shared between the 
  // source of this net and the sinks
  if(!currPNet->insertRegistersNewBanks(numMove - maxSharedRegs, netlist, criticalPathDelay)){
    return false;
  }
  
  //Update the latency of all the sinks
  currPNet->parentNet->increaseAllSinkLatency(numMove);
  
  //Go through all of the pipelining nets and update their latency
  currPNet->updateAllLatencies(0);
  return true;
}

//This net (and associated registers) used to be attached between this output 
// terminal and it's sink.  Put it back.
void Terminal::reattachNetForwards(PipeliningNet *oldPNet, Netlist *netlist, int numMove){
  //Make sure that this is the correct parent net
  if(destPipeliningNet->parentNet != oldPNet->parentNet){
    ERROR("Something is very, very wrong!");
  }

  //Get rid of the net that we created and take it out of circulation
  //cout << "Freeing net " << destPipeliningNet->name << endl;
  destPipeliningNet->parentNet->removePipelineNetFromPNList(destPipeliningNet);

  //Before we free this net, we have to account for the wire and timing cost we are
  // getting rid of.
  //cout << "Surplusing net's wire cost of " << destPipeliningNet->wireCost << " and timing cost of " 
  //     << destPipeliningNet->timingCost << endl;
  netlist->wireCostSurplus += destPipeliningNet->wireCost;
  netlist->timingCostSurplus += destPipeliningNet->timingCost;

  //I'd like to delete this net, but if it's a loopback net I can't
  // this is a small (hopefully) memory leak
  //delete destPipeliningNet;
  
  //Reattach the old net
  //cout << "Making " << name << " source of net " << oldPNet->name << endl;
  destPipeliningNet = oldPNet;

  //Add back the old nets and registers, plus reattach the old sinks
  destPipeliningNet->addBackNetAndRegistersForwards(netlist, numMove);

  //Update the latency of all the sinks
  destPipeliningNet->parentNet->increaseAllSinkLatency(numMove);
}

//This net (and associated registers) used to drive this input 
// terminal.  Put it back.
void Terminal::reattachNetBackwards(PipeliningNet *oldPNet, Netlist *netlist, int numMove){
  //Make sure that this is the correct parent net
  if(sourcePipeliningNet->parentNet != oldPNet->parentNet){
    ERROR("Something is very, very wrong!");
  }

  //Detach this sink from the existing source pipelining net
  sourcePipeliningNet->removeSinkPlusData(this);

  //Reattach the old net
  //cout << "Adding " << oldPNet->name<< " as source of terminal " << name;
  sourcePipeliningNet = oldPNet;

  //Add back the old nets and registers, plus reattach to the main net
  sourcePipeliningNet->addBackNetAndRegistersBackwards(this, netlist);

  //Add latency on this sink.
  sourcePipeliningNet->parentNet->increaseSinkLatency(this, numMove);
}

//This is different from detachAndReattachHereBackward since we don't know
// how many we should take out.  There may be 100 or 0, all we know is
// that we want to take out the nerfed registers.
//We also, obviously, don't care about undoing this move.
void Terminal::detachAndReattachHereBackwardsAborted(Netlist *netlist){
  //Starting at this sink, go upstream and start deleting nerfed registers until
  // we encounter a non-nerfed register.
  //Then make this a sink of that pnet.
  detachAndReattachHereBackwardAborted(this, netlist);

  //Correct the latency on this sink.
  sourcePipeliningNet->parentNet->setLatencyOfSink(this, sourcePipeliningNet->latency);
}

//Keep calling this function until we get an un-nerfed source.
void Terminal::detachAndReattachHereBackwardAborted(Terminal *sink, Netlist *netlist){
  Terminal *source;
  PipeliningNet *sPNet = sourcePipeliningNet;
 
  if(sPNet){
    //Detach this sink from the source net
    sPNet->removeSinkPlusData(this);

    source = sPNet->source;
    //Check to see if we have reached an un-nerfed source
    if(source->retimeNerfed){
      //We have not, so get rid of the net this register drives and keep going
      //Notice this is actually a single sink, since we just removed it above.
      if(sPNet->sinkVector.size() != 0){
	ERROR("We should not have fanout!");
      }

      //Remove this source register from it's register bank and the common lists
      source->getRegisterBank()->removeEndReference(source);
      sPNet->parentNet->removePipelineRegFromPRList(source);
      netlist->removePipelineRegisterFromPlacementPRVector(source);

      //Remove this net from the parent net's list of pipelining nets
      sPNet->parentNet->removePipelineNetFromPNList(sPNet);

      source->detachAndReattachHereBackwardAborted(sink, netlist);

      //Delete the net and the register
      //cout << "Removing register " << source->name << endl;
      //cout << "Removing net " << sPNet->name << endl;
      delete sPNet;
      delete source;
    }
     
    else{
      //We have reached the net that we should reattach this sink to.
      sPNet->pushSinkPlusData(sink, 0.0, sink->requiredTime - sPNet->source->departureTime);
      sink->sourcePipeliningNet = sPNet;
    }
  }
  else{
    ERROR("This source has no upstream net!");
  }
}

//This is different from detachAndReattachHereForward since we don't know
// how many we should take out.  There may be 100 or 0, all we know is
// that we want to take out the nerfed registers.
//We also, obviously, don't care about undoing this move.
void Terminal::detachAndReattachHereForwardsAborted(Netlist *netlist){
  //Go through the destination net, follow all of the sinks' nets until they are 
  // no longer nerfed, then attach them to this terminal.
  detachAndReattachHereForwardAborted(this, netlist);

  //Update the latency on all of the sinks of this net.
  net->updateLatencyAllSinks();
}

//Keep calling this function until we get an un-nerfed sinks
void Terminal::detachAndReattachHereForwardAborted(Terminal *source, Netlist *netlist){
  if(destPipeliningNet){
    for(vector<Terminal*>::iterator iter = destPipeliningNet->sinkVector.begin(); 
	iter != destPipeliningNet->sinkVector.end(); iter++){
      //Detach this sink from the net
      destPipeliningNet->removeSinkPlusData(*iter);

      //Check to see if we have reached an un-nerfed sink
      if((*iter)->retimeNerfed){
	//This sink is nerfed, so call this function recursively,
	//then get rid of this register and the net it drives.
      
	//Remove this source register from it's register bank and the common lists
	(*iter)->getRegisterBank()->removeStartReference(*iter);

	destPipeliningNet->parentNet->removePipelineRegFromPRList(*iter);
	netlist->removePipelineRegisterFromPlacementPRVector(*iter);

	//Remove the net this register drives from the parent net's list of pipelining nets
	destPipeliningNet->parentNet->removePipelineNetFromPNList((*iter)->destPipeliningNet);

	(*iter)->detachAndReattachHereForwardAborted(source, netlist);

	//Delete the net and the register
	cout << "Removing register " << (*iter)->name << endl;
	cout << "Removing net " << (*iter)->destPipeliningNet->name << endl;
	delete (*iter)->destPipeliningNet;
	delete *iter;
      }
     
      else{
	//We have reached a sink that we should reattach to the logic block output
	source->destPipeliningNet->pushSinkPlusData(*iter, 0.0, (*iter)->requiredTime - source->departureTime);
	(*iter)->sourcePipeliningNet = source->destPipeliningNet;

	if((*iter)->getRegisterBank()){
	  (*iter)->getRegisterBank()->updateRoutingNodes();
	}
      }
    }
  }
  else{
    ERROR("This sink has no downstream net!");
  }
}

//Does this terminal connect, eventually, to this logic block?
void Terminal::loopsBackTo(logicBlock *lb, int *latency, bool *fanOut, bool *loopsBack){
  if(!rb){
    //We reached the end of the line.  We can tell if this logic block matches the one provided.
    if(lBlock == lb){
      *loopsBack = true;
      *latency = net->returnLatencyOfSink(this);
    }
  }
  else{
    //We have not reached the end of the line yet, so call this recursively on all of the
    // sinks of the destination net of last register in this register bank.
    if(rb->getLastSource()->destPipeliningNet->sinkVector.size() > 1){
      *fanOut = true;
    }
    for(vector<Terminal*>::iterator iter = rb->getLastSource()->destPipeliningNet->sinkVector.begin();
	iter != rb->getLastSource()->destPipeliningNet->sinkVector.end(); iter++){
      (*iter)->loopsBackTo(lb, latency, fanOut, loopsBack);
    }
  }
}


//We have decided to accept the newly created registers upstream of this block, so
// un-nerf them.
void Terminal::clearNerfedBlocksBackward(){
  if(sourcePipeliningNet){
    if(sourcePipeliningNet->source->retimeNerfed){
      //cout << "Accepting register " << sourcePipeliningNet->source->name << endl;
      sourcePipeliningNet->source->retimeNerfed = false;
      sourcePipeliningNet->source->clearNerfedBlocksBackward();
    }
  }
  else{
    ERROR("This should only be called on pipelined nets!");
  }
}

//We have decided to accept the newly created registers downstream of this block, so
// un-nerf them.
void Terminal::clearNerfedBlocksForward(){
  if(destPipeliningNet){
    for(vector<Terminal*>::iterator iter = destPipeliningNet->sinkVector.begin();
	iter != destPipeliningNet->sinkVector.end(); iter++){
      if((*iter)->retimeNerfed){
	//cout << "Accepting register " << (*iter)->name << endl;
	(*iter)->retimeNerfed = false;
	(*iter)->clearNerfedBlocksForward();
      }
    }
  }
  else{
    ERROR("This should only be called on pipelined nets!");
  }
}


//This terminal is the input port of a logic block.  We have recently
// increased the number of registers on this link, so we may have
// increased the potential number of registers we can push backwards
// through the logic block that generated this signal.
void Terminal::checkBackwardsCapacity(){
  net->source->lBlock->calculateRegPushBackward();
}

//This terminal is the output port of a logic block.  We have recently
// increased the number of registers on this link, so we may have
// increased the potential number of registers we can push forwards
// through the logic block that generated this signal.
void Terminal::checkForwardsCapacity(){
  for(vector<Terminal*>::iterator iter = net->sinkVector.begin();
      iter != net->sinkVector.end(); iter++){
    (*iter)->lBlock->calculateRegPushForward();
  }
}

//Make sure there is a source and dest net for all of these registers,
// then check to make sure that the first 5 characters of the name matches 
// the name of this register.
void Terminal::checkSourceAndDestNet(){
  if(!sourcePipeliningNet || !destPipeliningNet){
    ERROR("Register should have source and dest net!");
  }
  else{
//     if(name.compare(0, 5, sourcePipeliningNet->name, 0, 5) != 0){
//       ERROR("Register should have source net with matching name!");
//     }
//     if(name.compare(0, 5, destPipeliningNet->name, 0, 5) != 0){
//       ERROR("Register should have dest net with matching name!");
//     }
  }

  //Check to see if this register is assigned to a register bank
  if(!rb){
    ERROR("Register should have a register bank!");
  }
  else{
    if(rb->getFirstSink() != this){
      //If this is not the first sink in the register bank, the source of our source pnet should
      // be mapped to this bank.
      if(sourcePipeliningNet->source->rb != rb){
	ERROR("Source register should be in same bank!");
      }
    }
    if(rb->getLastSource() != this){
      //If this is not the last source in the register bank, our destination pnet should not have
      // any fanout
      if(destPipeliningNet->sinkVector.size() != 1){
	ERROR("Non-terminal register in bank should not have fan-out!");
      }
    }
  }
}

//Return the type of block that drives this pipelining register
//Could be a LUT, a register bank, an I/O pin, RAM, MULTIPLIER or another pipelining register.
enum e_genericBlockType Terminal::returnDriverType(){
  //Pipelining register must be driven by pipelining nets
  if(!sourcePipeliningNet){
    ERROR("This pipelining register is not driven by a pipelined net!");
  }

  if(!sourcePipeliningNet->source->lBlock){
    //If this pipelining net is not driven by a logic block,
    // we know that this pipelining register is driven by another pipelining register.
    return GBT_NO_TYPE;
  }
  else{
    return (sourcePipeliningNet->source->lBlock->get_blockType());
  }
}

//Put this register bank in the same CLB as it's driver, if there is room.
//Otherwise, just put it in an empty register bank.
void Terminal::placeWithParent(){
  Terminal *sourcePR;
  logicBlock *sourceLB;
  vprCLB *sourceCLB;
  registerBank *emptyRB;

  if(!sourcePipeliningNet){
    ERROR("This pipelining register is not driven by a pipelined net!");
  }

  //This logic block is driven by a pipelining net. 
  if(!sourcePipeliningNet->source->lBlock){
    //If this pipelining net is not driven by a logic block,
    // we know that this pipelining register is driven by another pipelining register.
    sourcePR = sourcePipeliningNet->source;

    //Make sure this pipelining register is actually placed
    if(sourcePR->getLogicBlockRow() == -1 || sourcePR->getLogicBlockCol() == -1){
      ERROR("Pipelining register is not placed!");
    }

    //This is a child pipelining register of the parent pipelining register.
    //If there is room, try placing it in the same register bank.
    if(!sourcePR->rb->bankFull()){
      sourcePR->rb->placeTerminal(this);
      return;
    }
    else{
      //If the register bank is full, try to put it in the same CLB.
      sourceCLB = FPGA->getCLBAt(sourcePR->getLogicBlockRow(), sourcePR->getLogicBlockCol());
    }
  }
  else{
    //This pipelining register is driven by a logic block
    sourceLB = sourcePipeliningNet->source->lBlock;

    if(sourceLB->get_blockType() == INPUTPAD){
      emptyRB = FPGA->getRandomFreeRegBank();
      emptyRB->placeTerminal(this);
      return;
    }

    //Make sure this logic block is actually placed
    if(sourceLB->row == -1 || sourceLB->col == -1){
      ERROR("Logic block is not placed!");
    }
    
    sourceCLB = FPGA->getCLBAt(sourceLB->row, sourceLB->col);
  }

  //Find out if there is a free register bank in the same CLB as the source
  emptyRB = sourceCLB->getFirstEmptyRegisterBank();

  if(emptyRB){
    emptyRB->placeTerminal(this);
  }
  else{
    emptyRB = FPGA->getRandomFreeRegBank();
    emptyRB->placeTerminal(this);
  }
}

//See if the driver of this pipelining register has been placed
bool Terminal::parentPlaced(){
  Terminal *sourcePR;
  logicBlock *sourceLB;
  vprCLB *sourceCLB;
  registerBank *emptyRB;

  if(!sourcePipeliningNet){
    ERROR("This pipelining register is not driven by a pipelined net!");
  }

  //This logic block is driven by a pipelining net. 
  if(!sourcePipeliningNet->source->lBlock){
    //If this pipelining net is not driven by a logic block,
    // we know that this pipelining register is driven by another pipelining register.
    sourcePR = sourcePipeliningNet->source;

    //Make sure this pipelining register is actually placed
    if(sourcePR->getLogicBlockRow() == -1 || sourcePR->getLogicBlockCol() == -1){
      return false;
    }
  }
  else{
    //This pipelining register is driven by a logic block
    sourceLB = sourcePipeliningNet->source->lBlock;

    //Make sure this logic block is actually placed
    if(sourceLB->row == -1 || sourceLB->col == -1){
      return false;
    }
  }
  return true;
}

//Go through all of the sinks of this terminal.
//
void Terminal::printDistanceToAllSinks(){
  if(destPipeliningNet){
    destPipeliningNet->printDistanceToAllSinks();
  }

  else if(net){
    net->printDistanceToAllSinks();
  }
}

//Go through all of the sinks of this terminal and figure out what the closest sink is
int Terminal::minDistanceToAllSinks(){
  if(destPipeliningNet){
    return(destPipeliningNet->minDistanceToAllSinks());
  }

  else if(net){
    return(net->minDistanceToAllSinks());
  }
}
