// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern vprFPGA *FPGA;

//Remove the placement for all luts and FF defined in this logical CLB
// from the host physical CLB of this register bank
void c_clb::unPlaceLogicBlocksFromLUTsAndFF(registerBank *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->getHostCLB()->row;
  int targetCol = block->getHostCLB()->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  registerBank* targetRB;
  int i;

  //Clear the CLB of any location information
  clearLocation(block);  

  //Go through all of the LUTs associated with this clb
  if(lutVector.size() > targetCLB->getNumLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }
  for(iter = lutVector.begin(), i = 0; iter != lutVector.end(); iter++, i++){
    if((*iter)->row != targetRow || (*iter)->col != targetCol){
      ERROR("LUT does not have correct current placement!");
    }
    else{
      unPlaceBlock((*iter), (void *) targetCLB, LB, (*iter)->ptc_index);
    }
  }

  if(registerVector.size() > targetCLB->getNumRegBanks()){
    ERROR("We cannot handle that many FFs in this CLB!");
  }
  for(iter = registerVector.begin(), i = 0; iter != registerVector.end(); iter++, i++){
    if((*iter)->row != targetRow || (*iter)->col != targetCol){
      ERROR("FF does not have correct current placement!");
    }
    else{
      targetRB = targetCLB->getRegBank(i);
      targetRB->unPlaceLogicBlock(*iter);
    }
  }
}

//Remove the placement for all luts and FF defined in this logical CLB
// from the host physical CLB of this LUT
void c_clb::unPlaceLogicBlocksFromLUTsAndFF(vprLUT *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->row;
  int targetCol = block->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  registerBank* targetRB;
  int i;

 //Clear the CLB of any location information
  clearLocation(block);  

  //Go through all of the LUTs associated with this clb
  if(lutVector.size() > targetCLB->getNumLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }
  for(iter = lutVector.begin(), i = 0; iter != lutVector.end(); iter++, i++){
    if((*iter)->row != targetRow || (*iter)->col != targetCol){
      ERROR("LUT does not have correct current placement!");
    }
    else{
      unPlaceBlock((*iter), (void *) targetCLB, LB, (*iter)->ptc_index);
    }
  }

  if(registerVector.size() > targetCLB->getNumRegBanks()){
    ERROR("We cannot handle that many FFs in this CLB!");
  }
  for(iter = registerVector.begin(), i = 0; iter != registerVector.end(); iter++, i++){
    if((*iter)->row != targetRow || (*iter)->col != targetCol){
      ERROR("FF does not have correct current placement!");
    }
    else{
      targetRB = targetCLB->getRegBank(i);
      targetRB->unPlaceLogicBlock(*iter);
    }
  }
}

//Remove the placement for all luts defined in this logical CLB
// from the host physical CLB of this LUT
void c_clb::unPlaceLogicBlocksFromLUTs(vprLUT *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->row;
  int targetCol = block->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  int i;

 //Clear the CLB of any location information
  clearLocation(block);  

  //Go through all of the LUTs associated with this clb
  if(lutVector.size() > targetCLB->getNumLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }
  for(iter = lutVector.begin(), i = 0; iter != lutVector.end(); iter++, i++){
    if((*iter)->row != targetRow || (*iter)->col != targetCol){
      ERROR("LUT does not have correct current placement!");
    }
    else{
      unPlaceBlock((*iter), (void *) targetCLB, LB, (*iter)->ptc_index);
    }
  }
}

//Place all LUTs and FF defined in this logical CLB
//into the host CLB of this register bank
void c_clb::placeLogicBlocksToLUTsAndFF(registerBank *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->getHostCLB()->row;
  int targetCol = block->getHostCLB()->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  registerBank* targetRB;
  int i;

  //Update the position
  updateLocation(block);  

  //Go through all of the LUTs associated with this CLB
  if(lutVector.size() >  targetCLB->getNumLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }
  for(iter = lutVector.begin(), i = 0; iter != lutVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      placeBlock((*iter), (void *) targetCLB, LB, i);
    }
    else{
      ERROR("LUT has current placement!");
    }
  }
  if(registerVector.size() > targetCLB->getNumRegBanks()){
    ERROR("We cannot handle that many FFs in this CLB!");
  }
  
  for(iter = registerVector.begin(), i = 0; iter != registerVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      targetRB = targetCLB->getRegBank(i);
      targetRB->placeLogicBlock(*iter);
    }
    else{
      ERROR("FF has current placement!");
    }
  }
}

//Place all LUTs and FF defined in this logical CLB
//into the host CLB of this LUT
void c_clb::placeLogicBlocksToLUTsAndFF(vprLUT *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->row;
  int targetCol = block->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  registerBank* targetRB;
  int i;

  //Update the position
  updateLocation(block);  

  //Go through all of the LUTs associated with this CLB
  if(lutVector.size() >  targetCLB->getNumLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }
  for(iter = lutVector.begin(), i = 0; iter != lutVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      placeBlock((*iter), (void *) targetCLB, LB, i);
    }
    else{
      ERROR("LUT has current placement!");
    }
  }
  if(registerVector.size() > targetCLB->getNumRegBanks()){
    ERROR("We cannot handle that many FFs in this CLB!");
  }
  
  for(iter = registerVector.begin(), i = 0; iter != registerVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      targetRB = targetCLB->getRegBank(i);
      targetRB->placeLogicBlock(*iter);
    }
    else{
      ERROR("FF has current placement!");
    }
  }
}

//Place all LUTs defined in this logical CLB
//into the host CLB of this LUT
void c_clb::placeLogicBlocksToLUTs(vprLUT *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->row;
  int targetCol = block->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  int i;

  //Update the position
  updateLocation(block);  

  //Go through all of the LUTs associated with this CLB
  if(lutVector.size() >  targetCLB->getNumLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }
  for(iter = lutVector.begin(), i = 0; iter != lutVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      placeBlock((*iter), (void *) targetCLB, LB, i);
    }
    else{
      ERROR("LUT has current placement!");
    }
  }
}

//Update the sink delays and slacks for all blocks in this logical CLB
void c_clb::updateSinkDelaysAndSlacksForLUTsAndFF(){
  vector<logicBlock *>::iterator iter;

  //Go through all of the LUTs associated with this CLB
  for(iter = lutVector.begin(); iter != lutVector.end(); iter++){
    (*iter)->updateSinkDelaysAndSlacks();
  }

  //Do the same for the FFs
  for(iter = registerVector.begin(); iter != registerVector.end(); iter++){
    (*iter)->updateSinkDelaysAndSlacks();
  }
}

//Update the sink delays and slacks for the LUTs in this logical CLB
void c_clb::updateSinkDelaysAndSlacksForLUTs(){
  vector<logicBlock *>::iterator iter;

  //Go through all of the LUTs associated with this CLB
  for(iter = lutVector.begin(); iter != lutVector.end(); iter++){
    (*iter)->updateSinkDelaysAndSlacks();
  }
}

//Update the wire and timing costs for all blocks in this logical CLB
void c_clb::updateCostsAndDeltasForLUTsAndFF(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vector<logicBlock *>::iterator iter;
  
  //Go through all of the LUTs associated with this CLB
  for(iter = lutVector.begin(); iter != lutVector.end(); iter++){
    (*iter)->updateCostsAndDeltas(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  }

  //Do the same for the FFs
  for(iter = registerVector.begin(); iter != registerVector.end(); iter++){
    (*iter)->updateCostsAndDeltas(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  }
}

//Update the wire and timing costs for all blocks in this logical CLB
void c_clb::updateCostsAndDeltasForLUTsAndFFCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  vector<logicBlock *>::iterator iter;
  
  //Go through all of the LUTs associated with this CLB
  for(iter = lutVector.begin(); iter != lutVector.end(); iter++){
    (*iter)->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  }

  //Do the same for the FFs
  for(iter = registerVector.begin(); iter != registerVector.end(); iter++){
    (*iter)->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  }
}

//Update the wire and timing costs for the LUTs in this logical CLB
void c_clb::updateCostsAndDeltasForLUTs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vector<logicBlock *>::iterator iter;
  
  //Go through all of the LUTs associated with this CLB
  for(iter = lutVector.begin(); iter != lutVector.end(); iter++){
    (*iter)->updateCostsAndDeltas(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  }
}

//Update the wire and timing costs for the LUTs in this logical CLB
void c_clb::updateCostsAndDeltasForLUTsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  vector<logicBlock *>::iterator iter;
  
  //Go through all of the LUTs associated with this CLB
  for(iter = lutVector.begin(); iter != lutVector.end(); iter++){
    (*iter)->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  }
}
