// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

// Set name of LUT/Register
void mcncBlock::set_name(char* n){
  if(magicNum!=12118)
    ERROR("trying to access a deleted node...");
  name = n;
}

// Add a net to inputList
void mcncBlock::addto_inputList(mcncNet* p_net){
  //if(!p_net)
  //  ERROR("cannot add NULL input net...");
  if(magicNum!=12118)
    ERROR("trying to access a deleted node...");
  inputList.push_back(p_net);
  numInputNets++;
}

// Add a net to outputList
void mcncBlock::addto_outputList(mcncNet* p_net){
  //if(!p_net)
  //  ERROR("cannot add NULL output net...");
  if(magicNum!=12118)
    ERROR("trying to access a deleted node...");
  outputList.push_back(p_net);
  numOutputNets++;
}

void mcncBlock::set_blockType(enum e_mcncBlockType t){
  if(magicNum!=12118)
    ERROR("trying to access a deleted node...");
  if(t < MCNC_INPUT || t > MCNC_REG)
    ERROR("Invalid clb type");
  blockType = t;
}

enum e_mcncBlockType mcncBlock::get_blockType(){
  if(magicNum!=12118)
    ERROR("trying to access a deleted node...");
  return (blockType);
}

// Get name of block
string mcncBlock::get_name(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  return name;
}

// Get num input nets
int mcncBlock::get_numInputNets(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  return inputList.size();
}

// Get num output nets
int mcncBlock::get_numOutputNets(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  return outputList.size();
}

// Create a new iterator for inputList
void mcncBlock::create_inputListItr(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  if(input_itr)
    ERROR("an iterator already exists...");
  input_itr = new list<mcncNet*>::iterator;
  *input_itr = inputList.begin();
}

// Remove an already existing inputList iterator
void mcncBlock::remove_inputListItr(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  if(input_itr == 0)
    ERROR("cannot delete NULL iterator...");
  delete input_itr;
  input_itr = 0;
}

// Get next element from inputList
mcncNet* mcncBlock::getnext_inputList(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");

  if(*input_itr == inputList.end())
    return 0;
  else{
    mcncNet* ret_ptr = **input_itr;
    (*input_itr)++;
    return ret_ptr;
  }

}

// Create a new iterator for outputList
void mcncBlock::create_outputListItr(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  if(output_itr)
    ERROR("an iterator already exists...");
  output_itr = new list<mcncNet*>::iterator;
  *output_itr = outputList.begin();
}

// Remove an already existing outputList iterator
void mcncBlock::remove_outputListItr(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");
  if(output_itr == 0)
    ERROR("cannot delete NULL iterator...");
  delete output_itr;
  output_itr = 0;
}

// Get next element from outputList
mcncNet* mcncBlock::getnext_outputList(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");

  if(*output_itr == outputList.end())
    return 0;
  else{
    mcncNet* ret_ptr = **output_itr;
    (*output_itr)++;
    return ret_ptr;
  }
}


// Print info about this clb. Print name, input and output nets.
void mcncBlock::print_Info(){
  if(magicNum!=12118)
    ERROR("trying to access deleted node...");

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR "<<name;

  //print type
  if(blockType == MCNC_INPUT)
    cout<<endl<<"INPUT BLOCK";
  else if(blockType == MCNC_OUTPUT)
    cout<<endl<<"OUTPUT BLOCK";
  else if(blockType == MCNC_CLB)
    cout<<endl<<"LOGIC BLOCK";
  else if(blockType == MCNC_LUT)
   cout<<endl<<"\tLUT";
  else if(blockType == MCNC_REG)
   cout<<endl<<"\tREGISTER";
  else{
    ERROR("Unknown block type!");
  }  


  if(blockType != MCNC_CLB){
    //print input list
    cout<<endl<<"INPUT NETS -->";
    list<mcncNet*>::iterator input_itr = inputList.begin();
    vector<int>::iterator latencyIter = inputLatencyVector.begin();
    for(;input_itr!=inputList.end() && latencyIter != inputLatencyVector.end(); input_itr++, latencyIter++){
      if(*input_itr){
	cout<<endl<<"\t"<<(*input_itr)->get_name() << " delay " << *latencyIter;
      }
    }

    //print output list
    cout<<endl<<"OUTPUT NETS -->";
    list<mcncNet*>::iterator output_itr = outputList.begin();
    for(;output_itr!=outputList.end();output_itr++){
      if(*output_itr){
	cout<<endl<<"\t"<<(*output_itr)->get_name(); 
      }
    }
  }

  else{
    cout << endl << "LUTS -->";
    for(vector <mcncBlock*>::iterator iter = luts.begin(); iter != luts.end(); iter++){
      (*iter)->print_Info();
    }
    cout << endl << "REGISTERS -->";
    for(vector <mcncBlock*>::iterator iter = registers.begin(); iter != registers.end(); iter++){
      (*iter)->print_Info();
    }
  }
}

void mcncBlock::pushBackInputLatencyVector(int latency){
  inputLatencyVector.push_back(latency);
}
