// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

//extern bool pipeline_aware;
extern bool placementRetiming;

// Set name of mcnc netlist to n.
void mcncNetlist::set_name(char* n){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  name.assign(n);
}

// Add p_block to list of blocks.
// This function assumes that p_block has been updated
// to reflect whether it is a logic, input-only or
// output-only block. 
void mcncNetlist::addto_mcncBlockList(mcncBlock* p_block){
  if(!p_block)
    ERROR("block cannot be NULL...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  mcncBlockList.push_back(p_block);

  if(p_block->get_blockType() == MCNC_CLB)
    numClbs++;
  else if(p_block->get_blockType() == MCNC_INPUT)
    numInputs++;
  else if(p_block->get_blockType() == MCNC_OUTPUT)
    numOutputs++;
  else{
    ERROR("Trying to push non-CLB/IO type into list!");
  }
}

// Add p_net to list of nets and increment num_nets.
void mcncNetlist::addto_mcncNetList(mcncNet* p_net){
  if(!p_net)
    ERROR("net cannot be NULL...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  mcncNetList.push_back(p_net);
  numNets++;
}

// Get total num clbs in this netlist.
int mcncNetlist::get_numClbs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numClbs;
}

// Get total num input-only IO blocks in this netlist.
int mcncNetlist::get_numInputs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get total num output-only IO blocks in this netlist.
int mcncNetlist::get_numOutputs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Get total num nets in this netlist.
int mcncNetlist::get_numNets(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numNets;
}

// Get the name of this netlist
string mcncNetlist::get_name(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Print details of each block in BlockList.
void mcncNetlist::print_mcncBlockList(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT BLOCK LIST";
  cout<<endl<<"***************************************************************";

  list<mcncBlock *>::iterator block_itr = mcncBlockList.begin();
  for(;block_itr!=mcncBlockList.end();block_itr++)
    (*block_itr)->print_Info();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT BLOCK LIST";
  cout<<endl<<"***************************************************************";

}

// Print details of each net in NetList.
void mcncNetlist::print_mcncNetList(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT NET LIST";
  cout<<endl<<"***************************************************************";

  list<mcncNet*>::iterator net_itr = mcncNetList.begin();
  for(;net_itr!=mcncNetList.end();net_itr++)
    (*net_itr)->print_NetInfo();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT NET LIST";
  cout<<endl<<"***************************************************************";

}

// Print the name of this mcnc netlist.
void mcncNetlist::print_name(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  cout<<endl<<endl<<"NAME OF NETLIST IS "<<name;
}

// Print the details of block pointed to by p_block.
void mcncNetlist::print_mcncBlock(mcncBlock* p_block){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!p_block)
    ERROR("block cannot be NULL...");
  p_block->print_Info();
}

// Print the details of net pointed to by p_net.
void mcncNetlist::print_mcncNet(mcncNet* p_net){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!p_net)
    ERROR("net cannot be NULL...");
  p_net->print_NetInfo();
}

// Look for block with block_name in BlockList. If found, return the block.
mcncBlock* mcncNetlist::found_mcncBlock(char* block_name){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!block_name)
    ERROR("block name cannot be NULL...");
  list<mcncBlock*>::iterator block_itr = mcncBlockList.begin();
  for(;block_itr!=mcncBlockList.end();block_itr++){
    if(strcmp((*block_itr)->get_name().c_str(),block_name)==0)
      return *block_itr;
  }
  return 0;
}

// Look for net with net_name in NetList. If found, return the net.
mcncNet* mcncNetlist::found_mcncNet(char* net_name){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!net_name)
    ERROR("net name cannot be NULL...");
  list<mcncNet*>::iterator net_itr = mcncNetList.begin();
  for(;net_itr!=mcncNetList.end();net_itr++){
    if(strcmp((*net_itr)->get_name().c_str(),net_name)==0)
      return *net_itr;
  }
  return 0;
}

// Convert this mcnc netlist to a generic netlist.
// This function will have to be defined and called by a user routine.
// The objective here is to convert the user's netlist data-structs
// to the genric format that Independence will work with.
void mcncNetlist::convert2Generic(Netlist *netlist){
  logicBlock* lb;
  c_clb* parentCLB;

  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  // give the generic netlist a name
  netlist->set_name(name);

  // for each block in the mcnc netlist, create a matching one in generic netlist
  for(list<mcncBlock*>::iterator block_itr = mcncBlockList.begin(); block_itr!=mcncBlockList.end(); block_itr++){
    //cout << endl << "Converting block " << (*block_itr)->get_name() << endl;
    
    //If it is a I/O pin (or at some point a multiplier/RAM), just make a logic block
    if((*block_itr)->blockType != MCNC_CLB){
      lb = new logicBlock;
      
      // give the logic block a name
      lb->set_name((*block_itr)->get_name());

      if((*block_itr)->blockType == MCNC_INPUT){
	lb->set_blockType(INPUTPAD);
      }
      else if((*block_itr)->blockType == MCNC_OUTPUT){
      	lb->set_blockType(OUTPUTPAD);
      }
      else{
	ERROR("Unknown block type!");
      }
      addInputsOutputsFromMCNC(netlist, lb, *block_itr);

      // add lb to list of logicBlocks in n
      netlist->addto_lbList(lb);
    }
    //However, if it is a CLB we should make a new clb, then create logic blocks inside of it
    else{
      parentCLB = new c_clb((*block_itr)->get_name());
      netlist->clbList.push_back(parentCLB);

      //Go through all of the LUTs in the clb
      for(vector <mcncBlock*>::iterator lutIter = (*block_itr)->luts.begin(); 
	  lutIter != (*block_itr)->luts.end(); lutIter++){
	lb = new logicBlock;
      
	// give the logic block a name
	lb->set_name((*lutIter)->get_name());

	lb->set_blockType(LUT);

	addInputsOutputsFromMCNC(netlist, lb, *lutIter);
	netlist->addto_lbList(lb);
	
	parentCLB->lutVector.push_back(lb);
	lb->parentCLB = parentCLB;
      }
      
      //Go through all of the registers in the clb
      for(vector <mcncBlock*>::iterator regIter = (*block_itr)->registers.begin(); 
	  regIter != (*block_itr)->registers.end(); regIter++){
	lb = new logicBlock;
      
	// give the logic block a name
	lb->set_name((*regIter)->get_name());

	lb->set_blockType(REGISTER_BANK);

	addInputsOutputsFromMCNC(netlist, lb, *regIter);
	netlist->addto_lbList(lb);

	parentCLB->registerVector.push_back(lb);
	lb->parentCLB = parentCLB;
      }

    }
  }

  //cout << "Creating lbVector" << endl;
  //Now create a vector of logic blocks so that we can more easily access these items
  netlist->create_lbVector();

  /* for each net, create vector of sinks, and populate */
  //netlist->create_netSinkVectors();

  netlist->calculateNumBlocks();
  netlist->create_slackVectors();
  netlist->create_linkDelayVectors();
  if(placementRetiming){
    netlist->calculateRegPushes();
  }
}      
 
//Add the inputs and outputs from this mcnc block
void addInputsOutputsFromMCNC(Netlist *netlist, logicBlock* lb, mcncBlock* block){
  int count;
  int y;
  list<mcncNet*>::iterator n_itr;
  mcncNet* mcnc_net;
  Net* in_net;
  Net* out_net;
  Terminal* input_terminal;
  Terminal* output_terminal;
 
  //Go through input nets of mcnc block
  count = 0;
  for(y = 0, n_itr = block->inputList.begin(); n_itr !=block->inputList.end(); n_itr++, y++){
    mcnc_net = *n_itr;

    // find input net
    //The input net might be open
    if(!mcnc_net){
      continue;
    }

    //If not, then process it
    //cout << endl << "Looking at input # " << y<<endl;
    //cout << "Net = " << mcnc_net->get_name() <<endl;
    //cout << "Latency = " << block->inputLatencyVector.at(y) <<endl;

    in_net = netlist->found_Net(mcnc_net->get_name());

    if(!in_net){
      in_net = createNet(mcnc_net);
      netlist->addto_netList(in_net);
    }
      
    // create a new input terminal
    input_terminal = createTerminal(lb, in_net, INPUT_TERMINAL, false,count);
    lb->addto_inputList(input_terminal);
      //lb->print_Info();

    in_net->addto_sinkVector(input_terminal);
    in_net->addto_latencyVector(block->inputLatencyVector.at(y));
      
    //Check to see if this net has pipelined signals in it
    if(block->inputLatencyVector.at(y) != 0){
      in_net->pipelined = true;
    }
    count++;
  }

  //Go through output nets of mcnc block
  count = 0;
  for(y = 0, n_itr = block->outputList.begin(); n_itr !=block->outputList.end(); n_itr++, y++){
    mcnc_net = *n_itr;
    
    // find output net
    //The output net might be open
    if(!mcnc_net){
      continue;
    }
    
    //If not, then process it
    //cout << endl << "Looking at output # " << y<<endl;
    //cout << "Net = " << mcnc_net->get_name() <<endl;

    out_net = netlist->found_Net(mcnc_net->get_name());

    if(!out_net){
      out_net = createNet(mcnc_net);
      netlist->addto_netList(out_net);
    }
      
    // create a new output terminal
    output_terminal = createTerminal(lb, out_net, OUTPUT_TERMINAL, true, count);
    lb->addto_outputList(output_terminal);
    out_net->set_source(output_terminal);
    //lb->print_Info();

    count++;
  }

  //Set all outputs to depend upon all inputs
  //and all input to fan out to all outputs
  lb->setOutputDependency();
  lb->setInputFanOut();  
}


// this function creates a new Net object based on the mcncNet pointed
// to by n. 
Net* createNet(mcncNet* n){
  if(n==0)
    ERROR("NULL mcnc net here...");
  Net* new_net = new Net;
  new_net->set_name(n->get_name());
  return new_net;
}

// create a new Terminal object. this object represents a terminal of the logicBlock
// lb which is a source / sink of tnet. term_type is a user-defined terminal type,
// is_source indicates whether this terminal is a source or sink of tnet, and 
// term_num is used to derive the name of this terminal.
Terminal* createTerminal(logicBlock* lb, Net* tnet, enum e_terminalType term_type,
                                           bool is_source, int term_num){
  //char term_name [MAX_STR_LEN];
  string term_name;

  if(lb==0)
    ERROR("NULL logic block...");
  if(tnet==0)
    ERROR("NULL net here...");
  if(term_type != INPUT_TERMINAL && term_type != OUTPUT_TERMINAL)
    ERROR("faulty terminal types...");
  
  if(term_type == INPUT_TERMINAL){
    //sprintf(term_name,"%s_INPUT%d",lb->get_name(),term_num);
    term_name = lb->get_name() + "_INPUT" + itoa(term_num);
  }
  else{
    //sprintf(term_name,"%s_OUTPUT%d",lb->get_name(),term_num);
    term_name = lb->get_name() + "_OUTPUT" + itoa(term_num);
  }

  Terminal* new_terminal = new Terminal(term_name, lb, tnet, term_type, is_source);
  
  return new_terminal;
}


