// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition for an mcnc netlist. Note that this class 
definition is specific to mcnc netlists. For every new benchmark
set, a separate class definition will have to be created and
implemented.
****************************************************************/ 
#ifndef MCNC_NETLIST_H
#define MCNC_NETLIST_H

class mcncBlock;
class mcncNet;
class Netlist;
class Terminal;
class logicBlock;

class mcncNetlist{
  //char name [MAX_STR_LEN];  //name of mcnc netlist
  string name;  //name of mcnc netlist
  list<mcncBlock*> mcncBlockList;  //list of I/O pins and clbs in mcnc netlist
  list<mcncNet*> mcncNetList;  //list of nets in mcnc netlist
  int numClbs;  //num clbs in mcnc netlist
  int numInputs;  //num input-only IO blocks
  int numOutputs;  //num output-only IO blocks
  int numNets;  //num nets in mcnc netlist
  int magicNum;
 public:
  mcncNetlist(){
    numClbs=0;
    numNets=0;
    numInputs=numOutputs=0;
    magicNum=11111;
  }

  ~mcncNetlist(){
    if(magicNum!=11111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
    
    for(list<mcncBlock*>::iterator blockIter = mcncBlockList.begin(); blockIter != mcncBlockList.end(); blockIter++){
      delete *blockIter;
    }
    for(list<mcncNet*>::iterator netIter = mcncNetList.begin(); netIter != mcncNetList.end(); netIter++){
      delete *netIter;
    }
  }

  void set_name(char* n);  //set the name of mcnc netlist
  void addto_mcncBlockList(mcncBlock* p_clb);  //add a clb to mcncClbList
  void addto_mcncNetList(mcncNet* p_net);  //add a net to mcncNetList

  int get_numClbs();  //get num clbs in mcnc netlist
  int get_numNets();  //get num nets in mcnc netlist
  string get_name();  //get the name of mcnc netlist
  int get_numInputs();  //total num input-only IO blocks
  int get_numOutputs();  //total num output-only IO blocks

  void print_mcncBlockList();  //print the entire list of clbs & I/O pins
  void print_mcncNetList();  //print the entire list of nets
  void print_name();  //print the name of mcnc netlist
  void print_mcncBlock(mcncBlock* p_block);  //print the details of clb pointed to by p_block
  void print_mcncNet(mcncNet* p_net);    //print the details of net pointed to by p_net

  mcncBlock* found_mcncBlock(char* block_name);  //look for block in BlockList. if found, return block.
  mcncNet* found_mcncNet(char* net_name);  //look for net in NetList. if found, return net.

  void convert2Generic(Netlist *n);  //convert this mcnc netlist to a generic netlist
};

void addInputsOutputsFromMCNC(Netlist* netlist, logicBlock* lb, mcncBlock* block);
Net* createNet(mcncNet* n);  //create a new Net object based on the mcncNet pointed
                             //to by n.
Terminal* createTerminal(logicBlock* lb,Net* tnet, enum e_terminalType term_type, bool is_source,
			 int term_num);  //create a new Terminal object

#endif
