// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern int MAXID;
extern unsigned int *VECTORMASK;

int MAXIDpathVectorSize(){
  return ((MAXID/NUMVECTORBITS) + 1);
}

int pathVectorSize(vector <unsigned int> *pathVector){
  return (pathVector->size() * NUMVECTORBITS);
}

bool checkNodeID(vector <unsigned int> *pathVector, int nodeID){
//   cout << "\t\tCheckNodeID Path Vector = ";
//   for(vector <unsigned int>::iterator iter = pathVector->begin(); iter != pathVector->end(); iter++){
//     printf("%x ", *iter);
//   }
//   cout << endl;

  //This nodeID might be larger than what the pathvector can hold
  if(nodeID >= pathVectorSize(pathVector)){
    return false;
  }

  //printf("\t\tCheckNodeID comparision = %x\n", pathVector->at(nodeID / NUMVECTORBITS) & VECTORMASK[nodeID % NUMVECTORBITS]);
  if(pathVector->at(nodeID / NUMVECTORBITS) & VECTORMASK[nodeID % NUMVECTORBITS]){
    return true;
  }
  else{
    return false;
  }
}

void setNodeID(vector <unsigned int> *pathVector, int nodeID){
  pathVector->at(nodeID / NUMVECTORBITS) = pathVector->at(nodeID / NUMVECTORBITS) | VECTORMASK[nodeID % NUMVECTORBITS];

  //cout << "Path Vector = ";
  //for(vector <unsigned int>::iterator iter = pathVector->begin(); iter != pathVector->end(); iter++){
  //  printf("%x ", *iter);
  //}
  //cout << endl;
}

//Create the base path bit-vector for a given pipelined net.
//1) A search started from a pre-existing route should not be able to explore any node
//that is directly along it's path back to the net source.  It can explore nodes used along
//other branches, not directly upstream of itself even at earlier latencies or nodes used by
//"itself" at later latencies.
vector <unsigned int> * createPathVectorRoot(Net* net, PipeliningNet* pNet, list<pExplore> *cleanupList){
  vector <unsigned int> *pathVector;
  PipeliningNet* currPNet;
  Terminal* currSink;
  treeNode* currTreeNode;
  list<PipeliningNet *>::iterator pNetIter;

  //Make a new pathvector
  pathVector = new vector<unsigned int>(MAXIDpathVectorSize());
  
  //Go through all pipelined nets and add in all nets at the same latency as the given one
  //for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
  //  if((*pNetIter)->latency == pNet->latency){
  //    addTree2PathVector((*pNetIter)->rTree.root, pathVector, cleanupList);
  //  }
  //}
  
  //Add in all nodes directly upstream of this pipelined net
  currSink = pNet->source;
  currPNet = currSink->sourcePipeliningNet;
  
  //Keep going until we don't have an upstream pipelined net
  while(currPNet){
    //Find the current sink in the list of this pipelining nets
    currTreeNode = currPNet->rTree.getNode(currSink->rNode);

    //Add this node to the pathvector
    addNode2PathVector(currTreeNode->rNode, pathVector, cleanupList);

    //Follow this back to the source of the pipelining net
    while(currTreeNode->parent){
      currTreeNode = currTreeNode->parent;
      addNode2PathVector(currTreeNode->rNode, pathVector, cleanupList);
    }
    currSink = currPNet->source;
    currPNet = currSink->sourcePipeliningNet;
  }
  
//   cout << "\tCreatePathVectorRoot Path Vector = ";
//   for(vector <unsigned int>::iterator iter = pathVector->begin(); iter != pathVector->end(); iter++){
//     printf("%x ", *iter);
//   }
//   cout << endl;

  return pathVector;
}

void addTree2PathVector(treeNode* tree_node, vector <unsigned int> *pathVector, list<pExplore> *cleanupList){
  treeNode* t = tree_node;
  treeNode* child;
  
  //This net may have nothing in it's routing tree
  if(!t){
    //cout << "Net does not have anything in routing tree!" << endl;
    return;
  }

  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");

  //Add this node to the pathvector
  addNode2PathVector(t->get_rNode(), pathVector, cleanupList);

  //Call this recursively on the rest of the tree
  t->create_childListItr();
  for(child = t->getnext_childList(); child!=0; child=t->getnext_childList()){
    addTree2PathVector(child, pathVector, cleanupList);
  }
  t->remove_childListItr();
}


void addNode2PathVector(routingNode *rNode, vector <unsigned int> *pathVector, list<pExplore> *cleanupList){
  //Check to see if the rNode has an ID
  if(rNode->nodeID == 0){
    rNode->nodeID = MAXID + 1;
    MAXID++;
    //cout << "\taddNode2PathVector Routing Node " << rNode->name << " assigned ID " << rNode->nodeID << endl;

    //Since we assigned this routing node an ID, we should add it to the cleanup list
    cleanupList->push_back(pExplore(rNode, 0, 0, 0));
  }

  //See if the pathvector can fit this node's ID
  if(rNode->nodeID >= pathVectorSize(pathVector)){
    //Increase the size of this vector to be able to handle the maximum node ID
    pathVector->resize(MAXIDpathVectorSize());
  }

  //Add this node into the pathvector
  setNodeID(pathVector, rNode->nodeID);
}

void fillPathVectorMask(){
  VECTORMASK = (unsigned int *) malloc(NUMVECTORBITS * sizeof(unsigned int));
  
  for(int i = 0; i < NUMVECTORBITS; i++){
    VECTORMASK[i] = 1 << i;
  }
}
