// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern int MAXLATENCY;
extern vprFPGA *FPGA;

void initializeAndCheckArch(Netlist *netlist, bool pipeline_aware){
  double averageIntrinsicDelay;

  //FPGA->print_info();

  //Calculate the average intrinsic delay of all routing nodes
  //and use this value to scale base cost values
  averageIntrinsicDelay = FPGA->calculateAverageIntrinsicDelay();
  
  cout << "Average Intrinsic Delay of routing nodes = " << averageIntrinsicDelay << endl;
  FPGA->scaleBaseCosts(averageIntrinsicDelay);
  
  //Count number of alu, multipliers, reg and rams in both the netlist and vpr architecture
  //to make sure that we have enough of everything
  cout << "Netlist has " << netlist->countLUT() << " LUTs" << endl;
  cout << "Netlist has " << netlist->countMult() << " Multipliers" << endl;
  cout << "Netlist has " << netlist->countFixedReg() << " Fixed Registers" << endl;
  cout << "Netlist has " << netlist->countRAM() << " RAM Blocks" << endl;
  cout << "Netlist has " << netlist->get_numInputs() << " inputs" << endl;
  cout << "Netlist has " << netlist->get_numOutputs() << " outputs" << endl;

  //Set up a bunch of pipeline-aware data structures
  if(pipeline_aware){
    cout << "Netlist requires at least " << netlist->countDelayReg() << " Delay Registers" << endl;
    fillPathVectorMask();
  }
  
  cout << "Max latency of any sink " << MAXLATENCY << endl;
  
  if(netlist->countLUT() > FPGA->countLUT())
    ERROR("Architecture does not have enough LUTs");
  else if(netlist->countMult() > FPGA->countMult())
    ERROR("Architecture does not have enough multipliers");
  else if(netlist->countFixedReg() > FPGA->countRegBanks())
    ERROR("Architecture does not have enough registers");
  else if(netlist->countRAM() > FPGA->countRAM())
    ERROR("Architecture does not have enough RAMs");
  else if(netlist->get_numInputs() + netlist->get_numOutputs() > FPGA->get_numIOs())
    ERROR("Architecture does not have enough input/output pins");

  if(pipeline_aware){
    if(netlist->countFixedReg() + netlist->countDelayReg() > FPGA->countDelayReg())
      ERROR("Architecture does not have enough registers to perform pipeline-aware placement/routing!");
  }
}
