// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern vprFPGA *FPGA;
//extern long int wireCost;
extern long int sharingCost;
extern bool lockRegPlacement;

void place(Netlist *netlist, bool pipeline_aware, bool timing_driven, FILE* fp){
  time_t start_time,end_time;

  fprintf(fp,"\nTotal num blocks (LOGIC + IO) = %d and total num nets = %d",
	  netlist->getNumBlocks(), netlist->get_numNets());
  fprintf(fp,"\nTarget is a %d x %d device.", FPGA->get_numRows(), FPGA->get_numCols());

  //See if we should be doing pipeline aware or Independence-style placement
  if(pipeline_aware){ 
    // start measuring the anneal runtime
    start_time=time(0);

    pipeline_placement(netlist, timing_driven);
	
    //printPlacement(netlist);
  }
  else{
    //independencePlacement(netlist, timing_driven, fp);
  }
  end_time=time(0);

  cout<<endl<<endl<<"TOTAL PLACEMENT RUNTIME IN SECONDS : "<<end_time-start_time << endl;
  fprintf(fp,"\n\nTOTAL ANNEAL RUNTIME IN SECONDS : %d\n",end_time-start_time);
}

void ltAnneal(Netlist *netlist, bool pipeline_aware, bool timing_driven){
  time_t start_time,end_time;

  //See if we should be doing pipeline aware or Independence-style placement
  if(pipeline_aware){ 
    // start measuring the anneal runtime
    start_time=time(0);

    pipelineLowTempAnneal(netlist, timing_driven);
  }
  else{
  }
  end_time=time(0);

  cout<<endl<<endl<<"Total low-temp anneal runtime in seconds : "<<end_time-start_time << endl;
}


// void independencePlacement(Netlist *netlist, bool timing_driven, FILE* fp){
//   double criticalPathDelay;
//   float annealTemp=0.;

//   wireCost = 0;
//   initPlacement(netlist);
  
// #if defined(PLACER_DEBUG)
//   printPlacement();
// #endif
  	
//   /* to facilitate A*, sort sinks in increasing order of separation from source */
//   netlist->sort_netSinksDistance();
	
//   /* Initial Pass...route each net in the netlist and calculate sharing and wire costs */
//   //Since we don't know the critical path delay yet, just fill in a really small number to prevent a divide by 0
//   routeNets(netlist, false, 1,true, timing_driven, DBL_MIN);

//   //Even if we are not in timing-driven mode we should still calculate timing numbers
//   criticalPathDelay = calculateCriticalPathDelay(netlist, false);
	
//   /* calculate starting temperature */
//   annealTemp = calcStartTemperature(netlist, timing_driven);

//   /* enter simulated annealing placement routine */
//   annealDesign(netlist, fp, timing_driven, annealTemp);
	
//   /* rip up all nets */
//   cout<<endl<<endl<<"RIPPING UP THE ENTIRE ROUTING....";
//   netlist->create_netListItr();
//   for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){
//     //if(net->sinkVector==0)
//     //continue;
//     net->rip_rTree();
//   }
//   netlist->remove_netListItr();
//   cout<<"DONE!! WIRE USAGE = "<<wireCost<<" AND SHARING = "<<sharingCost; 
// }  


// // This function will create a random initial placement of the netlist on the
// // FPGA architecture
// void initPlacement(Netlist *netlist){
//   int num_rows=FPGA->get_numRows();
//   int num_cols=FPGA->get_numCols();
//   int c,r;
//   int i;
//   vprIOB* io_block;vprLUT* lut;

//   netlist->create_lbListItr();
//   for(logicBlock* lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
//     while(1){
//       //First check to see if this is an input or output block
//       //Those need to be locked into place specially
//       if(lb->get_blockType() == INPUTPAD){
// 	io_block=(vprIOB*)(FPGA->get_blockAt(1,0));
// 	for(i=0;i<io_block->get_numIOs();i++){
// 	  if(io_block->getfrom_currBlockArray(i)==0){
// 	    /* found an empty IO location */
// 	    placeBlock(lb,(void *)(io_block), IOPIN,i);
// 	    break;
// 	  }
// 	}
// 	break;
//       }
//       else if(lb->get_blockType() == OUTPUTPAD){
// 	io_block=(vprIOB*)(FPGA->get_blockAt(1, num_cols+1));
// 	for(i=0;i<io_block->get_numIOs();i++){
// 	  if(io_block->getfrom_currBlockArray(i)==0){
// 	    /* found an empty IO location */
// 	    placeBlock(lb,(void *)(io_block), IOPIN,i);
// 	    break;
// 	  }
// 	}
// 	break;
//       }
//       else{
// 	//This is a logic block
//      r=lrand48()%num_rows;
// 	c=lrand48()%num_cols;

// 	lut=(vprLUT*)(FPGA->get_blockAt(r+1,c+1));
// 	if(lut->get_currBlock()){
// 	  /* a logic block has already been mapped to this location...try another coord */
// 	  continue;
// 	}
// 	//Make sure that the logic block type of lb matches that of the lut
// 	//that we're trying to map it to
// 	else if(lut->get_blockType() != lb->get_blockType()){
// 	  //The block type of the lb under question cannot be mapped to the lut we have selected
// 	  //cout << "LUT block type " << lut->get_blockType() << " does not match LB block type" << lb->get_blockType() << endl;
// 	  continue;
// 	}
// 	else{
// 	  /* found a location that this logic block can be placed on */
// 	  placeBlock(lb,(void *)(lut), LB,-1);
// 	  break;
// 	}
//       }
//     }
//   }
//   netlist->remove_lbListItr();

// }

// float calcStartTemperature(Netlist *netlist, bool timing_driven){
//   void* block0;void* block1;
//   vector<float> costArray(netlist->getNumBlocks());
//   float prev_wireCost=(float)(wireCost);
//   int index0,index1,block_type;

//   for(int i=0;i<netlist->getNumBlocks();i++){

//     /* randomly generate a pair of FPGA tiles that will be involved in the move */
//     generateMove(netlist, &block0,&index0,&block1,&index1,&block_type,(float)(FPGA->get_numRows()));

//     /* make the move i.e. actually move the logic/IO blocks from one location to the other */
//     makeMove(&block0,&index0,&block1,&index1,&block_type,1, timing_driven);

// #if defined(PLACER_DEBUG)
//     actual_wireCost=calcWireCost(netlist);
//     actual_sharingCost=FPGA->countSharing(0);
//     if(wireCost!=actual_wireCost)
//       ERROR("wire calculation mismatch...");
//     if(sharingCost!=actual_sharingCost)
//       ERROR("sharing calculation mismatch...");
// #endif

//     /* sort sinks for A* purposes */
//     sortSinks(&block0,&index0,&block1,&index1,&block_type);

//     costArray.at(i)=SHARE_FAC*(float)(sharingCost) + (float)(wireCost);

//   }

//   float avg_cost=0.;float cov=0.;
//   /* calculate the average cost */
//   for(int j=0;j<netlist->getNumBlocks();j++)
//     avg_cost+=costArray.at(j)/(float)(netlist->getNumBlocks());
//   /* calculate cov */
//   for(int k=0;k<netlist->getNumBlocks();k++)
//     cov+=(costArray.at(k) - avg_cost)*(costArray.at(k) - avg_cost);

//   return((float)(0.001 * pow((double)cov/(double)(netlist->getNumBlocks()),0.5)));
// }

// // This function generates an anneal move. When the function is done executing
// // block0 and block1 point to the 2 FPGA locations that will be involved in the
// // move. block0 always points to a location that holds a logic or IO block, while 
// // block1 may or may not be currently occupied. If the blocks involved are IOBs,
// // then index0 and index1 hold the indices of the specific IOs involved. block_type
// // indicates whether the involved blocks are LUTs or IOBs. 
// void generateMove(Netlist *netlist, void** block0,int* index0,void** block1,int* index1,int* block_type, float dLimit){
//   int num_rows=FPGA->get_numRows();
//   int num_cols=FPGA->get_numCols();
//   int num_els = netlist->getNumBlocks();

//   int col0,row0,col1,row1;
//   double selector;
//   logicBlock* lb1;
//   int dlim = (int)(floor((double)(dLimit)));

//   //if(dlim>FPGA->get_numRows() || dlim>FPGA->get_numCols())
//   //  dlim=FPGA->get_numRows();
//   //else if(dlim<1)
//   //  dlim=1;
  
//   //Remove windowing
//   dlim = FPGA->get_numCols();

//   logicBlock* lb0=netlist->get_logicBlock(lrand48()%num_els);
//   //This filters out everything but luts
//   while(movable_block(lb0)==false){
//     /* randomly select a logic block */
//     lb0 = netlist->get_logicBlock(lrand48()%num_els);
//   }
  
//   /* record its current coordinates in the vpr FPGA */
//   col0=lb0->col;row0=lb0->row;*index0=lb0->ptc_index;
  
// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"LOGIC BLOCK "<<lb0->get_name()<<" AT COL="<<col0<<" AND ROW="<<row0;
// #endif

//   /* assign the tile at these coordinates to block0 */
//   *block0=FPGA->get_blockAt(row0,col0);

//   //if(lb0->get_isLogic()!=true){
//   //  ERROR("Trying to move a input or output pin!");
//   //}

//   //else{
//     *block_type=1;

//     /* lb0 is a logic block...randomly select another logic block location */
//     while(1){
//       /* randomly select a logic block location in the FPGA */
//       col1=1 + lrand48()%num_cols;
//       row1=1 + lrand48()%num_rows;
//       *index1=-1;

//       /* check to see whether randomly generated location is the same as the other */ 
//       if((row0==row1 && col0==col1) || abs(row0-row1)>dlim || abs(col0-col1)>dlim)
//         continue;
//       //Check to see if the new location is of the same block type as the other location
//       else if(((vprLUT*)(*block0))->get_blockType() != ((vprLUT*)FPGA->get_blockAt(row1, col1))->get_blockType()){
// 	//cout << "Block type " << ((vprLUT*)(*block0))->get_blockType() << " != Block type " << ((vprLUT*)FPGA->get_blockAt(row1, col1))->get_blockType() << endl;
// 	continue;
//       }
//       else
//         break;

//     }
//     *block1=FPGA->get_blockAt(row1,col1);

// #if defined(PLACER_DEBUG)
//     if(lb1=((vprLUT*)(*block1))->get_currBlock()){
//       cout<<endl<<endl<<"LOGIC BLOCK "<<lb1->get_name()<<" AT COL="<<lb1->col<<" AND ROW="<<
//         lb1->row;
//     }
// #endif

//     //}
// }

// // This function checks to see if lb is a movable logic block. The
// // reason it is here is because there are 'global' nets that are 
// // sourced at input blocks that have no sinks (e.g. clock, reset etc)
// /* Modified to not allow the movement of input or output pins */
// bool movable_block(logicBlock* lb){

//   /* if we have an input block that sources a net with no sinks, */
//   /* then the block is not movable. */
//   /*   if(lb->get_isInput()){ */
//   /*     lb->create_outputListItr(); */
//   /*     net=lb->getnext_outputList()->get_net(); */
//   /*     lb->remove_outputListItr(); */
//   /*     if(net->sinkVector==0) */
//   /*       return false; */
//   /*   } */
//   if(lb->get_isInput() || lb->get_isOutput())
//     return false;
//   return true;
// }

// void makeMove(void** block0,int* index0,void** block1,int* index1,int* block_type,int iter, bool timing_driven){
//   Terminal* term;
//   logicBlock* lb0;logicBlock* lb1;

//   if(*block_type==0){
//     // We have an IOB
//     lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
//     lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
//   }
//   else{
//     // We have a LUT
//     lb0=((vprLUT*)(*block0))->get_currBlock();
//     lb1=((vprLUT*)(*block1))->get_currBlock();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"IN MAKE MOVE...";
// #endif

//   /* first, rip up the entire net that is sourced at a logic block. also, rip up the */
//   /* branch of each net that has a sink at the logic block. the occupancy of the nodes */  
//   /* in the ripped-up net segments are accordingly decremented. */
//   ripRoutes(lb0,lb1);

//   /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
//   if(*block_type==1){
//     //((vprLUT*)(*block0))->get_sourceNode()->clear_terminalList();
//     //((vprLUT*)(*block0))->get_sinkNode()->clear_terminalList();
//     //((vprLUT*)(*block1))->get_sourceNode()->clear_terminalList();
//     //((vprLUT*)(*block1))->get_sinkNode()->clear_terminalList();
//   }
//   else{
//     //((vprIOB*)(*block0))->getfrom_sourceArray(*index0)->clear_terminalList();
//     //((vprIOB*)(*block0))->getfrom_sinkArray(*index0)->clear_terminalList();
//     //((vprIOB*)(*block1))->getfrom_sourceArray(*index1)->clear_terminalList();
//     //((vprIOB*)(*block1))->getfrom_sinkArray(*index1)->clear_terminalList();
//   }

//   /* translate/swap lb0 and/or lb1 */
//   if(*block_type==0){
//     placeBlock(lb0,*block1, IOPIN,*index1);
//     if(lb1)
//       placeBlock(lb1,*block0, IOPIN,*index0);
//     else
//       ((vprIOB*)(*block0))->addto_currBlockArray(0,*index0);
//   }
//   else{
//     placeBlock(lb0,*block1, LB,*index1);
//     if(lb1)
//       placeBlock(lb1,*block0, LB,*index0);
//     else
//       ((vprLUT*)(*block0))->set_currBlock(0);
//   }

//   /* now, re-route the nets that connect to lb0 and lb1 */
//   if(lb0){
//     /* route the output nets of lb0 */
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
//       /* only consider those nets that have not already been ripped up */
//       routeNet(iter,term->get_net(),1,false, timing_driven, 0.0);
//       /* this net has now been counted */
//       term->get_net()->set_isCounted(true);
//     }
//     lb0->remove_outputListItr();
//   }

//   /* now, re-route the nets that connect to lb0 and lb1 */
//   if(lb1){
//     /* route the output nets of lb0 */
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
//       /* only consider those nets that have not already been ripped up */
//       routeNet(iter,term->get_net(),1,false, timing_driven, 0.0);
//       /* this net has now been counted */
//       term->get_net()->set_isCounted(true);
//     }
//     lb1->remove_outputListItr();
//   }

//   if(lb0){
//     /* now, route the input terminals of lb0 */ 
//     lb0->create_inputListItr();
//     for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false)
//         routeSink(iter,term->get_net(),term->get_rNode(),1, timing_driven);
//     }
//     lb0->remove_inputListItr();
//   }

//   if(lb1){
//     /* now, route the input terminals of lb0 */ 
//     lb1->create_inputListItr();
//     for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false)
//         routeSink(iter,term->get_net(),term->get_rNode(),1, timing_driven);
//     }
//     lb1->remove_inputListItr();
//   }

//   /* now that the nets have been ripped up, reset the output nets of lb0 and lb1 */
//   if(lb0){
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb0->remove_outputListItr();
//   }
//   if(lb1){
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb1->remove_outputListItr();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<"EXITING MAKE MOVE...";
// #endif
// }

// void undoMove(void** block0,int* index0,void** block1,int* index1,int* block_type,list<cachedObj*>* src_list0,
//               list<cachedObj*>* src_list1,list<cachedObj*>* sink_list0,
//               list<cachedObj*>* sink_list1,int iter, bool timing_driven){
//   logicBlock* lb0;logicBlock* lb1;
//   Terminal* term;
//   int connected=0;

//   if(*block_type==0){
//     // We have an IOB
//     lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
//     lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
//   }
//   else{
//     // We have a LUT
//     lb0=((vprLUT*)(*block0))->get_currBlock();
//     lb1=((vprLUT*)(*block1))->get_currBlock();
//   }

//   /* alright, rip up the source nets and sink segments that connect to lb0 and/or lb1 */
//   ripRoutes(lb0,lb1);

//   /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
//   if(*block_type==1){
//     //((vprLUT*)(*block0))->get_sourceNode()->clear_terminalList();
//     //((vprLUT*)(*block0))->get_sinkNode()->clear_terminalList();
//     //((vprLUT*)(*block1))->get_sourceNode()->clear_terminalList();
//     //((vprLUT*)(*block1))->get_sinkNode()->clear_terminalList();
//   }
//   else{
//     //((vprIOB*)(*block0))->getfrom_sourceArray(*index0)->clear_terminalList();
//     //((vprIOB*)(*block0))->getfrom_sinkArray(*index0)->clear_terminalList();
//     //((vprIOB*)(*block1))->getfrom_sourceArray(*index1)->clear_terminalList();
//     //((vprIOB*)(*block1))->getfrom_sinkArray(*index1)->clear_terminalList();
//   }

//   /* translate/swap lb0 and/or lb1 */
//   if(*block_type==0){
//     placeBlock(lb1,*block0, IOPIN,*index0);
//     if(lb0)
//       placeBlock(lb0,*block1, IOPIN,*index1);
//     else
//       ((vprIOB*)(*block1))->addto_currBlockArray(0,*index1);
//   }
//   else{
//     placeBlock(lb1,*block0, LB,*index0);
//     if(lb0)
//       placeBlock(lb0,*block1, LB,*index1);
//     else
//       ((vprLUT*)(*block1))->set_currBlock(0);
//   }

//   if(*block_type==0){
//     // We have an IOB
//     lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
//     lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
//   }
//   else{
//     // We have a LUT
//     lb0=((vprLUT*)(*block0))->get_currBlock();
//     lb1=((vprLUT*)(*block1))->get_currBlock();
//   }

//   /* now reconnect the previously cached nets and net segments to the terminals of lb0 */
//   /* and/or lb1. Route trees that were cached in cacheRoutes() will be used here. */

//   if(lb0){
//     /* loop thru the output terminals of lb0. for each output terminal, reconnect the */
//     /* cached net back to lb0. */
//     list<cachedObj*>::iterator list_itr=src_list0->begin();
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
//       if(term->get_net()->get_isCounted()==false){
//         /* only consider those nets that have not already been ripped up */

// 	if((*list_itr)->rNode!=term->get_rNode())
// 	  ERROR("terminal mismatch while undoing move...");
// 	/* repopulate the net's route tree with previously cached route tree. */
// 	(term->get_net()->get_rTree())->set_root(((*list_itr)->rTree).get_root());
// 	/* increment occupancies */
// 	(term->get_net()->get_rTree())->incrementOccupancies(0);
// 	term->get_net()->set_isCounted(true);
// 	/* increment the source list iterator */
// 	list_itr++;
 
//       }
//     }
//     lb0->remove_outputListItr();
//   }

//   if(lb1){
//     /* loop thru the output terminals of lb1. for each output terminal, reconnect the */
//     /* cached net back to lb1. */
//     list<cachedObj*>::iterator list_itr=src_list1->begin();
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
//       if(term->get_net()->get_isCounted()==false){
//         /* only consider those nets that have not already been ripped up */

// 	if((*list_itr)->rNode!=term->get_rNode())
// 	  ERROR("terminal mismatch while undoing move...");
// 	/* repopulate the net's route tree with previously cached route tree. */
// 	(term->get_net()->get_rTree())->set_root(((*list_itr)->rTree).get_root());
// 	/* increment occupancies. */
// 	(term->get_net()->get_rTree())->incrementOccupancies(0);
// 	term->get_net()->set_isCounted(true);
// 	/* increment the source list iterator */
// 	list_itr++;
 
//       }
//     }
//     lb1->remove_outputListItr();
//   }

//   if(lb0){
//     /* now, loop thru the input terminals of lb0. for each input terminal, reconnect */
//     /* the previously cached net segment back to the sink terminal of the logic block */
//     list<cachedObj*>::iterator list_itr=sink_list0->begin();
//     lb0->create_inputListItr();
//     for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false){
//         /* only consider those nets that have not already been ripped up */

// 	if((*list_itr)->rNode!=term->get_rNode())
// 	  ERROR("terminal mismatch while undoing move...");
// 	/* now reconnect the previously cached net segment to the sink terminal. */
// 	/* while doing so, increment occupancies. */
// 	connected=term->get_net()->reconnect(((*list_itr)->rTree).get_root());

//         /* if unable to connect, store the terminal and branch. the connection will be made later. */
//         if(connected==-1){
//           routeSink(iter,term->get_net(),term->get_rNode(),1, timing_driven);
//           (*list_itr)->clearTree=1;
//         }

// 	/* increment the source list iterator */
// 	list_itr++;

//       }
//     }
//     lb0->remove_inputListItr();
//   }

//   if(lb1){
//     /* now, loop thru the input terminals of lb1. for each input terminal, reconnect */
//     /* the previously cached net segment back to the sink terminal of the logic block */
//     list<cachedObj*>::iterator list_itr=sink_list1->begin();
//     lb1->create_inputListItr();
//     for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false){
//         /* only consider those nets that have not already been ripped up */

// 	if((*list_itr)->rNode!=term->get_rNode())
// 	  ERROR("terminal mismatch while undoing move...");
// 	/* now reconnect the previously cached net segment to the sink terminal. */
// 	/* while doing so, increment occupancies. */
// 	connected=term->get_net()->reconnect(((*list_itr)->rTree).get_root());

//         /* if unable to connect, store the terminal and branch. the connection will be made later. */
//         if(connected==-1){
//           routeSink(iter,term->get_net(),term->get_rNode(),1, timing_driven);
//           (*list_itr)->clearTree=1;
//         }

// 	/* increment the source list iterator */
// 	list_itr++;

//       }
//     }
//     lb1->remove_inputListItr();
//   }

//   /* reset the output nets of lb0 and lb1 */
//   if(lb0){
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb0->remove_outputListItr();
//   }
//   if(lb1){
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb1->remove_outputListItr();
//   }
// }


// // This function rips up the entire source net, and the branch of each net that sinks at
// // lb0 and/or lb1. Occupancies are appropriately decremented.
// void ripRoutes(logicBlock* lb0,logicBlock* lb1){
//   Terminal* term;

// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"IN RIP ROUTES...";
// #endif

//   if(lb0){
//     /* loop thru the output nets of lb0 and delete route trees */
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
//       /* only consider those nets that have not already been ripped up */
//       term->get_net()->rip_rTree();
//       /* this net has now been counted */
//       term->get_net()->set_isCounted(true);
//     }
//     lb0->remove_outputListItr();
//   }

//   if(lb1){
//     /* loop thru the output nets of lb1 and delete route trees */
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
//       /* only consider those nets that have not already been ripped up */
//       term->get_net()->rip_rTree();
//       /* this net has now been counted */
//       term->get_net()->set_isCounted(true);
//     }
//     lb1->remove_outputListItr();
//   }

//   if(lb0){
//     /* now, loop thru the input nets of lb0 and rip up the branch that connects */
//     /* lb0 to the route tree of each input net */
//     lb0->create_inputListItr();
//     for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false)
//         term->get_net()->rip_rTreeBranch(term->get_rNode());
//     }
//     lb0->remove_inputListItr();
//   }

//   if(lb1){
//     /* now, loop thru the input nets of lb1 and rip up the branch that connects */
//     /* lb1 to the route tree of each input net */
//     lb1->create_inputListItr();
//     for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false)
//         term->get_net()->rip_rTreeBranch(term->get_rNode());
//     }
//     lb1->remove_inputListItr();
//   }

//   /* now that the nets have been ripped up, reset the output nets of lb0 and lb1 */
//   if(lb0){
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb0->remove_outputListItr();
//   }
//   if(lb1){
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb1->remove_outputListItr();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"EXITING RIP ROUTES...";
// #endif

// }

//   This function places the logic block lb on FPGA block b. The parameter
//   t indicates whether this is a vprCLB or vprIOB. In case we have a vprIOB,
//   then index corresponds to the specific IO at which lb is placed. 
//   In the case we have a vprCLB, the index corresponds to the specific LUT we want 
//   to place the logic block into.
//   Note that we assume here that there is no other logic block already mapped
//   to the FPGA block b...the calling function MUST check occupancy.
void placeBlock(logicBlock* lb,void* b,enum e_timingMoveBlockType t, int index){
  vprLUT* lut;
  vprIOB* iob;
  routingNode* rnode;
  Terminal* term;

  if(lb==0 || b==0)
    ERROR("NULL parameters passed to function...");

  if(t == LB){
    /* We have a LUT */
    lut = ((vprCLB*)(b))->getLUT(index);

    if(lut->get_currBlock()){
      ERROR("Trying to place over an occupied LUT location!");
    }

    /*      if(lut->get_currBlock()) */
    /*        ERROR("hey, this LUT is already taken...");  */

    lut->set_currBlock(lb);
    lb->col = lut->get_col();
    lb->row = lut->get_row();
    lb->ptc_index = index;

    /* first map the input nets of lb to the sink node of lut */
    rnode = lut->get_sinkNode();
    lb->create_inputListItr();
    for(term = lb->getnext_inputList(); term != 0; term = lb->getnext_inputList()){
      //rnode->addto_terminalList(term);
      term->set_rNode(rnode);
    }
    lb->remove_inputListItr();

    /* now map the output nets of lb to the source node of lut */
    rnode=lut->get_sourceNode();
    lb->create_outputListItr();
    for(term = lb->getnext_outputList(); term != 0; term = lb->getnext_outputList()){
      //rnode->addto_terminalList(term);
      term->set_rNode(rnode);
    }
    lb->remove_outputListItr();

  }
  else if(t==IOPIN){
    /* We have an IOB */
    iob=(vprIOB*)(b);

    /*      if(iob->getfrom_currBlockArray(index)) */
    /*        ERROR("hey, this IOB is already taken..."); */

    if(iob->getfrom_currBlockArray(index)){
      ERROR("Trying to place over an occupied IOB location!");
    }

    iob->addto_currBlockArray(lb, index);
    lb->col=iob->get_col();
    lb->row=iob->get_row();
    lb->ptc_index=index;

    if(lb->get_blockType() == OUTPUTPAD){
      /* We have an output pad */
      rnode=iob->getfrom_sinkArray(index);
      lb->create_inputListItr();
      term=lb->getnext_inputList();
      //rnode->addto_terminalList(term);
      term->set_rNode(rnode);
      lb->remove_inputListItr();
    }
    else{
      /* We have an input pad */
      rnode=iob->getfrom_sourceArray(index);
      lb->create_outputListItr();
      term=lb->getnext_outputList();
      //rnode->addto_terminalList(term);
      term->set_rNode(rnode);
      lb->remove_outputListItr();
    }
  }
  else{
    ERROR("invalid FPGA block type...");
  }
}

//   This function unmaps the logic block lb from FPGA block b. The parameter
//   t indicates whether this is a vprLUT or vprIOB. In case we have a vprIOB,
//   then io_index corresponds to the specific IO at which lb is placed. 
//   Note that we assume here that there is no other logic block already mapped
//   to the FPGA block b...the calling function MUST check occupancy.
void unPlaceBlock(logicBlock* lb, void* b, enum e_timingMoveBlockType t, int index){
  vprLUT* lut;
  vprIOB* iob;
  Terminal* term;

  if(lb==0 || b==0)
    ERROR("NULL parameters passed to function...");

  if(t == LB){
    //We have a LUT
    lut = ((vprCLB*)(b))->getLUT(index);

    if(lb != lut->get_currBlock()){
      ERROR("unPlaceBlock: trying to unmap a different logic block!");
    }

    //First, unmap the input nets of the logic block from the sink node of the LUT
    //lut->get_sinkNode()->clear_terminalList();
    lb->create_inputListItr();
    for(term=lb->getnext_inputList();term!=0;term=lb->getnext_inputList()){
      term->clearRNode();
    }
    lb->remove_inputListItr();

    //Now, unmap the output nets of the logic block from the source node the LUT
    //lut->get_sourceNode()->clear_terminalList();
    lb->create_outputListItr();
    for(term=lb->getnext_outputList();term!=0;term=lb->getnext_outputList()){
      term->clearRNode();
    }
    lb->remove_outputListItr();

    lut->clearCurrBlock();
    lb->row = -1;
    lb->col = -1;
    lb->ptc_index = -1;
  }
  else if(t == IOPIN){
    /* We have an IOB */
    iob=(vprIOB*)(b);

    if(lb != iob->getfrom_currBlockArray(index)){
      ERROR("unPlaceBlock: trying to unmap a different I/O block!");
    }

    //If this block is an output pad
    if(lb->get_blockType() == OUTPUTPAD){
      //Unmap the input nets of the logic block from the sink node of the output pin
      //iob->getfrom_sinkArray(io_index)->clear_terminalList();
      lb->create_inputListItr();
      for(term=lb->getnext_inputList();term!=0;term=lb->getnext_inputList()){
	term->clearRNode();
      }
      lb->remove_inputListItr();
    }
    //If this block is an input pad
    else{
      //Unmap the output nets of the logic block from the source node the input pin
      //iob->getfrom_sourceArray(io_index)->clear_terminalList();
      lb->create_outputListItr();
      for(term=lb->getnext_outputList();term!=0;term=lb->getnext_outputList()){
	term->clearRNode();
      }
      lb->remove_outputListItr();    
    }

    iob->clearCurrBlockArray(index);
    lb->row = -1;
    lb->col = -1;
  }
  else{
    ERROR("invalid FPGA block type...");
  }
}

// void annealDesign(Netlist *netlist, FILE* fp, bool timing_driven, float annealTemp){
//   int bf_sharingCost,bf_wireCost,af_sharingCost,af_wireCost,prev_wireCost;
//   int iter=1;float deltaC;double i;
//   void* block0;void* block1;
//   int index0,index1;int block_type;
//   list<cachedObj*> sink_list0;list<cachedObj*> sink_list1;
//   list<cachedObj*> src_list0;list<cachedObj*> src_list1;
//   double num_acc=0.;double fracn_acc=1.;double alpha=0.;
//   float placementCost=0.;
//   time_t startTime,stopTime;

//   //Size of the move window
//   float dLimit=(float)(FPGA->get_numRows());

//   cout<<endl<<endl;
//   cout<<endl<<"temp.\t\tw_use\t\tshare\t\taccept\t\tdlim\t    cost\tALPHA";
//   cout<<endl<<"===================================================================================================";

//   fprintf(fp,"\n\n\ntemp.\t\tw_use\t\tshare\t\taccept\t\tdlim\t    cost\tALPHA");
//   fprintf(fp,"\n===================================================================================================");

//   while(fracn_acc > 0.05){
//     iter++;

//     FPGA->updateHistoryCosts(iter,true);

//     prev_wireCost=wireCost;

//     num_acc=0.;

//     for(i=0.;i<3.0*pow(netlist->getNumBlocks(),1.33);i+=1.0){

// #if defined(PLACER_DEBUG)
//       cout<<endl<<endl<<"********************** GENERATING NEW MOVE ************************";
// #endif


//       generateMove(netlist, &block0,&index0,&block1,&index1,&block_type, dLimit);


//       bf_sharingCost=sharingCost;bf_wireCost=wireCost;


//       cacheRoutes(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1);

 
//       makeMove(&block0,&index0,&block1,&index1,&block_type,iter, timing_driven);


//       af_sharingCost=sharingCost;af_wireCost=wireCost;

//       //I need to change this to make the placement timing driven!
//       deltaC=SHARE_FAC*(float)(af_sharingCost-bf_sharingCost)/(float)(prev_wireCost) + 
// 	(float)(af_wireCost-bf_wireCost)/(float)(prev_wireCost);

//       if(deltaC > 1.0e-10){
// 	// this is a bad move

//         if(drand48() < exp(-0.1*deltaC/annealTemp)){
//           // accept this move
// 	  sortSinks(&block0,&index0,&block1,&index1,&block_type);
//           clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);
//           num_acc+=1.0;

// #if defined(PLACER_DEBUG)
//           actual_wireCost=calcWireCost(netlist);
//           actual_sharingCost=FPGA->countSharing(0);
//           if(wireCost!=actual_wireCost)
//             ERROR("wire calculation mismatch...");
//           if(sharingCost!=actual_sharingCost)
//             ERROR("sharing calculation mismatch...");
// #endif

//         }
//         else{
//           // reject this move
//           undoMove(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1,iter, timing_driven);
//           clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,0);

// #if defined(PLACER_DEBUG)
//           actual_wireCost=calcWireCost(netlist);
//           actual_sharingCost=FPGA->countSharing(0);
//           if(wireCost!=actual_wireCost)
//             ERROR("wire calculation mismatch...");
//           if(sharingCost!=actual_sharingCost)
//             ERROR("sharing calculation mismatch...");
// #endif

//         }

//       }
//       else{
// 	// this is a good move...keep it
// 	sortSinks(&block0,&index0,&block1,&index1,&block_type);
// 	clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);

// 	if(deltaC < -1.0e-10)
// 	  num_acc+=1.0;
// 	/*  	num_acc+=1.0; */

// #if defined(PLACER_DEBUG)
//         actual_wireCost=calcWireCost(netlist);
//         actual_sharingCost=FPGA->countSharing(0);
//         if(wireCost!=actual_wireCost)
//           ERROR("wire calculation mismatch...");
//         if(sharingCost!=actual_sharingCost)
//           ERROR("sharing calculation mismatch...");
// #endif

//       }

//     }

//     stopTime=time(0);

//     fracn_acc=num_acc/i;
//     placementCost=SHARE_FAC*(float)(sharingCost) + (float)(wireCost);
//     printf("\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t%10.2f\t%4.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,ALPHA);
//     fprintf(fp,"\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t%10.2f\t%4.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,ALPHA);

//     fflush(stdout);
//     fflush(fp);


//     if(fracn_acc>0.96)
//       alpha=0.5;
//     else if(fracn_acc<=0.96 && fracn_acc>0.8)
//       alpha=0.9;
//     else if(fracn_acc<=0.8 && fracn_acc>0.15)
//       alpha=0.95;
//     else
//       alpha=0.8;
//     annealTemp=alpha*annealTemp;

//     dLimit = dLimit*(1-0.44+fracn_acc);
//     if(dLimit > FPGA->get_numRows())
//       dLimit = FPGA->get_numRows();
//     if(dLimit < 1)
//       dLimit=1;

//     routeNets(netlist, false, iter,true, timing_driven, DBL_MIN);
//   }

//   /* quench now */
//   for(int n=0;n<3;n++){

//     /* increment iteration */
//     iter++;
//     /* update history cost */
//     FPGA->updateHistoryCosts(iter,true);
//     /* init normalizing factor */
//     prev_wireCost=wireCost;
//     /* init num moves accepted to 0 */
//     num_acc=0.;

//     for(i=0.;i<5.0*pow(netlist->getNumBlocks(),1.33);i+=1.0){

// #if defined(PLACER_DEBUG)
//       cout<<endl<<endl<<"********************** GENERATING NEW MOVE ************************";
// #endif

//       /* randomly generate a pair of FPGA tiles that will be involved in the move */
//       generateMove(netlist, &block0,&index0,&block1,&index1,&block_type, dLimit);

//       /* cache sharing and wire costs before making the move */
//       bf_sharingCost=sharingCost;bf_wireCost=wireCost;

//       /* cache route trees & terminal mapping of the nets that connect to the moved blocks */
//       cacheRoutes(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1);

//       /* make the move i.e. actually move the logic/IO blocks from one location to the other */
//       makeMove(&block0,&index0,&block1,&index1,&block_type,iter, timing_driven);

//       /* cache sharing and wire costs after making the move */
//       af_sharingCost=sharingCost;af_wireCost=wireCost;

//       deltaC=SHARE_FAC*(float)(af_sharingCost-bf_sharingCost)/(float)(prev_wireCost) + 
// 	(float)(af_wireCost-bf_wireCost)/(float)(prev_wireCost);

//       if(deltaC > 1.0e-10){
// 	/* this is a bad move....reject */
// 	undoMove(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1,iter, timing_driven);
//         clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,0);

// #if defined(PLACER_DEBUG)
//         actual_wireCost=calcWireCost(netlist);
//         actual_sharingCost=FPGA->countSharing(0);
//         if(wireCost!=actual_wireCost)
//           ERROR("wire calculation mismatch...");
//         if(sharingCost!=actual_sharingCost)
//           ERROR("sharing calculation mismatch...");
// #endif

//       }
//       else{
// 	/* this is a good move...keep it */
// 	sortSinks(&block0,&index0,&block1,&index1,&block_type);
// 	clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);

// 	if(deltaC < -1.0e-10)
// 	  num_acc+=1.0;
// 	/*  	num_acc+=1.0; */

// #if defined(PLACER_DEBUG)
//         actual_wireCost=calcWireCost(netlist);
//         actual_sharingCost=FPGA->countSharing(0);
//         if(wireCost!=actual_wireCost)
//           ERROR("wire calculation mismatch...");
//         if(sharingCost!=actual_sharingCost)
//           ERROR("sharing calculation mismatch...");
// #endif

//       }

//     }

//     fracn_acc=num_acc/i;
//     placementCost=SHARE_FAC*(float)(sharingCost) + (float)(wireCost);
//     printf("\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t\t%10.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost);
//     fprintf(fp,"\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t\t%10.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost);

//     routeNets(netlist, false, iter,true, timing_driven, DBL_MIN);

//   }

// }

// void cacheRoutes(void** block0,int* index0,void** block1,int* index1,int* block_type,list<cachedObj*>* src_list0,
//                  list<cachedObj*>* src_list1,list<cachedObj*>* sink_list0,
//                  list<cachedObj*>* sink_list1){
//   logicBlock* lb0;logicBlock* lb1;
//   Terminal* term;
//   cachedObj* cache_obj;

//   if(*block_type==0){
//     // We have an IOB
//     lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
//     lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
//   }
//   else{
//     // We have a LUT
//     lb0=((vprLUT*)(*block0))->get_currBlock();
//     lb1=((vprLUT*)(*block1))->get_currBlock();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"IN CACHE ROUTES...";
// #endif

//   if(lb0){

// #if defined(PLACER_DEBUG)
//     cout<<endl<<"LOOKING AT "<<lb0->get_name();
// #endif

//     /* loop thru the output terminals of lb0. for each output terminal, store the routing */
//     /* node that the terminal maps to and the ENTIRE route tree of the net that it drives */
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){ 
//       /* create new cache object */
//       cache_obj=new cachedObj;
//       /* store routing node that the terminal maps to */
//       cache_obj->rNode=term->get_rNode();
//       /* copy ENTIRE route tree */
//       (cache_obj->rTree).populateWith(term->get_net()->get_rTree()->get_root());
//       /* add to list of cached objects */
//       src_list0->push_back(cache_obj);

// #if defined(PLACER_DEBUG)
//       (cache_obj->rTree).print_info();
// #endif

//       term->get_net()->set_isCounted(true);
//     }
//     lb0->remove_outputListItr();
//   }

//   if(lb1){

// #if defined(PLACER_DEBUG)
//     cout<<endl<<"LOOKING AT "<<lb1->get_name();
// #endif

//     /* loop thru the output terminals of lb1. for each output terminal, store the routing */
//     /* node that the terminal maps to and the ENTIRE route tree of the net that it drives */
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
//       /* create new cache object */
//       cache_obj=new cachedObj;
//       /* store routing node that the terminal maps to */
//       cache_obj->rNode=term->get_rNode();
//       /* copy ENTIRE route tree */
//       (cache_obj->rTree).populateWith((term->get_net()->get_rTree())->get_root());
//       /* add to list of cached objects */
//       src_list1->push_back(cache_obj);

// #if defined(PLACER_DEBUG)
//       (cache_obj->rTree).print_info();
// #endif

//       term->get_net()->set_isCounted(true);
//     }
//     lb1->remove_outputListItr();
//   }

//   if(lb0){

// #if defined(PLACER_DEBUG)
//     cout<<endl<<"LOOKING AT "<<lb0->get_name();
// #endif

//     /* now, loop thru the input terminals of lb0. for each input terminal, store the */
//     /* routing node that it maps to and the segment of the net's route tree that sinks */
//     /* at this input terminal */
//     lb0->create_inputListItr();
//     for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false){
//         /* only consider those nets that have not already been ripped up */

// 	/* create new cache object */
// 	cache_obj=new cachedObj;
// 	/* store routing node that the terminal maps to */
// 	cache_obj->rNode=term->get_rNode();
// 	/* get the sub-tree that sinks at this terminal and store a copy */
// 	(cache_obj->rTree).populateWith(
// 					term->get_net()->get_rTree()->get_subTree(term->get_rNode()),term->get_rNode());

// #if defined(PLACER_DEBUG)
// 	(cache_obj->rTree).print_info();
// #endif

// 	/* add to list of cached objects */
// 	sink_list0->push_back(cache_obj);

//       }
//     }
//     lb0->remove_inputListItr();
//   }

//   if(lb1){

// #if defined(PLACER_DEBUG)
//     cout<<endl<<"LOOKING AT "<<lb1->get_name();
// #endif

//     /* now, loop thru the input terminals of lb1. for each input terminal, store the */
//     /* routing node that it maps to and the segment of the net's route tree that sinks */
//     /* at this input terminal */
//     lb1->create_inputListItr();
//     for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
//       if(term->get_net()->get_isCounted()==false){
//         /* only consider those nets that have not already been ripped up */

// 	/* create new cache object */
// 	cache_obj=new cachedObj;
// 	/* store routing node that the terminal maps to */
// 	cache_obj->rNode=term->get_rNode();
// 	/* get the sub-tree that sinks at this terminal and store a copy */
// 	(cache_obj->rTree).populateWith(
// 					term->get_net()->get_rTree()->get_subTree(term->get_rNode()),term->get_rNode());

// #if defined(PLACER_DEBUG)
// 	(cache_obj->rTree).print_info();
// #endif

// 	/* add to list of cached objects */
// 	sink_list1->push_back(cache_obj);

//       }
//     }
//     lb1->remove_inputListItr();
//   }

//   /* reset the output nets of lb0 and lb1 */
//   if(lb0){
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb0->remove_outputListItr();
//   }
//   if(lb1){
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
//       term->get_net()->set_isCounted(false);
//     lb1->remove_outputListItr();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<"EXITING CACHE ROUTES...";
// #endif

// }

// void sortSinks(void** block0,int* index0,void** block1,int* index1,int* block_type){
//   logicBlock* lb0;logicBlock* lb1;
//   Terminal* term;

//   if(*block_type==0){
//     // We have an IOB
//     lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
//     lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
//   }
//   else{
//     // We have a LUT
//     lb0=((vprLUT*)(*block0))->get_currBlock();
//     lb1=((vprLUT*)(*block1))->get_currBlock();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"IN SORT SINKS...";
// #endif

//   /* now, sort the nets that connect to lb0 and lb1 */
//   if(lb0){
//     lb0->create_outputListItr();
//     for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
//       term->get_net()->sort_sinkVectorDistance();
//     lb0->remove_outputListItr();

//     lb0->create_inputListItr();
//     for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList())
//       term->get_net()->sort_sinkVectorDistance();
//     lb0->remove_inputListItr();
//   }

//   if(lb1){
//     lb1->create_outputListItr();
//     for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
//       term->get_net()->sort_sinkVectorDistance();
//     lb1->remove_outputListItr();

//     lb1->create_inputListItr();
//     for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList())
//       term->get_net()->sort_sinkVectorDistance();
//     lb1->remove_inputListItr();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<"EXITING SORT SINKS...";
// #endif
// }

// void clearCachedRoutes(list<cachedObj*>* src_list0,list<cachedObj*>* src_list1,
// 		       list<cachedObj*>* sink_list0,list<cachedObj*>* sink_list1,int clear_tree){
//   cachedObj* ptr;

//   list<cachedObj*>::iterator list_itr=src_list0->begin();
//   for(;list_itr!=src_list0->end();list_itr++){
//     if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
//       ((*list_itr)->rTree).clear();
//     ptr=*list_itr;
//     delete ptr;
//   }
//   src_list0->clear();

//   list_itr=src_list1->begin();
//   for(;list_itr!=src_list1->end();list_itr++){
//     if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
//       ((*list_itr)->rTree).clear();
//     ptr=*list_itr;
//     delete ptr;
//   }
//   src_list1->clear();

//   list_itr=sink_list0->begin();
//   for(;list_itr!=sink_list0->end();list_itr++){
//     if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
//       ((*list_itr)->rTree).clear();
//     ptr=*list_itr;
//     delete ptr;
//   }
//   sink_list0->clear();

//   list_itr=sink_list1->begin();
//   for(;list_itr!=sink_list1->end();list_itr++){
//     if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
//       ((*list_itr)->rTree).clear();
//     ptr=*list_itr;
//     delete ptr;
//   }
//   sink_list1->clear();
// }

// //This does the initial placement for all IO pins
// //Notice, this is Rapid-specific!
// void placeIO(Netlist *netlist){
//   int num_rows=FPGA->get_numRows();
//   int num_cols=FPGA->get_numCols();
//   int c,r;
//   int i;
//   vprIOB* io_block;vprLUT* lut;

//   netlist->create_lbListItr();
//   for(logicBlock* lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
//     while(1){
//       //First check to see if this is an input or output block
//       //Those need to be locked into place specially
//       if(lb->get_blockType() == INPUTPAD){
// 	io_block=(vprIOB*)(FPGA->get_blockAt(1,0));
// 	for(i=0;i<io_block->get_numIOs();i++){
// 	  if(io_block->getfrom_currBlockArray(i)==0){
// 	    /* found an empty IO location */
// 	    //cout << "Placing LB " << lb->get_name() << endl;
// 	    //cout << "To " << i << endl;
// 	    placeBlock(lb,(void *)(io_block), IOPIN,i);
// 	    break;
// 	  }
// 	}
// 	break;
//       }
//       else if(lb->get_blockType() == OUTPUTPAD){
// 	io_block=(vprIOB*)(FPGA->get_blockAt(1, num_cols+1));
// 	for(i=0;i<io_block->get_numIOs();i++){
// 	  if(io_block->getfrom_currBlockArray(i)==0){
// 	    /* found an empty IO location */
// 	    //cout << "Placing LB " << lb->get_name() << endl;
// 	    //cout << "To " << i << endl;
// 	    placeBlock(lb,(void *)(io_block),IOPIN,i);
// 	    break;
// 	  }
// 	}
// 	break;
//       }
//       else{
// 	break;
//       }
//     }
//   }
//   netlist->remove_lbListItr();
// }

/***
 * This function prints the current placement of the netlist on the FPGA
 ***/
void printPlacement(Netlist *netlist){
  vprIOB* iob;
  vprLUT* lut;
  registerBank* rb;
  routingNode* rnode;
  Terminal* term;

  cout<<endl<<endl;
  cout<<endl<<"<<<<<<<<<<<<<<<<<<<<< BEGIN PRINT PLACEMENT >>>>>>>>>>>>>>>>>>>>>";

  netlist->create_lbListItr();
  for(logicBlock* lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    cout<<endl<<lb->get_name()<<" AT POSITION ["<<lb->col<<","<<lb->row<<"]"<<" IO_INDEX "<<
      lb->ptc_index << endl;

    /* CORRESPONDENCE CHECK */

    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      iob = FPGA->getIOBAt(lb->row, lb->col);
      if(lb != iob->getfrom_currBlockArray(lb->ptc_index))
	ERROR("correspondence violation in placement phase...");
    }
    else if(lb->get_blockType() == LUT){
      lut = FPGA->getCLBAt(lb->row, lb->col)->getLUT(lb->ptc_index);
      if(lb != lut->get_currBlock())
	ERROR("correspondence violation in placement phase...");
    }
    else{
      rb = FPGA->getCLBAt(lb->row, lb->col)->getRegBank(lb->ptc_index);
      if(lb != rb->get_currBlock())
	ERROR("correspondence violation in placement phase...");
    }

    /* print the nets that connect to the inputs of the logic block mapped to tile */

    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      rnode = iob->getfrom_sinkArray(lb->ptc_index);
    }
    else if(lb->get_blockType() == LUT){
      rnode = lut->get_sinkNode();
    }
    else if(lb->get_blockType() == REGISTER_BANK){
      rnode = rb->getFirstRoutingNode();
    }
    else{
      ERROR("Unknown logic block type!");
    }

    if(lb->get_blockType() == LUT || lb->get_blockType() == OUTPUTPAD || lb->get_blockType() == REGISTER_BANK){
      cout<<"SINK NODE IS "<<rnode->get_name()<<". NETS MAPPED TO SINK NODE -->"; 
      //rnode->create_terminalListItr();
      //       for(term=rnode->getnext_terminalList();term!=0;term=rnode->getnext_terminalList()){
      // 	/* CORRESPONDENCE CHECK */
      // 	if(term->get_rNode()!=rnode)
      // 	  ERROR("correspondence violation in placement phase...");
      // 	cout<<endl<<"\t"<<term->get_net()->get_name();
      //}
      //       rnode->remove_terminalListItr();
      cout << endl;
    }
    /* print the nets that connect to the outputs of the logic block mapped to tile */
    
    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      rnode = iob->getfrom_sourceArray(lb->ptc_index);
    }
    
    else if(lb->get_blockType() == LUT){
      rnode = lut->get_sourceNode();
    }
    else if(lb->get_blockType() == REGISTER_BANK){
      rnode = rb->getFirstRoutingNode();
    }
    else{
      ERROR("Unknown logic block type!");
    }    

    if(lb->get_blockType() == LUT || lb->get_blockType() == INPUTPAD || lb->get_blockType() == REGISTER_BANK){
      cout<<"SOURCE NODE IS "<<rnode->get_name()<<". NETS MAPPED TO SOURCE NODE -->"; 
      //       rnode->create_terminalListItr();
      //       for(term=rnode->getnext_terminalList();term!=0;term=rnode->getnext_terminalList()){
      // 	/* CORRESPONDENCE CHECK */
      // 	if(term->get_rNode()!=rnode)
      // 	  ERROR("correspondence violation in placement phase...");
      // 	cout<<endl<<"\t"<<term->get_net()->get_name();
      //       }
      //       rnode->remove_terminalListItr();
    }
    cout<<endl;
  }
  netlist->remove_lbListItr();
  
  netlist->create_netListItr();
  for(Net* net = netlist->getnext_netList(); net != 0; net = netlist->getnext_netList()){
    for(list<Terminal *>::iterator regIter = net->pipelineRegisterList.begin(); regIter != net->pipelineRegisterList.end(); regIter++){
      cout <<endl << (*regIter)->name << " AT DELAY BLOCK [";
      cout << (*regIter)->getRegisterBank()->getHostCLB()->get_col() << ", " 
	   << (*regIter)->getRegisterBank()->getHostCLB()->get_row() << "] bank " 
	   << (*regIter)->getRegisterBank()->getHostBankNum();
    }
  }
  netlist->remove_netListItr();
  
  cout<<endl<<"<<<<<<<<<<<<<<<<<<<<< END PRINT PLACEMENT >>>>>>>>>>>>>>>>>>>>>" <<endl;
}


void printVPRPlacement(Netlist *netlist, long int random_seed,char* rr_echo, char* arguments, FILE* out_fp){
  char filename [256];
  FILE* fp;
  list<Net*>::iterator netIter;
  int clockRow, clockCol, clockPTC;

  sprintf(filename,"%s_%s_%s_%ld.p",netlist->get_name().c_str(),rr_echo, arguments, random_seed);
  fp=fopen(filename,"w");

  fprintf(fp,"\n");
  fprintf(fp,"# Netlist file: %s  Architecture file: %s",netlist->get_name().c_str(), rr_echo);
  fprintf(fp,"\n# Array size: %d x %d logic blocks\n",FPGA->get_numRows(), FPGA->get_numCols());

  FPGA->lineup_IOBs();

  //Figure out if we need a clock
  if(netlist->needClock()){
    FPGA->placeClockPin(&clockRow, &clockCol, &clockPTC);
    fprintf(fp,"\nPLACEMENT: pclk\t\t%d\t%d\t%d", clockCol, clockRow, clockPTC);
  }


  netlist->create_lbListItr();
  for(logicBlock* lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    fprintf(fp,"\nPLACEMENT: %s\t\t%d\t%d",lb->get_name().c_str(),lb->col,lb->row);
    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      fprintf(fp,"\t%d",lb->ptc_index);
    }
    else if(lb->get_blockType() == REGISTER_BANK || lb->get_blockType() == LUT){
      fprintf(fp,"\t%d", lb->ptc_index);
    }
    else{
      ERROR("Unknown logic block type!");
    }
  }
  netlist->remove_lbListItr();

  //Now go through all of the nets and dump the current placement of all pipelining registers
  for(netIter = netlist->netList.begin(); netIter != netlist->netList.end(); netIter++){
    (*netIter)->printVPRPlacement(fp);
  }

  /*    cout<<endl; */
  fprintf(fp,"\n\n");
  //fprintf(out_fp,"\n\n");
  fclose(fp);
}

// void swapBlocks(void** block0, int* index0, void** block1, int* index1, int* block_type){
//   Terminal* term;
//   logicBlock* lb0;logicBlock* lb1;

//   if(*block_type==0){
//     // We have an IOB
//     lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
//     lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
//   }
//   else{
//     // We have a LUT
//     lb0=((vprLUT*)(*block0))->get_currBlock();
//     lb1=((vprLUT*)(*block1))->get_currBlock();
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<endl<<"IN MAKE MOVE...";
// #endif

//   /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
//   if(*block_type==1){
// //     ((vprLUT*)(*block0))->get_sourceNode()->clear_terminalList();
// //     ((vprLUT*)(*block0))->get_sinkNode()->clear_terminalList();
// //     ((vprLUT*)(*block1))->get_sourceNode()->clear_terminalList();
// //     ((vprLUT*)(*block1))->get_sinkNode()->clear_terminalList();
//   }
//   else{
// //     ((vprIOB*)(*block0))->getfrom_sourceArray(*index0)->clear_terminalList();
// //     ((vprIOB*)(*block0))->getfrom_sinkArray(*index0)->clear_terminalList();
// //     ((vprIOB*)(*block1))->getfrom_sourceArray(*index1)->clear_terminalList();
// //     ((vprIOB*)(*block1))->getfrom_sinkArray(*index1)->clear_terminalList();
//   }

//   /* translate/swap lb0 and/or lb1 */
//   if(*block_type==0){
//     placeBlock(lb0,*block1, IOPIN,*index1);
//     if(lb1)
//       placeBlock(lb1,*block0, IOPIN,*index0);
//     else
//       ((vprIOB*)(*block0))->addto_currBlockArray(0,*index0);
//   }
//   else{
//     placeBlock(lb0,*block1, LB,*index1);
//     if(lb1)
//       placeBlock(lb1,*block0, LB,*index0);
//     else
//       ((vprLUT*)(*block0))->set_currBlock(0);
//   }

// #if defined(PLACER_DEBUG)
//   cout<<endl<<"EXITING MAKE MOVE...";
// #endif
// }


void findCostAndCPofPlacement(Netlist *netlist){
  double criticalPathDelay;
  
  //Calculate the critical path for the netlist given the initial placement
  netlist->updatePlacementTiming();
  
  criticalPathDelay = calculateCriticalPathDelay(netlist, true);

  //Now calculate the cost of the placement
  netlist->calculatePlacementCost(true, criticalPathDelay);
  
  printf("Placement wire Cost = %f\n", netlist->wireCost);
  printf("Placement timing Cost = %e\n", netlist->timingCost);
  printf("Estimated critical path delay = %e\n", criticalPathDelay);
}
