// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
//extern long int wireCost;
extern long int sharingCost;

// Erase the child node pointed to by ch_itr
void treeNode::eraseChild(list<treeNode*>::iterator* ch_itr){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(ch_itr==0)
    ERROR("child node has a NULL iterator...");
  childList.erase(*ch_itr);
} 

//Go through child list and try to find this child
void treeNode::removeChild(treeNode* child){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");

  //Make sure that no one has an iterator out right now
  if(child_itr==0){
    childList.remove(child);
  }
  else{
    ERROR("Cannot remove child from child list - iterator is being used");
  }

}

// Get the iterator that points to the node in childList that holds child_node  
list<treeNode*>::iterator treeNode::get_childItr(treeNode* child_node){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(child_node==0)
    ERROR("cannot look for NULL node in childList...");

  list<treeNode*>::iterator ch_itr=childList.begin();
  for(;ch_itr!=childList.end();ch_itr++){
    if(*ch_itr==child_node){
      return ch_itr;
    }
  }

  ERROR("could not find tree node in childList...");
}

// Set the parent node of this tree node
void treeNode::set_parent(treeNode* t){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(t==0)
    ERROR("setting parent of tree node to NULL...");
  parent=t;
}

// Set the routing node that this tree node represents
void treeNode::set_rNode(routingNode* r){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(r==0)
    ERROR("tree node cannot hold a NULL routing node...");
  rNode=r;

  /* at this point, set the inrTree member of rNode to indicate that rNode has been */
  /* associated with a node in the route tree of the net being currently routed */
//   rNode->set_inrTree(true);

}

// Set the cost of the path from the root to this tree node
// void treeNode::set_pathCost(double p){
//   if(magicNum!=31111)
//     ERROR("trying to access already deleted node...");
//   if(p<0.)
//     ERROR("path cost to a tree node cannot be NULL...");
//   pathCost=p;
// }

// Add a tree node to childList
void treeNode::addto_childList(treeNode* t){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(t==0)
    ERROR("setting child of tree node to NULL...");
  childList.push_back(t);
}

// Create a new childList iterator
void treeNode::create_childListItr(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(child_itr)
    ERROR("child iterator already exists...");
  child_itr=new list<treeNode*>::iterator;
  *child_itr = childList.begin();
}

// Get the number of children of this tree node
int treeNode::get_numChildren(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return childList.size();
} 

// Remove childList iterator
void treeNode::remove_childListItr(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(child_itr==0)
    ERROR("trying to delete NULL child iterator...");
  delete child_itr;
  child_itr=0;
}

// Get parent of this tree node
treeNode* treeNode::get_parent(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return parent;
}

// Get the routing node that this tree node represents
routingNode* treeNode::get_rNode(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return rNode;
}

// Get the cost of the path from the root to this tree node
// double treeNode::get_pathCost(){
//   if(magicNum!=31111)
//     ERROR("trying to access already deleted node...");
//   return pathCost;
// }

// Get the next element in childList
treeNode* treeNode::getnext_childList(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(*child_itr == childList.end())
    return 0;
  else{
    treeNode* ret_ptr = **child_itr;
    (*child_itr)++;
    return ret_ptr;
  }
}

// Get the root of the subtree that sinks at sink_rnode
treeNode* routeTree::get_subTree(routingNode* sink_rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(sink_rnode==0)
    ERROR("sink routing node cannot be NULL...");

  /* first locate the tree node that points to sink_rnode */
  treeNode* sink_tree_node=findNode(root,sink_rnode);
  if(sink_tree_node==0)
    ERROR("could not find sink routing node in this tree...");

  /* now iteratively traverse the parent of tree nodes until the root of the subtree */
  /* that sinks at sink_rnode is reached. */
  treeNode* tree_node;
  for(tree_node=sink_tree_node;tree_node->get_numChildren()<=1 &&
      tree_node->get_rNode()->get_type()!=SOURCE;tree_node=tree_node->get_parent()){ 
    ;
  }

  return tree_node;

}

// Make this route tree a copy of the tree rooted at tree_root.
// Special case, because we want only the tree branch between tree_root and
// the tree node that points to sink_rnode
void routeTree::populateWith(treeNode* tree_root,routingNode* sink_rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(tree_root==0 || sink_rnode==0)
    ERROR("route tree's root cannot be NULL...");

  treeNode* tree_node=findNode(tree_root,sink_rnode);
  treeNode* tree_node_copy;treeNode* prev_tree_node=0;

  for(;tree_node->get_numChildren()<=1 && tree_node->get_rNode()->get_type()!=SOURCE;
      tree_node=tree_node->get_parent()){
    /* create a new tree node that will be a copy of tree node */
    tree_node_copy=new treeNode;
    tree_node_copy->set_rNode(tree_node->get_rNode());
    tree_node_copy->linkDelay = tree_node->linkDelay;

    if(prev_tree_node){
      prev_tree_node->set_parent(tree_node_copy);
      tree_node_copy->addto_childList(prev_tree_node);
    }

    prev_tree_node=tree_node_copy;
  }

  tree_node_copy=new treeNode;
  tree_node_copy->set_rNode(tree_node->get_rNode());
  prev_tree_node->set_parent(tree_node_copy);
  tree_node_copy->addto_childList(prev_tree_node);

  root=tree_node_copy;

}



// Make this route tree a copy of the tree rooted at tree_root.
void routeTree::populateWith(treeNode* tree_root){

//   cout<<endl<<endl<<"IN POPULATE WITH...";

  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(tree_root==0)
    ERROR("route tree's root cannot be NULL...");

  root=makeCopy(tree_root);

//   cout<<endl<<"EXITING POPULATE WITH...";

}

// Find tree node that points to routing node rnode
treeNode* routeTree::getNode(routingNode* rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(rnode==0)
    ERROR("trying to look for NULL routing node...");

  treeNode* ret_node=findNode(root,rnode);
  if(ret_node==0){

#if defined(PLACER_DEBUG)
    cout<<endl<<endl<<"COULD NOT FIND "<<rnode->name;
    print_info();
#endif

  }

  return ret_node;
}

// Recursively build a copy of this route tree
treeNode* routeTree::makeCopy(treeNode* r){
  treeNode* child;treeNode* child_copy;
  treeNode* tree_node=r;treeNode* tree_node_copy;

  /* create a new tree node that will be a copy of tree node */
  tree_node_copy=new treeNode;
  tree_node_copy->set_rNode(tree_node->get_rNode());
  tree_node_copy->linkDelay = tree_node->linkDelay;

//   cout<<endl<<"\tcreated "<<tree_node_copy->get_rNode()->get_name();

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    /* recursively create a copy of the subtree rooted at this node */
    child_copy=makeCopy(child);
    tree_node_copy->addto_childList(child_copy);
    child_copy->set_parent(tree_node_copy);
  }
  tree_node->remove_childListItr();

  /* return the newly created copy of tree node */
  return tree_node_copy;

}

// Clean the routing nodes in this tree. This is a post-BFS reinit operation.
void routeTree::clean_rNodes(treeNode* r){
  treeNode* tree_node=r; treeNode* child;
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");

  tree_node->get_rNode()->isExplored=false;
  tree_node->get_rNode()->inrTree=false;
  tree_node->get_rNode()->predecessor=0;
  tree_node->get_rNode()->pathCost=DBL_MAX;
  tree_node->get_rNode()->netDelay=DBL_MAX;
  tree_node->get_rNode()->isRouted=false;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    clean_rNodes(child);
  tree_node->remove_childListItr();

}

// Set the root of this tree
void routeTree::set_root(treeNode* rt){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
//   if(rt==0)
//     ERROR("route tree's root cannot be NULL...");
  root=rt;
}

// Get the root of this tree
treeNode* routeTree::get_root(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  return root;
}

// Print this tree and info relating to it 
void routeTree::print_info(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  //if(root==0)
  //ERROR("trying to print NULL tree here...");
  //cout<<endl<<"PRINT ROUTE TREE -->"<<endl;
  if(root){
    preorderTraverse(root,0);
  }
  cout<<endl;
}

// Clean up this routing tree
void routeTree::clear(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(root){
    postorderTraverse(root,1);
    root=0;
  }
}

// Increment the occupancies of the routing nodes that constitute this tree
void routeTree::incrementOccupancies(int how){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(how!=0 && how!=1)
    ERROR("spurious parameter value here...");

  if(how==0)
    postorderTraverse(root,3);
  else
    postorderTraverse(root,4);
}

// Traverse this tree preorder. The actions to be taken at each node are specified by do_what
void routeTree::preorderTraverse(treeNode* r,int do_what){
  treeNode* child;
  treeNode* tree_node=r;

  if(do_what==0){
    /* print this tree node */
    cout<<"\t\t"<<tree_node->get_rNode()->get_name();
    cout<< " in routing tree? " << tree_node->get_rNode()->inrTree;
    cout << " in PRoute Tree? ";
    for(vector <unsigned int>::iterator iter = tree_node->get_rNode()->inPRTree.begin(); iter != tree_node->get_rNode()->inPRTree.end(); iter++){
      printf("%x ", *iter);
    }
    cout << endl;
  }

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    preorderTraverse(child,do_what);
  tree_node->remove_childListItr();

}

// Traverse this tree postorder. The action to be taken at each node are specified by do_what
void routeTree::postorderTraverse(treeNode* r,int do_what){
  treeNode* child;
  treeNode* tree_node=r;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    if(do_what == 87 || do_what == 88 || do_what == 89){
      postorderTraverse(child,86);
    }
    else if(do_what == 234){
      postorderTraverse(child, 233);
    }
    else{
      postorderTraverse(child,do_what);
    }
  }
  tree_node->remove_childListItr();

  if(do_what==0)
    /* print this tree node */
    //cout<<endl<<tree_node->get_rNode()->get_name()<<"  "<<tree_node->get_pathCost();
    cout<<tree_node->get_rNode()->get_name()<<"  "<<tree_node->get_rNode()->get_pathCost() <<endl;
  else if(do_what==1)
    /* delete this tree node */
    delete tree_node;
  else if(do_what==2 || do_what == 93){
    /* change sharing and wire cost to refelct the deletion of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost-=1;
    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE){
      //wireCost-=1;
    }
//     wireCost-=1;

    /* decrement occupancy */
    //If we are doing type "93" that means that we should not decrement the occupancy of 
    //source nodes that are also delay nodes
    if(do_what == 2 || !((tree_node->get_rNode() == root->get_rNode()) && (tree_node->get_rNode()->get_type() == DELAY))){
      tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()-1);
#if defined(PLACER_DEBUG)
    cout<<endl<<"\tdecrement and delete "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy();
#endif
    }

    /* delete this tree node */
    delete tree_node;
  }
  else if(do_what==3){
    /* increment occupancy of all routing nodes*/
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
    /* change sharing and wire cost to reflect the addition of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity()){
      sharingCost+=1;
      cout << "Increasing sharing cost due to node " <<  tree_node->get_rNode()->name << endl;
    }
//     wireCost+=1;

    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE){
      //wireCost+=1;
    }
  }
  else if(do_what==4){
    /* increment occupancy of non source/sink routing nodes*/
    if(tree_node->get_rNode()->get_type()!=SOURCE && tree_node->get_rNode()->get_type()!=SINK)
      tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
  }

  else if(do_what==5)
    tree_node->get_rNode()->set_counted(false);

  //Added this in!
  //Decrement occupancy and delete node
  else if(do_what == 86 || do_what == 88){

    /* change sharing and wire cost to reflect the deletion of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost-=1;
    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE){
      //wireCost-=1;
    }
    /* decrement occupancy */
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()-1);

#if defined(PLACER_DEBUG)
    cout<<"\tdecrement and deleted "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy() << endl;;
#endif

    //Clean up hanging nodes!
    //Make sure that we remove this node from the parent's child list
    if(do_what == 88 && tree_node->parent){
      tree_node->parent->removeChild(tree_node);
    }

    /* delete this tree node */
    if(root == tree_node){
      root = 0;
    }
    delete tree_node;
  }
  
  //Don't decrement occupancy
  else if(do_what == 87 || do_what == 89){
#if defined(PLACER_DEBUG)
    cout<<"\tdeleted "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy() << endl;
#endif

    //Clean up hanging nodes!
    //Make sure that we remove this node from the parent's child list
    if(do_what == 89 && tree_node->parent){
      tree_node->parent->removeChild(tree_node);
    }

    /* delete this tree node */
    if(root == tree_node){
      root = 0;
    }
    delete tree_node;
  }

  //Increment occupancy
  else if(do_what == 233){
    /* increment occupancy of all routing nodes*/
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
    /* change sharing and wire cost to reflect the addition of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity()){
      sharingCost+=1;
      cout << "Increasing sharing cost due to node " << tree_node->get_rNode()->name << endl;
    }

    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE){
      //wireCost+=1;
    }
#if defined(PLACER_DEBUG)
    cout<<"\tincremented "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy() << endl;;
#endif
  }
}

// Find rnode in this tree and return the relevant tree node
treeNode* routeTree::findNode(treeNode* r,routingNode* rnode){
  treeNode* tree_node=r;treeNode* ret_node=0;
  treeNode* child;

//   cout<<endl<<"this is tree node "<<tree_node->get_rNode()->get_name();

  if(tree_node->get_rNode()==rnode){
    /* found rnode...return this node */
    return tree_node;
  }

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    ret_node=findNode(child,rnode);
    if(ret_node){
      tree_node->remove_childListItr();
      return ret_node;
    }
  }
  tree_node->remove_childListItr();

  /* rnode does not exist in any node in the sub-tree rooted at tree_node */
  return 0;

}


// Make a copy of this tree from the given tree node down
routeTree* routeTree::copyPortion(treeNode* r){
  routeTree* copy = new routeTree();

  copy->populateWith(r);
  return copy;
}

//Increment the new occupancy of this node and it's children
void routeTree::reCalculateOccupancy(treeNode *r){
  r->rNode->newOccupancy++;

  for(list<treeNode*>::iterator childIter = r->childList.begin(); childIter != r->childList.end(); childIter++){
    reCalculateOccupancy(*childIter);
  }
}

void routeTree::clearInPRTree(treeNode *r){
  for(vector <unsigned int>::iterator vectIter = r->rNode->inPRTree.begin(); vectIter != r->rNode->inPRTree.end(); vectIter++){
    (*vectIter) = 0;
  }

  r->create_childListItr();
  for(treeNode *child=r->getnext_childList(); child!=0; child=r->getnext_childList()){
    clearInPRTree(child);
  }
  r->remove_childListItr();
}

void routeTree::printPath(Terminal *sink){
  treeNode *tempTreeNode;

  //If there is a route tree, print out the sequence
  if(root){
    tempTreeNode = returnTreeNode(root, sink->rNode);
    
    while(tempTreeNode){
      cout << "Node " << tempTreeNode->rNode->name << endl;
      tempTreeNode = tempTreeNode->parent;
    }
  }
}

treeNode *routeTree::returnTreeNode(treeNode* tNode, routingNode *rNode){
  treeNode *tempTreeNode;

  if(tNode->rNode == rNode){
    return tNode;
  }
  else{
    for(list<treeNode*>::iterator iter = tNode->childList.begin(); iter != tNode->childList.end(); iter++){
      tempTreeNode = returnTreeNode(*iter, rNode);
      if(tempTreeNode){
	return(tempTreeNode); 
      }
    }
    return NULL;
  }
}
