// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern bool critical_path_headroom;
extern double critical_path_headroom_fac;
extern int pipelined_cost_model;

std::string itoa(const int& x){
  std::ostringstream o;
  if (!(o << x)) return "ERROR";
  return o.str();
}

/* return the larger of two floating point numbers */
double doubmax(double x,double y){
  double diff=x-y;
  if(diff < 0.0)
    return y;
  else
    return x;
}

/* return the smaller of two floating point numbers */
double doubmin(double x,double y){
  double diff=x-y;
  if(diff < 0.0)
    return x;
  else
    return y;
}

/* return the larger of two floating point numbers */
float flomax(float x, float y){
  float diff=x-y;
  if(diff < 0.0)
    return y;
  else
    return x;
}

/* return the smaller of two floating point numbers */
float flomin(float x, float y){
  float diff=x-y;
  if(diff < 0.0)
    return x;
  else
    return y;
}

//Return the minimum of two integers
int min(int x, int y){
  if(x < y){
    return x;
  }
  else{
    return y;
  }
}

//Return the maximum of two integers
int max(int x, int y){
  if(x < y){
    return y;
  }
  else{
    return x;
  }
}


int KFUNCTION(int atLatency){
  return 1;
}

double convertACrit(int critIndex){
  if(critIndex < 0 || critIndex >= NUMASSUMEDCRIT){
    ERROR("Invalid assumed criticality");
  }
  else if(critIndex == NUMASSUMEDCRIT - 1){
    return(0.99);
  }
  else{
    return((double) (critIndex + 1) / (double) NUMASSUMEDCRIT);
  }
}

double calculateNewPathCost(double delay, double criticalPathDelay, double congestionCost, double additionalCosts){
  double crit;

  if(pipelined_cost_model == 0){
    //pathfinder cost model
    crit = doubmin((delay / criticalPathDelay), 0.99);
    return((crit * delay) + ((1 - crit) * congestionCost) + additionalCosts);
  }
  else if(pipelined_cost_model == 1){
    //modified pathfinder cost model
    crit = doubmin((delay / criticalPathDelay), 0.99);
    crit = crit / (1 - crit);
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else if(pipelined_cost_model == 2){
    //linear cost model
    crit = delay / criticalPathDelay;
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else if(pipelined_cost_model == 3){
    //quadratic cost model
    crit = delay / criticalPathDelay;
    crit = crit * crit;
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else if(pipelined_cost_model == 4){
    //cubic cost model
    crit = delay / criticalPathDelay;
    crit = crit * crit * crit;
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else{
    ERROR("Unknown pipelined cost model!");
  }
}

double calculateNewPathCostAC(int ACIndex, double delay, double congestionCost, double additionalCosts){
  double crit;
  
  crit = convertACrit(ACIndex);

  if(pipelined_cost_model == 0){
    //pathfinder cost model
    return((crit * delay) + ((1 - crit) * congestionCost) + additionalCosts);
  }
  else if(pipelined_cost_model == 1){
    //modified pathfinder cost model
    crit = crit / (1 - crit);
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else if(pipelined_cost_model == 2){
    //linear cost model
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else if(pipelined_cost_model == 3){
    //quadratic cost model
    crit = crit * crit;
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else if(pipelined_cost_model == 4){
    //cubic cost model
    crit = crit * crit * crit;
    return((crit * delay) + congestionCost + additionalCosts);
  }
  else{
    ERROR("Unknown pipelined cost model!");
  }
}
