// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern vprFPGA *FPGA;
extern bool maxCutCost;

#ifdef PLACER_DEBUG
extern int numRegistersEstimatedFreed;
extern int numRegistersEstimatedNeeded;
#endif
extern bool lut_level_placement;

vprBaseBlock::vprBaseBlock(int r, int c){
  if(r<0 || c<0)
    ERROR("row/col coords and channel width cannot be -ve...");
  row=r;col=c;
  magicNum=21111;
}

vprBaseBlock::~vprBaseBlock(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;
}

int vprBaseBlock::get_row(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return row;
}

int vprBaseBlock::get_col(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return col;
}

void vprBaseBlock::print_info(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  cout<<endl<<endl<<"FPGA LOCATION AT x = "<<col<<" AND y = "<<row;
}

/******************************************************************/

vprLUT::vprLUT(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum=21112;
  sourceNode=sinkNode=clock_sinkNode=0;
  //ipin_itr=opin_itr=0;
  currBlock=0;
}

vprLUT::~vprLUT(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  else
    derived_magicNum=0;
}

void vprLUT::set_sourceNode(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("source routing node cannot be NULL...");
  sourceNode=r;
}

void vprLUT::set_sinkNode(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("sink routing node cannot be NULL...");
  sinkNode=r;
}

void vprLUT::set_currBlock(logicBlock* l){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  currBlock=l;
}

/*void vprLUT::addto_inputPinList(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("input routing node cannot be NULL...");
  inputPinList.push_back(r);
  }*/

/*void vprLUT::addto_outputPinList(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("output routing node cannot be NULL...");
  outputPinList.push_back(r);
  }*/

/*void vprLUT::create_inputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(ipin_itr)
    ERROR("an iterator already exists...");
  ipin_itr = new list<routingNode*>::iterator;
  *ipin_itr = inputPinList.begin();
  }*/

/*void vprLUT::create_outputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(opin_itr)
    ERROR("an iterator already exists...");
  opin_itr = new list<routingNode*>::iterator;
  *opin_itr = outputPinList.begin();
  }*/

/*void vprLUT::remove_inputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!ipin_itr)
    ERROR("there is no iterator to remove...");
  delete ipin_itr;
  ipin_itr = 0;
  }*/

/*void vprLUT::remove_outputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!opin_itr)
    ERROR("there is no iterator to remove...");
  delete opin_itr;
  opin_itr = 0;
  }*/

/*routingNode* vprLUT::getnext_inputPinList(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!ipin_itr)
    ERROR("NULL iterator here...");

  if(*ipin_itr == inputPinList.end())
    return 0;
  else{
    routingNode* ret_ptr = **ipin_itr;
    (*ipin_itr)++;
    return ret_ptr;
  }
  }*/

/*routingNode* vprLUT::getnext_outputPinList(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!opin_itr)
    ERROR("NULL iterator here...");

  if(*opin_itr == outputPinList.end())
    return 0;
  else{
    routingNode* ret_ptr = **opin_itr;
    (*opin_itr)++;
    return ret_ptr;
  }

  }*/

logicBlock* vprLUT::get_currBlock(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return currBlock;
}

routingNode* vprLUT::get_sourceNode(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return sourceNode;
}

routingNode* vprLUT::get_sinkNode(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return sinkNode;
}

void vprLUT::set_blockType(enum e_genericBlockType t){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(t != LUT)
    ERROR("Invalid block type");
  blockType = t;
}

enum e_genericBlockType vprLUT::get_blockType(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return blockType;
}

void vprLUT::print_info(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");

  vprBaseBlock::print_info();

  if(currBlock)
    cout<<endl<<"LOGIC BLOCK : "<<currBlock->get_name();
  else
    cout<<endl<<"LOGIC BLOCK : ----";

  cout<<endl<<"SOURCE NODE : "<<sourceNode->get_name()<<", SINK NODE : "<<sinkNode->get_name();

  /*list<routingNode*>::iterator input_itr = inputPinList.begin();
  cout<<endl<<"INPUT ROUTING NODES  :";
  for(;input_itr!=inputPinList.end();input_itr++)
  cout<<"  "<<(*input_itr)->get_name();*/

  /*list<routingNode*>::iterator output_itr = outputPinList.begin();
  cout<<endl<<"OUTPUT ROUTING NODES  :";
  for(;output_itr!=outputPinList.end();output_itr++)
  cout<<"  "<<(*output_itr)->get_name();*/

}

void vprLUT::clearCurrBlock(){
  currBlock=0;  
}

//Are the LUTs and registers in the delay block at the same location empty?
bool vprLUT::emptyCLB(){
  if(!FPGA->getCLBAt(row, col)->areLUTsEmpty() || !FPGA->getCLBAt(row, col)->areRegistersEmpty())
    return false;
  return true;
}

//Are all of the LUTs in thei CLB empty?
bool vprLUT::emptyCLBLUTS(){
  if(!FPGA->getCLBAt(row, col)->areLUTsEmpty())
    return false;
  return true;
}

//We will swap the mappings of the two LUTs.
//If we are doing lut-level placement, this is all we will do.
//If we are not doing lut-level placement, we will also swap the
// mapping of any other LUTs in the CLB.
//myLB will point to the logic block currently mapped to this LUT.
//otherLB will point to the logic block currently mapped to the other LUT.
//Notice that if we are doing a move, myLB MUST point to something,
// by definition of use, but otherLB doesn't have to.
//Also notice, this relies on the netlist CLB packing, but that shouldn't matter
// since LUTs are always packed.
// void vprLUT::swapMapping(int myIndex, logicBlock* myLB, vprLUT* otherLUT, int otherIndex, logicBlock* otherLB){
//   c_clb* myCLB;
//   c_clb* otherCLB;

//   if(lut_level_placement){
//     //We are not maintaining the LUT packing, so just swap the two selected LUTs
//     unPlaceBlock(myLB, (void*) this, LB, myIndex);
//     if(otherLB){
//       unPlaceBlock(otherLB, (void *) otherLUT, LB, otherIndex);
//     }
//     placeBlock(myLB, (void *) otherLUT, LB, otherIndex);
//     if(otherLB){
//       placeBlock(otherLB, (void *) this, LB, myIndex);
//     }
//     myLB->updateSinkDelaysAndSlacks();
//     if(otherLB){
//       otherLB->updateSinkDelaysAndSlacks();
//     }
//   }
//   else{
//     //Regardless of whether or not we are doing clb-level placement,
//     // if we are maintaining the LUT packing we should move all LUTs in the same
//     // CLB together.
//     //Thus, this logic block will have an associated netlist parent CLB.
//     if(myLB){
//       myCLB = myLB->parentCLB;
//     }
//     else{
//       //Does this actually ever get used?
//       myCLB = returnMappedCLBFromLUTs();
//       ERROR("Why is this used?");
//     }
//     if(!myCLB){
//       ERROR("This should be mapped to something!");
//     }
    
//     //The other LUT might or might not be currently occupied.
//     if(otherLB){
//       otherCLB = otherLB->parentCLB;
//     }
//     //If not, we should check the other LUTs in the cluster.
//     //Just because this LUT is empty doesn't mean all the other ones are.
//     else{
//       otherCLB = otherLUT->returnMappedCLBFromLUTs();
//     }
    
//     //Remove the logic blocks from the old locations
//     myCLB->unPlaceLogicBlocksFromLUTs(this);
//     if(otherCLB){    
//       otherCLB->unPlaceLogicBlocksFromLUTs(otherLUT);
//     }
    
//     //Put them into the new locations
//     myCLB->placeLogicBlocksToLUTs(otherLUT);
//     if(otherCLB){
//       otherCLB->placeLogicBlocksToLUTs(this);
//     }
//     //Update the sink timings and slack for the nets we just moved
//     // The slack calculations are not exact since the critical path might have changed
//     myCLB->updateSinkDelaysAndSlacksForLUTs();
//     if(otherCLB)
//       otherCLB->updateSinkDelaysAndSlacksForLUTs();
//   }
// }
void vprLUT::swapMapping(int myIndex, logicBlock* myLB, vprLUT* otherLUT, int otherIndex, logicBlock* otherLB){
  c_clb* myCLB;
  c_clb* otherCLB;

  if(lut_level_placement){
    //We are not maintaining the LUT packing, so just swap the two selected LUTs
    unPlaceBlock(myLB, (void*) this, LB, myIndex);
    if(otherLB){
      unPlaceBlock(otherLB, (void *) otherLUT, LB, otherIndex);
    }
    placeBlock(myLB, (void *) otherLUT, LB, otherIndex);
    if(otherLB){
      placeBlock(otherLB, (void *) this, LB, myIndex);
    }
    myLB->updateSinkDelaysAndSlacks();
    if(otherLB){
      otherLB->updateSinkDelaysAndSlacks();
    }
  }
  else{
    //Regardless of whether or not we are doing clb-level placement,
    // if we are maintaining the LUT packing we should move all LUTs in the same
    // CLB together.
    //The CLB this LUT belongs to must have at least one occupied LUT
    myCLB = returnMappedCLBFromLUTs();
    if(!myCLB){
      ERROR("This should be mapped to something!");
    }
    
    //The other clb might or might not have an occupied LUT
    otherCLB = otherLUT->returnMappedCLBFromLUTs();
    
    //Remove the logic blocks from the old locations
    myCLB->unPlaceLogicBlocksFromLUTs(this);
    if(otherCLB){    
      otherCLB->unPlaceLogicBlocksFromLUTs(otherLUT);
    }
    
    //Put them into the new locations
    myCLB->placeLogicBlocksToLUTs(otherLUT);
    if(otherCLB){
      otherCLB->placeLogicBlocksToLUTs(this);
    }
    //Update the sink timings and slack for the nets we just moved
    // The slack calculations are not exact since the critical path might have changed
    myCLB->updateSinkDelaysAndSlacksForLUTs();
    if(otherCLB)
      otherCLB->updateSinkDelaysAndSlacksForLUTs();
  }
}

//Swap the entire contents of the two physical CLBs these luts belong to.
//Notice that since we are not doing strict CLB-level moves, we can't rely on the netlist clb information
// to find FFs.
//We have to find them manually.
void vprLUT::swapStickyMapping(vprLUT* otherLUT){
  c_clb* myCLB;
  c_clb* otherCLB;

  if(lut_level_placement){
    //We would like to be able to move a single LUT and any directly
    // associated FFs.
    //However, the problem is that we might have problems undoing the move.
    //For example, if this clb has a lut and a child FF, but the other
    // clb has a lone lut and is completely filled with unrelated FF
    // we will need to move one of those unrelated FFs to this clb to
    // make room.
    //However, if we do that, we won't know which FF to drag back
    // if we would like to undo the move.
    ERROR("We don't support lut-level placement with sticky LUTs yet!");
  }
  else{
    //Migrate all of the LUTs in this CLB to the other one.
    myCLB = returnMappedCLBFromLUTs();
    if(!myCLB){
      ERROR("This should be mapped to something!");
    }
    
    //The other clb might or might not have an occupied LUT
    otherCLB = otherLUT->returnMappedCLBFromLUTs();
    
    //Remove the luts from the old locations
    myCLB->unPlaceLogicBlocksFromLUTs(this);
    if(otherCLB){    
      otherCLB->unPlaceLogicBlocksFromLUTs(otherLUT);
    }
    
    //Put them into the new locations
    myCLB->placeLogicBlocksToLUTs(otherLUT);
    if(otherCLB){
      otherCLB->placeLogicBlocksToLUTs(this);
    }

    //Swap any FFs that might be mapped to the host physical clbs
    // for these two luts.
    FPGA->getCLBAt(get_row(), get_col())->swapAllFFs(FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col()));

    //Update the sink timings and slack for the nets we just moved
    // The slack calculations are not exact since the critical path might have changed
    myCLB->updateSinkDelaysAndSlacksForLUTs();
    if(otherCLB)
      otherCLB->updateSinkDelaysAndSlacksForLUTs();

    FPGA->getCLBAt(get_row(), get_col())->updateSinkDelaysAndSlacksForAllFFs();
    FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col())->updateSinkDelaysAndSlacksForAllFFs();
  }
}

//This will swap the contents of the two array CLBs that hold this LUT
// and the other LUT.
//Also notice, this relies on the netlist CLB packing, so if we are doing this on
// an "unpacked" netlist, the FFs in this physical CLB will not be moved since they
// are not in any netlist CLB.
// void vprLUT::swapEntireCLBMappings(logicBlock* myLB, vprLUT* otherLUT, logicBlock* otherLB){
//   c_clb* myCLB;
//   c_clb* otherCLB;

//   //Even though it was used to call this function, this LUT will not always be occupied by a logic block.
//   //The original call to this will, but when we undo the move, since we swapped the whole CLB, there is no guarantee that
//   // this LUT location will still be occupied.
//   //However, if possible, we should check since this is faster than looking through all of the items in the CLB
//   // to find something that is occupied.
//   if(myLB){
//     myCLB = myLB->parentCLB;
//   }
//   else{
//     myCLB = returnMappedCLBFromLUTs();
//   }
//   if(!myCLB){
//     ERROR("This should be mapped to something!");
//   }

//   //The other LUT might or might not be currently occupied.
//   if(otherLB){
//     otherCLB = otherLB->parentCLB;
//   }
//   //If not, we should check the other LUTs and FFs in the cluster.
//   //Just because this LUT is empty does not mean that the other LUTs are.
//   //Also, the other CLB could be inhabited by a netlist CLB that is all FFs.
//   else{
//     otherCLB = otherLUT->returnMappedCLBFromLUTsAndFF();
//   }

//   myCLB->unPlaceLogicBlocksFromLUTsAndFF(this);
//   if(otherCLB)
//     otherCLB->unPlaceLogicBlocksFromLUTsAndFF(otherLUT);

//   myCLB->placeLogicBlocksToLUTsAndFF(otherLUT);
//   if(otherCLB)
//     otherCLB->placeLogicBlocksToLUTsAndFF(this);

//   myCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
//   if(otherCLB)
//     otherCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
// }
void vprLUT::swapEntireCLBMappings(vprLUT* otherLUT){
  c_clb* myCLB;
  c_clb* otherCLB;

  //The clb this LUT belongs to MUST have something mapped to it.
  //At least one logic block LUT is mapped here.
  myCLB = returnMappedCLBFromLUTs();

  if(!myCLB){
    ERROR("This should be mapped to something!");
  }

  //The clb the other LUT belongs to might or might not be currently occupied.
  //We have to check both the LUTS and FFs just in case.
  otherCLB = otherLUT->returnMappedCLBFromLUTsAndFF();

  myCLB->unPlaceLogicBlocksFromLUTsAndFF(this);
  if(otherCLB)
    otherCLB->unPlaceLogicBlocksFromLUTsAndFF(otherLUT);

  myCLB->placeLogicBlocksToLUTsAndFF(otherLUT);
  if(otherCLB)
    otherCLB->placeLogicBlocksToLUTsAndFF(this);

  myCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
  if(otherCLB)
    otherCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
}

//We only moved the LUTs in the clbs associated with these two LUTs.
void vprLUT::updateCostAndDeltasForAllCLBLUTs(vprLUT* otherLUT, 
					     bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  c_clb* otherCLB;
  c_clb* myCLB;
  
  //Since we swapped them, the clb that the other lut belongs to
  // must have at least one lut mapped to it.
  otherCLB = otherLUT->returnMappedCLBFromLUTs();

  if(!otherCLB){
    ERROR("This should be mapped to something!");
  }
  
  //However, this one does not necessarily.
  myCLB = returnMappedCLBFromLUTs();

  otherCLB->updateCostsAndDeltasForLUTs(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  if(myCLB)
    myCLB->updateCostsAndDeltasForLUTs(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
}

//We only moved the LUTs in the clbs associated with these two LUTs.
void vprLUT::updateCostAndDeltasForAllCLBLUTsCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, 
							    double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  c_clb* otherCLB;
  c_clb* myCLB;
  
  //Since we swapped them, the clb that the other lut belongs to
  // must have at least one lut mapped to it.
  otherCLB = otherLUT->returnMappedCLBFromLUTs();

  if(!otherCLB){
    ERROR("This should be mapped to something!");
  }
  
  //However, this one does not necessarily.
  myCLB = returnMappedCLBFromLUTs();

  otherCLB->updateCostsAndDeltasForLUTsCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  if(myCLB)
    myCLB->updateCostsAndDeltasForLUTsCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
}

//We moved both the FFs and LUTs in the CLBs associated with these luts.
void vprLUT::updateCostAndDeltasForEntireCLB(vprLUT* otherLUT, 
					     bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  c_clb* otherCLB;
  c_clb* myCLB;

  //Since we swapped them, the other CLB must have at least one LUT that is occupied
  otherCLB = otherLUT->returnMappedCLBFromLUTs();

  if(!otherCLB){
    ERROR("This should be mapped to something!");
  }  

  //However, then other one doesn't necessarily.
  myCLB = returnMappedCLBFromLUTsAndFF();

  otherCLB->updateCostsAndDeltasForLUTsAndFF(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  if(myCLB)
    myCLB->updateCostsAndDeltasForLUTsAndFF(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
}

//We moved both the FFs and LUTs in the CLBs associated with these luts.
void vprLUT::updateCostAndDeltasForEntireCLBCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, 
							   double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  c_clb* otherCLB;
  c_clb* myCLB;

  //Since we swapped them, the other CLB must have at least one LUT that is occupied
  otherCLB = otherLUT->returnMappedCLBFromLUTs();

  if(!otherCLB){
    ERROR("This should be mapped to something!");
  }  

  //However, then other one doesn't necessarily.
  myCLB = returnMappedCLBFromLUTsAndFF();

  otherCLB->updateCostsAndDeltasForLUTsAndFFCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  if(myCLB)
    myCLB->updateCostsAndDeltasForLUTsAndFFCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
}


//Go through all of the register banks and LUTs associated with the vprCLB at this location.
//If there are any logic blocks mapped to these locations, return the CLB associated with them
c_clb* vprLUT::returnMappedCLBFromLUTsAndFF(){
  c_clb* mappedCLB = NULL;
  vprCLB *myCLB = FPGA->getCLBAt(row, col);

  //Try to extract a CLB from the FFs at this locations
  for(vector <registerBank*>::iterator rbIter = myCLB->registerBankVector.begin(); 
      rbIter != myCLB->registerBankVector.end(); rbIter++){
    if((*rbIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*rbIter)->get_currBlock()->parentCLB){
	  ERROR("Register bank is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*rbIter)->get_currBlock()->parentCLB;
      }
    }
  }
  
  //Try to extract a CLB from the LUTs at this locations
  for(vector <vprLUT *>::iterator lutIter = myCLB->lutVector.begin(); 
      lutIter != myCLB->lutVector.end(); lutIter++){
    if((*lutIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*lutIter)->get_currBlock()->parentCLB){
	  ERROR("LUT is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*lutIter)->get_currBlock()->parentCLB;
      }
    }
  }
  return mappedCLB;
}

//Go through all of the register banks and LUTs associated with the vprCLB at this location.
//If there are any logic blocks mapped to these locations, return the CLB associated with them
c_clb* vprLUT::returnMappedCLBFromLUTs(){
  c_clb* mappedCLB = NULL;
  vprCLB *myCLB = FPGA->getCLBAt(row, col);

  //Try to extract a CLB from the LUTs at this locations
  for(vector <vprLUT *>::iterator lutIter = myCLB->lutVector.begin(); 
      lutIter != myCLB->lutVector.end(); lutIter++){
    if((*lutIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*lutIter)->get_currBlock()->parentCLB){
	  ERROR("Register bank is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*lutIter)->get_currBlock()->parentCLB;
      }
    }
  }
  return mappedCLB;
}

/******************************************************************/

vprIOB::vprIOB(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum=21113;
  //sourceArray=sinkArray=inputPinArray=outputPinArray=0;
  sourceArray=sinkArray=0;
  currBlockArray=0;
  numIOs=0;
}

vprIOB::vprIOB(int r,int c,int n) : vprBaseBlock(r,c) {
  derived_magicNum=21113;
  numIOs=n;
  sourceArray=new vector<routingNode*>(n);
  sinkArray=new vector<routingNode*>(n);
  //inputPinArray=new vector<routingNode*>(n);
  //outputPinArray=new vector<routingNode*>(n);
  currBlockArray=new vector<logicBlock*>(n);

  for(int i=0;i<n;i++){
    sourceArray->at(i)=0;
    sinkArray->at(i)=0;
    //inputPinArray->at(i)=0;
    //outputPinArray->at(i)=0;
    currBlockArray->at(i)=0;
  }

}

vprIOB::~vprIOB(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  else
    derived_magicNum=0;
}

void vprIOB::addto_sourceArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  sourceArray->at(i)=r;
}

void vprIOB::addto_sinkArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  sinkArray->at(i)=r;
}

/*void vprIOB::addto_inputPinArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  inputPinArray->at(i)=r;
  }*/

/*void vprIOB::addto_outputPinArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  outputPinArray->at(i)=r;
  }*/

void vprIOB::addto_currBlockArray(logicBlock* l,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  currBlockArray->at(i)=l;
}

int vprIOB::get_numIOs(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  return numIOs;
}

routingNode* vprIOB::getfrom_sourceArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return sourceArray->at(i);
}

routingNode* vprIOB::getfrom_sinkArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return sinkArray->at(i);
}

/*routingNode* vprIOB::getfrom_inputPinArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return inputPinArray->at(i);
  }*/

/*routingNode* vprIOB::getfrom_outputPinArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return outputPinArray->at(i);
  }*/

logicBlock* vprIOB::getfrom_currBlockArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return currBlockArray->at(i);
}

void vprIOB::print_info(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");

  vprBaseBlock::print_info();

  /*for(int i=0;i<numIOs;i++){
    cout<<endl<<"SOURCE: "<<sourceArray->at(i)<<", OUTPUT: "<<outputPinArray->at(i)<<
      ", SINK: "<<sinkArray->at(i)<<", INPUT: "<<inputPinArray->at(i);
    if(currBlockArray->at(i))
      cout<<", IO BLOCK: "<<currBlockArray->at(i)->get_name();
    else
      cout<<", IO BLOCK: ----";
      }*/

}

void vprIOB::lineup_IOs(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");

  logicBlock* lb;routingNode* rnode;Terminal* term;int index=0;
  for(int i=0;i<numIOs;i++){
    if(lb=currBlockArray->at(i)){
      currBlockArray->at(index)=lb;
      lb->ptc_index=index;
      if(lb->get_blockType() == OUTPUTPAD){
	/* We have an output pad */
	rnode=sinkArray->at(index);
	lb->create_inputListItr();
	term=lb->getnext_inputList();
	//rnode->addto_terminalList(term);
	term->set_rNode(rnode);
	lb->remove_inputListItr();
      }
      else{
	/* We have an input pad */
	rnode=sourceArray->at(index);
	lb->create_outputListItr();
	term=lb->getnext_outputList();
	//rnode->addto_terminalList(term);
	term->set_rNode(rnode);
	lb->remove_outputListItr();
      }
      index++;
    }
  }

}

void vprIOB::clearCurrBlockArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  currBlockArray->at(i)=0;
}

void vprIOB::swapMapping(int myIndex, logicBlock* myLB, vprIOB* otherIOB, int otherIndex, logicBlock* otherLB){
  //Remove the io pins from the old locations
  unPlaceBlock(myLB, this, IOPIN, myIndex);
  if(otherLB)
    unPlaceBlock(otherLB, otherIOB, IOPIN, otherIndex);	
  
  //Put them in the new locations
  placeBlock(myLB, otherIOB, IOPIN, otherIndex);
  if(otherLB)
    placeBlock(otherLB, this, IOPIN, myIndex);

  //Update the sink timings and slack for the nets we just moved
  // The slack calculations are not exact since the critical path might have changed
  myLB->updateSinkDelaysAndSlacks();
  if(otherLB)
    otherLB->updateSinkDelaysAndSlacks();
}

/*(void vprIOB::connectSourceArrays(){

}

void vprIOB::connectInputPinArrays(){

}*/

/******************************************************************/

void vprFPGA::lineup_IOBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  for(int i=1;i<=numCols;i++){
    getIOBAt(0, i)->lineup_IOs();
    getIOBAt(numRows + 1, i)->lineup_IOs();
  }

  for(int j=1;j<=numRows;j++){
    getIOBAt(j, 0)->lineup_IOs();
    getIOBAt(j, numCols + 1)->lineup_IOs();
  }

}

// long int vprFPGA::countSharing(int opt){
//   if(magicNum!=23111)
//     ERROR("trying to double delete node here...");

//   int sharing=0;long int totalSharing=0;
//   list<routingNode*>::iterator rnode_itr = rnodeList.begin();
//   for(;rnode_itr!=rnodeList.end();rnode_itr++){
//     sharing=(*rnode_itr)->get_occupancy() - (*rnode_itr)->get_capacity();
//     if(sharing<0)
//       sharing=0;
//     totalSharing+=sharing;
//     if(opt==1)
//       cout<<endl<<(*rnode_itr)->get_name()<<"  "<<sharing;
//     else if(opt==2){
//       if(sharing>0)
// 	cout<<endl<<(*rnode_itr)->get_name()<<"  "<<(*rnode_itr)->get_occupancy();
//     }
//   }
//   return totalSharing;
// }

long int vprFPGA::countCongestedNodeOccupancy(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
}

void vprFPGA::updateHistoryCosts(int iter){
  list<routingNode*>::iterator rnodelist_itr;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  for(rnodelist_itr = rnodeList.begin(); rnodelist_itr != rnodeList.end(); rnodelist_itr++){
    (*rnodelist_itr)->update_historyCost(iter);
  }
}

void vprFPGA::print_info(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  list<routingNode*>::iterator rn_itr=rnodeList.begin();
  for(;rn_itr!=rnodeList.end();rn_itr++){
    if((*rn_itr)->type == BUS_OR_WIRE){
      //if(strncmp((*rn_itr)->wireName, "data_reg", 8) != 0){
      //if((*rn_itr)->wireName.compare(0, 8, "data_reg") != 0){
      //cout << (*rn_itr)->wireName << " delay " << (*rn_itr)->intrinsicDelay << endl;
      //}
      cout << "Wire node " << (*rn_itr)->name << " has delay " << (*rn_itr)->intrinsicDelay << endl;
    }
    //(*rn_itr)->print_info();
  }
  
  printf("IOB list has %d blocks\n", iobList.size());
  for(int i = 0; i < iobList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (iobList.at(i))->get_row(), (iobList.at(i))->get_col());
  }

  printf("LUT list has %d blocks\n", lutList.size());
  for(int i = 0; i < lutList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (lutList.at(i))->get_row(), (lutList.at(i))->get_col());
  }
  printf("MULT list has %d blocks\n", multList.size());
  for(int i = 0; i < multList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (multList.at(i))->get_row(), (multList.at(i))->get_col());
  }
  printf("RAM list has %d blocks\n", ramList.size());
  for(int i = 0; i < ramList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (ramList.at(i))->get_row(), (ramList.at(i))->get_col());
  }

  cout << "Array has " << registerBankVector.size() << " routing delay register banks" << endl;
  for(vector<registerBank*>::iterator itr = registerBankVector.begin(); itr != registerBankVector.end(); itr++){
    (*itr)->print_info();
  }
}

vprFPGA::~vprFPGA(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;

  //Delete the array
  
 //Make sure we are asking for a CLB location
  for(int row = 0; row <= numRows + 1; row++){
    for(int col = 0; col <= numCols + 1; col++){
      if(row < 1 || row > numRows || col < 1 || col > numCols){
	if(array.at(row).at(col)){
	  delete (vprIOB*) array.at(row).at(col);
	}
      }
      else{
	delete (vprCLB*) array.at(row).at(col);
      }
    }
  }

  //Delete all of the routing nodes
  for(list<routingNode*>::iterator rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    delete *rnodeIter;
  } 
}

vprFPGA::vprFPGA(int num_rows,int num_cols,int num_ios){
  int i, j;

  numRows = num_rows; 
  numCols = num_cols; 
  numIOs_perIOB = num_ios;
  numCLBs = numRows * numCols; 
  numIOs = 2 * numIOs_perIOB * (numRows + numCols);
  magicNum = 23111;
  numFreeRegisters = 0;
}

// int vprFPGA::get_numRows(){
//   if(magicNum!=23111)
//     ERROR("trying to double delete node here...");
//   return numRows;
// }

// int vprFPGA::get_numCols(){
//   if(magicNum!=23111)
//     ERROR("trying to double delete node here...");
//   return numCols;
// }

int vprFPGA::get_numCLBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numCLBs;
}

int vprFPGA::get_numIOs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numIOs;
}

int vprFPGA::countLUT(){
  int r, c, count = 0;
  int i;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      for(i = 0; i < clb->getNumLUTs(); i++){
	lut = clb->getLUT(i);
	if(lut->get_blockType() == LUT){
	  count ++;
	}
      }
    }
  }
  return (count);
}

int vprFPGA::countMult(){
  int r, c, count = 0;
  int i;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      for(i = 0; i < clb->getNumLUTs(); i++){
	lut = clb->getLUT(i);
	if(lut->get_blockType() == MULTIPLIER){
	  count ++;
	}
      }
    }
  }
  return (count);
}

int vprFPGA::countRegBanks(){
  int r, c, count = 0;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      count += getCLBAt(r + 1, c + 1)->getNumRegBanks();
    }
  }
  return (count);
}

// int vprFPGA::countRegTotal(){
//   int count = 0;
//   vector<vector <delayBlock *> >::iterator outerIter;
//   vector <delayBlock *>::iterator innerIter;
//   vector <registerBank *>::iterator bankIter;

//   if(magicNum!=23111)
//     ERROR("trying to double delete node here...");

//   for(outerIter = delayBlockArray.begin(); outerIter != delayBlockArray.end(); outerIter++){
//     for(innerIter = (*outerIter).begin(); innerIter != (*outerIter).end(); innerIter++){
//       for(bankIter = (*innerIter)->registerBankVector.begin(); bankIter != (*innerIter)->registerBankVector.end(); bankIter++){
// 	count += (*bankIter)->getNumPossibleDelay();
//       }
//     }
//   }
//   return (count);
// }

int vprFPGA::countRAM(){
  int r, c, count = 0;
  int i;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      for(i = 0; i < clb->getNumLUTs(); i++){
	lut = clb->getLUT(i);
	if(lut->get_blockType() == RAM_BLOCK){
	  count ++;
	}
      }
    }
  }
  return (count);
}

void vprFPGA::addto_rnodeList(routingNode* r){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("trying to add NULL rnode...");
  rnodeList.push_back(r);
}

void vprFPGA::create_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(rnode_itr)
    ERROR("an iterator already exists...");
  rnode_itr=new list<routingNode*>::iterator;
  *rnode_itr=rnodeList.begin();
}

void vprFPGA::remove_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("there is no iterator to remove...");
  delete rnode_itr;
  rnode_itr=0;
}

routingNode* vprFPGA::getnext_rnodeList(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("trying to access NULL iterator here...");

  if(*rnode_itr == rnodeList.end())
    return 0;
  else{
    routingNode* ret_ptr = **rnode_itr;
    (*rnode_itr)++;
    return ret_ptr;
  }

}

//Go through all routing nodes in the architecure and determine the
//average intrinsic delay
double vprFPGA::calculateAverageIntrinsicDelay(){
  long numNodes;
  double intrinsicDelaySum;
  list<routingNode*>::iterator rnodeIter;

  numNodes = 0;
  intrinsicDelaySum = 0.0;
  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    intrinsicDelaySum += (*rnodeIter)->intrinsicDelay;
    numNodes++;
  }

  return (intrinsicDelaySum / (double) numNodes);
}

// Go through all routing nodes in the architecture and scale their base cost
// by the average intrinsic delay of all nodes
// Thus, since the base cost of a input pin is normally set to 0.95 and a output pin
// has a base cost of 1.0, all nodes will maintain their relative proportion to each other
void vprFPGA::scaleBaseCosts(double averageIntrinsicDelay){
  list<routingNode*>::iterator rnodeIter;

  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    (*rnodeIter)->baseCost = (*rnodeIter)->baseCost * averageIntrinsicDelay;
  }
}

//Initialize some of the placement timing data structures
void vprFPGA::initPlacementDS(){
  // First, build lists of various block types
  // This is relatively Rapid-specific
  create_block_type_lists();
  
  // Now create the timing table we'll use to estimate delay during placement.
  // We do this even if we are not doing timing-driven placement so that
  //  we can give an estimate of the critical path delay.
  createTimingTable();
      
  if(maxCutCost){
    createCutCostVector();
  }

  //print_info();
}

//Count the number of delay nodes in the architecture
int vprFPGA::countDelayReg(){
  list<routingNode*>::iterator rnodeIter;
  int count = 0;

  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    if((*rnodeIter)->type == DELAY){
      count++;
    }
  }

  numFreeRegisters = count;
  return (count);
}

//Go through and make all the occupancy of all of the routing nodes 0
void vprFPGA::zeroOccupancy(){
  list<routingNode*>::iterator rnodeIter;

  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    (*rnodeIter)->newOccupancy = 0;
  }
}

long int vprFPGA::countNewSharing(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  int sharing=0;long int totalSharing=0;
  list<routingNode*>::iterator rnode_itr;

  for(rnode_itr = rnodeList.begin(); rnode_itr != rnodeList.end(); rnode_itr++){
    if((*rnode_itr)->newOccupancy != (*rnode_itr)->occupancy){
      cout << "Node " << (*rnode_itr)->name << " has old occupancy " << (*rnode_itr)->occupancy << " and new occupancy " << (*rnode_itr)->newOccupancy << endl;
    }

    sharing=(*rnode_itr)->newOccupancy - (*rnode_itr)->get_capacity();
    if(sharing <= 0){
      sharing=0;
    }
    else{
      cout << "Node " << (*rnode_itr)->name << " is shared by " << sharing << endl;
    }
    totalSharing+=sharing;

    if(opt==1)
      cout<<endl<<(*rnode_itr)->get_name()<<"  "<<sharing;
    else if(opt==2){
      if(sharing>0)
	cout<<endl<<(*rnode_itr)->get_name()<<"  "<<(*rnode_itr)->get_occupancy();
    }
  }

  return totalSharing;

}

// The block-type lists let you know where the 5th ALU or 2nd Multiplier is in the system
void vprFPGA::create_block_type_lists(){
  vprCLB* clb;
  vector<vprLUT *>::iterator lutIter;

  for(int r = 0; r < get_numRows(); r++){
    for(int c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      
      //Put all of the LUTs into the list
      for(lutIter = clb->lutVector.begin(); lutIter != clb->lutVector.end(); lutIter++){
	if((*lutIter)->get_blockType() == LUT)
	  lutList.push_back(*lutIter);
	else if((*lutIter)->get_blockType() == MULTIPLIER)
	  multList.push_back(*lutIter);
	else if((*lutIter)->get_blockType() == RAM_BLOCK)
	  ramList.push_back(*lutIter);
	else
	  ERROR("Unknown block type in array!");
      }
      
      // Go through all the registers blocks in the system and put
      //  each register bank into a list
      clb->addAllRegBanksToRegisterBankVector();
      clb->addAllRegBanksToFreeRegisterBankList();
    }
  }

  //Put all of the IO pins into the IOB list
  for(int r = 1; r <= get_numRows(); r++){
    //Left side
    iobList.push_back(getIOBAt(r, 0));
    //Right side
    iobList.push_back(getIOBAt(r, get_numCols()+ 1));
  }

  for(int c = 1; c <= get_numCols(); c++){
    //Top
    iobList.push_back(getIOBAt(0, c));
    //Bottom
    iobList.push_back(getIOBAt(get_numRows()+ 1, c));
  }
}

void vprFPGA::createTimingTable(){
  //Initialize the timing table itself
  initializeTimingTable();
  
  //Create a table to keep track of how many times we find a given type of sink
  // at a given distance - we need this to calculate average time
  initializeTimingNumHitTable();
 
  //Launch searches from all of the input pins
  sweepTimingIOPins();
  
  //These aren't necessary since we assume that all blocks have the same
  // connectivity regardless of type.
  //This is fine until we introduce interconnect registers.
  //Launch searches from all of the logic blocks
  sweepTimingLogic();

  //Launch searches from all of the delay registers
  sweepTimingDelay();

  //Calculate average time given how many times we found each kind of sink
  calculateTimingTable();

  emptyTimingNumHitTable();
}

void vprFPGA::createCutCostVector(){
  //The cut cost vector should have N-1 entries since there are N-1 gaps
  // Num Cols only accounts for logic blocks, but we have I/O pins too
  cutCostVector.resize(get_numCols() + 1, 0.0);
}

void vprFPGA::clearCutCostVector(){
  for(vector<double>::iterator iter = cutCostVector.begin(); iter != cutCostVector.end(); iter++){
    *iter = 0.0;
  }
}

//Make enough space for all the values we'll need in the timing table
//The first dimension is the number of different kinds of sources we have
//The second dimension is the number of different source banks we might have
//The third dimension is the number of different kinds of sinks we have
//The fourth dimension is the number of different sink banks we might have
//The fifth is the col separation between the source and the sink
//The sixth is the row separation between the source and the sink
void vprFPGA::initializeTimingTable(){
  int maxBank;

  //Go through and figure out what the maximum delay block bank is.
  //maxBank = 0;
//   for(vector<vector <delayBlock *> >::iterator outerIter = delayBlockArray.begin(); 
//       outerIter != delayBlockArray.end(); outerIter++){
//     for(vector <delayBlock *>::iterator innerIter = outerIter->begin(); 
// 	innerIter != outerIter->end(); innerIter++){
//       if(*innerIter){
// 	if(maxBank < (*innerIter)->getNumRegBanks()){
// 	  maxBank = (*innerIter)->getNumRegBanks();
// 	}
//       }
//     }
//   }
  //Right now, we assume that all banks have the same connectivity
  //This is fine until we introduce interconnect registers.
  maxBank = 1;

  timingTable = new vector <vector <vector <vector <vector <vector <double> > > > > >;
  
  //We have three kinds of sources: logic blocks, input pins and register banks
  timingTable->resize(3);
  ////Right now, we assume that all kinds of blocks have the same connectivity
  //timingTable->resize(1);
  for(vector <vector <vector <vector <vector <vector <double> > > > > >::iterator iter = 
	timingTable->begin(); iter != timingTable->end(); iter++){
    //Each of these can have up to N banks
    (*iter).resize(maxBank);
    for(vector <vector <vector <vector <vector <double> > > > >::iterator iter2 =
	  (*iter).begin(); iter2 != (*iter).end(); iter2++){
      //Each of these three sources can drive one of three kinds of sinks:
      //a logic block, an output pin or a delay block
      (*iter2).resize(3);
      ////Right now, we assume that all kinds of blocks have the same connectivity
      //(*iter2).resize(1);
      for(vector <vector <vector <vector <double> > > >::iterator iter3 = 
	    (*iter2).begin(); iter3 != (*iter2).end(); iter3++){
	//Each of these can have up to N banks
	(*iter3).resize(maxBank);
	for(vector <vector <vector <double> > >::iterator iter4 = 
	      (*iter3).begin(); iter4 != (*iter3).end(); iter4++){
	  //Each of these source-sink pairs can have an col separation of
	  // between negative (num columns + 1) (input pin on right
	  // drives output pin on left
	  // and (num columns + 1) (input pin to output pin).
	  //Thus, if we make each of the third dimension vectors
	  // of size (num columns + 1) * 2 we can accomodate everyone
	  (*iter4).resize(((numCols + 1) * 2) + 1);
	  for(vector <vector <double> >::iterator iter5 = 
		(*iter4).begin(); iter5 != (*iter4).end(); iter5++){
	    //Initialize these values to be really big
	    (*iter5).resize(((numRows + 1) * 2) + 1, 1.0E20);
	  }
	}
      }
    }
  }
}

//Make enough space for all the values we'll need in the num hit table
//Basically, do exactly the same thing as above, only for the number of times we hit each type of sink at a given distance
void vprFPGA::initializeTimingNumHitTable(){
  int maxBank;

  //Go through and figure out what the maximum delay block bank is.
  //This takes out the idea that register banks have different connectivity.
  //I'll have to put it back in when I put registers into the interconnect.
  maxBank = 1;
  // maxBank = 0;
  //   for(vector<vector <delayBlock *> >::iterator outerIter = delayBlockArray.begin(); 
  //       outerIter != delayBlockArray.end(); outerIter++){
  //     for(vector <delayBlock *>::iterator innerIter = outerIter->begin(); 
  // 	innerIter != outerIter->end(); innerIter++){
  //       if(*innerIter){
  // 	if(maxBank < (*innerIter)->getNumRegBanks()){
  // 	  maxBank = (*innerIter)->getNumRegBanks();
  // 	}
  //       }
  //     }
  //   }

  timingNumHitTable = new vector <vector <vector <vector <vector <vector <int> > > > > >;
  
  //We have three kinds of sources: logic blocks, input pins and delay blocks
  timingNumHitTable->resize(3);
  ////Right now, we assume that all kinds of blocks have the same connectivity
  //timingNumHitTable->resize(1);

  for(vector <vector <vector <vector <vector <vector <int> > > > > >::iterator iter = 
	timingNumHitTable->begin(); iter != timingNumHitTable->end(); iter++){
    //Each of these can have up to N banks
    (*iter).resize(maxBank);
    for(vector <vector <vector <vector <vector <int> > > > >::iterator iter2 =
	  (*iter).begin(); iter2 != (*iter).end(); iter2++){
      //Each of these three sources can drive one of three kinds of sinks:
      //a logic block, an output pin or a delay block
      (*iter2).resize(3);
      ////Right now, we assume that all kinds of blocks have the same connectivity
      //(*iter2).resize(1);
      for(vector <vector <vector <vector <int> > > >::iterator iter3 = 
	    (*iter2).begin(); iter3 != (*iter2).end(); iter3++){
	//Each of these can have up to N banks
	(*iter3).resize(maxBank);
	for(vector <vector <vector <int> > >::iterator iter4 = 
	      (*iter3).begin(); iter4 != (*iter3).end(); iter4++){
	  //Each of these source-sink pairs can have an col separation of
	  // between negative (num columns + 1) (input pin on right
	  // drives output pin on left
	  // and (num columns + 1) (input pin to output pin).
	  //Thus, if we make each of the third dimension vectors
	  // of size (num columns + 1 * 2 we can accomodate everyone
	  (*iter4).resize(((numCols + 1) * 2) + 1);
	  for(vector <vector <int> >::iterator iter5 = 
		(*iter4).begin(); iter5 != (*iter4).end(); iter5++){
	    //Initialize these values to be 0 since we haven't found any sinks yet
	    (*iter5).resize(((numRows + 1) * 2) + 1, 0);
	  }
	}
      }
    }
  }
}

//Calculate the average timing for each source/dest type/bank
void vprFPGA::calculateTimingTable(){
  int A, B, C, D, E, F;
  vector <vector <vector <vector <vector <vector <int> > > > > >::iterator iter;
  vector <vector <vector <vector <vector <int> > > > >::iterator iter2;
  vector <vector <vector <vector <int> > > >::iterator iter3;
  vector <vector <vector <int> > >::iterator iter4;
  vector <vector <int> >::iterator iter5;
  vector <int>::iterator iter6;
  
  for(iter = timingNumHitTable->begin(), A = 0; iter != timingNumHitTable->end(); iter++, A++){
    for(iter2 = (*iter).begin(), B = 0; iter2 != (*iter).end(); iter2++, B++){
      for(iter3 = (*iter2).begin(), C = 0; iter3 != (*iter2).end(); iter3++, C++){
	for(iter4 = (*iter3).begin(), D = 0; iter4 != (*iter3).end(); iter4++, D++){
	  for(iter5 = (*iter4).begin(), E = 0; iter5 != (*iter4).end(); iter5++, E++){
	    for(iter6 = (*iter5).begin(), F = 0; iter6 != (*iter5).end(); iter6++, F++){
	      if(*iter6 > 0){
		timingTable->at(A).at(B).at(C).at(D).at(E).at(F) /= (double)(*iter6);
		//cout << "Timing Table[" << A << "][" << B << "][" << C <<"]["  << D;
		//cout << "]["<< E << "][" << F << "] = " << timingTable->at(A).at(B).at(C).at(D).at(E).at(F) << endl;
	      }
	    }
	  }
	}
      }
    }
  }
}

void vprFPGA::emptyTimingTable(){
//   for(vector <vector <vector <vector <vector <vector <double> > > > > >::iterator iter = 
// 	timingTable->begin(); iter != timingTable->end(); iter++){
//     for(vector <vector <vector <vector <vector <double> > > > >::iterator iter2 =
// 	  (*iter).begin(); iter2 != (*iter).end(); iter2++){
//       for(vector <vector <vector <vector <double> > > >::iterator iter3 = 
// 	    (*iter2).begin(); iter3 != (*iter2).end(); iter3++){
// 	for(vector <vector <vector <double> > >::iterator iter4 = 
// 	      (*iter3).begin(); iter4 != (*iter3).end(); iter4++){
// 	  for(vector <vector <double> >::iterator iter5 = 
// 		(*iter4).begin(); iter5 != (*iter4).end(); iter5++){
// 	    // Row sep
// 	    (*iter5).resize(0);
// 	  }
// 	  //Col sep
// 	  (*iter4).resize(0);
// 	}
// 	//Dest Bank
// 	(*iter3).resize(0);
//       }
//       //Dest Type
//       (*iter2).resize(0);
//     }
//     //Source Bank
//     (*iter).resize(0);
//   }
//   //Source Type
//   timingTable->resize(0);
  delete timingTable;
}

void vprFPGA::emptyTimingNumHitTable(){
 //  for(vector <vector <vector <vector <vector <vector <int> > > > > >::iterator iter = 
// 	timingNumHitTable->begin(); iter != timingNumHitTable->end(); iter++){
//     for(vector <vector <vector <vector <vector <int> > > > >::iterator iter2 =
// 	  (*iter).begin(); iter2 != (*iter).end(); iter2++){
//       for(vector <vector <vector <vector <int> > > >::iterator iter3 = 
// 	    (*iter2).begin(); iter3 != (*iter2).end(); iter3++){
// 	for(vector <vector <vector <int> > >::iterator iter4 = 
// 	      (*iter3).begin(); iter4 != (*iter3).end(); iter4++){
// 	  for(vector <vector <int> >::iterator iter5 = 
// 		(*iter4).begin(); iter5 != (*iter4).end(); iter5++){
// 	    // Row sep
// 	    (*iter5).resize(0);
// 	  }
// 	  //Col sep
// 	  (*iter4).resize(0);
// 	}
// 	//Dest Bank
// 	(*iter3).resize(0);
//       }
//       //Dest Type
//       (*iter2).resize(0);
//     }
//     //Source Bank
//     (*iter).resize(0);
//   }
//   //Source Type
//   timingNumHitTable->resize(0);
  delete timingNumHitTable;
}

//Start a to-all search from some representational IO pins
void vprFPGA::sweepTimingIOPins(){
//   //Start from all ipins
//   for(int r = 1; r <= get_numRows(); r++){
//     //Left side
//     findAllSinksFrom((void*) getIOBAt(r, 0), IOPIN, 0, r, 0);
//     //Right side
//     findAllSinksFrom((void*) getIOBAt(r, get_numCols()+ 1), IOPIN, 0, r, get_numCols()+ 1);
//   }
//   for(int c = 1; c <= get_numCols(); c++){
//     //Top
//     findAllSinksFrom((void*) getIOBAt(0, c), IOPIN, 0, 0, c);
//     //Bottom
//     findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, c), IOPIN, 0, get_numRows()+ 1, c);
//   }
  //We should launch a search from each of the 8 corners, and something from the middle of each side
  //Top left corner
  findAllSinksFrom((void*) getIOBAt(0, 1), IOPIN, 0, 0, 1); //TOP
  findAllSinksFrom((void*) getIOBAt(1, 0), IOPIN, 0, 1, 0); //LEFT

  //Top right corner
  findAllSinksFrom((void*) getIOBAt(0, get_numCols()), IOPIN, 0, 0, get_numCols()); //TOP
  findAllSinksFrom((void*) getIOBAt(1, get_numCols()+ 1), IOPIN, 0, 1, get_numCols()+ 1); //RIGHT

  //Bottom left corner
  findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, 1), IOPIN, 0, get_numRows()+ 1, 1); //BOTTOM
  findAllSinksFrom((void*) getIOBAt(get_numRows(), 0), IOPIN, 0, get_numRows(), 0); //LEFT
  
  //Bottom right corner
  findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, get_numCols()), IOPIN, 0, get_numRows()+ 1, get_numCols()); //BOTTOM
  findAllSinksFrom((void*) getIOBAt(get_numRows(), get_numCols()+ 1), IOPIN, 0, get_numRows(), get_numCols()+ 1); //RIGHT

  //Middle of left side
  findAllSinksFrom((void*) getIOBAt((get_numRows()+ 1) / 2, 0), IOPIN, 0, (get_numRows()+ 1) / 2, 0);
		   
  //Middle of right side
  findAllSinksFrom((void*) getIOBAt((get_numRows()+ 1) / 2, get_numCols()+ 1), IOPIN, 0, (get_numRows()+ 1) / 2, get_numCols()+ 1);

  //Middle of top
  findAllSinksFrom((void*) getIOBAt(0, (get_numCols()+ 1) / 2), IOPIN, 0, 0, (get_numCols()+ 1) / 2);

  //Middle of bottom
  findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, (get_numCols()+ 1) / 2), IOPIN, 0, get_numRows()+ 1, (get_numCols()+ 1) / 2);
}

//Start a to-all search from some representational logic blocks
void vprFPGA::sweepTimingLogic(){
  //   for(int r = 1; r <= get_numRows(); r++){
  //     for(int c = 1; c <= get_numCols(); c++){
  //       findAllSinksFrom((void*) (getCLBAt(r,c)->getLUT(0)), LB, 0, r, c);
  //     }
  //   }
  //
  //Launch a search across the diagonals of the chip - notice we're assuming it's a square here!
  //Since we are missing IO pins in the corners, we should also perform searches from the
  // 2 strips to either side of the diagonal too.
  if(get_numCols() != get_numRows()){
    ERROR("We can only currently handle timing estimation for square chips!");
  }
  //Top left to bottom right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(r, r)->getLUT(0)), LB, 0, r, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(r, r+1)->getLUT(0)), LB, 0, r, r+1);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(r+1, r)->getLUT(0)), LB, 0, r+1, r);
  }

  //Bottom left to top right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r)->getLUT(0)), LB, 0, get_numRows() - r + 1, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r, r)->getLUT(0)), LB, 0, get_numRows() - r, r);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r + 1)->getLUT(0)), LB, 0, get_numRows() - r + 1, r + 1);
  }
}

//Start a to-all search from each register bank
void vprFPGA::sweepTimingDelay(){
//   for(int r = 1; r <= get_numRows(); r++){
//     for(int c = 1; r <= get_numCols(); c++){
//       //This only launches a search from the first bank in a delay block.
//       //This is fine until we add interconnect registers in.
//       findAllSinksFrom((void*) (getCLBAt(r,c)->getRegBank(0)), REGISTERBANK, 0, r, c);
//     }
//   }
  //Launch a search across the diagonals of the chip - notice we're assuming it's a square here!
  //Since we are missing IO pins in the corners, we should also perform searches from the
  // 2 strips to either side of the diagonal too.
  if(get_numCols() != get_numRows()){
    ERROR("We can only currently handle timing estimation for square chips!");
  }
  //Top left to bottom right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(r, r)->getRegBank(0)), REGISTERBANK, 0, r, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(r, r+1)->getRegBank(0)), REGISTERBANK, 0, r, r+1);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(r+1, r)->getRegBank(0)), REGISTERBANK, 0, r+1, r);
  }

  //Bottom left to top right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r)->getRegBank(0)), REGISTERBANK, 0, get_numRows() - r + 1, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r, r)->getRegBank(0)), REGISTERBANK, 0, get_numRows() - r, r);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r + 1)->getRegBank(0)), REGISTERBANK, 0, get_numRows() - r + 1, r + 1);
  }
}

//Perform a to-all search from the source of this block.
void vprFPGA::findAllSinksFrom(void* sourceBlock, e_timingMoveBlockType sourceType, int sourceBank, int sourceRow, int sourceCol){
  priority_queue<pqNode> PQ;
  routingNode *source;
  routingNode *rnode;
  routingNode* nb_rnode;
  routingEdge* nb_edge;
  double pathDelay;
  double newPathDelay;
  list<routingNode*> cleanupList;

#ifdef PARANOID_CHECK_ROUTE
  //Go through the entire routing graph and make sure that all nodes are unexplored before we start
  for(list<routingNode*>::iterator rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }
  }
#endif

  if(sourceType == IOPIN){
    //Pick one of the source routing nodes to be the source for this entire search
    if(((vprIOB*)sourceBlock)->sourceArray->size() == 0){
      ERROR("Input pin has no source!");
    }
    source = ((vprIOB*)sourceBlock)->sourceArray->front();
  }
  else if(sourceType == LB){
    //Pick one of the source routing nodes to be the source for this entire search
    if(!((vprLUT*)sourceBlock)->sourceNode){
      ERROR("Block has no source!");
    }
    source = ((vprLUT*)sourceBlock)->sourceNode;
  }
  else if(sourceType == REGISTERBANK){
    //Pick the first routing node in the bank to be the source for this search
    source = ((registerBank*)sourceBlock)->getFirstRoutingNode();
  }

  //Push the source into the PQ at a cost of 1.0 * intrinsic delay of this node
  //We are trying to find the shortest timing distance between the source and all other nodes so 
  // we will be putting everything in at a cost of just the delay
  //PQ.push(pqNode(source, source->intrinsicDelay, 0.0, 0, source->intrinsicDelay));
  
  //Actually, don't push the source in to the PQ, push the neighbors of the source into the PQ.
  //Just in case the source is also the sink, we want to go through at least 1 edge in the
  // RR graph.
  source->create_fanoutListItr();
  for(nb_edge = source->getnext_fanoutList(); nb_edge != 0; nb_edge = source->getnext_fanoutList()){
    nb_rnode = nb_edge->destNode;
    //Only add unexplored neighbors to the PQ
    if(nb_rnode->isExplored == false){
      newPathDelay = source->intrinsicDelay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
      PQ.push(pqNode(nb_rnode, newPathDelay, 0.0 ,rnode, newPathDelay));
    }
  }
  source->remove_fanoutListItr();

  //Keep exploring till the PQ is empty
  while(!PQ.empty()){
    rnode=(PQ.top()).rNode;
    
    //We will only pay attention to a routing node the first time we see it
    if(rnode->isExplored == true){
    }
    else{
      //Mark the node explored
      rnode->isExplored=true;
      cleanupList.push_back(rnode);

      pathDelay = (PQ.top()).netDelay;

      //See if this routing node belongs to something interesting (logic block sink, output pin sink, register bank terminal)
      if(rnode->type == SINK){
	//Figure out if this is an I/O pin or a logic block
	if(rnode->col_low == 0 || rnode->col_low == (FPGA->numCols + 1) || 
	   rnode->row_low == 0 || rnode->row_low == (FPGA->numRows+1)){
	  //cout << "Source at " << source->col_low << ", " << source->row_low << " found IO pin at " << rnode->col_low << ", " << rnode->row_low << endl;
	  addTimingTable(pathDelay, sourceType, sourceBank, IOPIN, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	  incNumHitTable(sourceType, sourceBank, IOPIN, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);

	}
	else{
	  addTimingTable(pathDelay, sourceType, sourceBank, LB, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	  incNumHitTable(sourceType, sourceBank, LB, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	}
      }
      else if(rnode->type == DELAY){
	//Found a delay node that is not the source
	addTimingTable(pathDelay, sourceType, sourceBank, REGISTERBANK, rnode->getBank(), 
		       rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	incNumHitTable(sourceType, sourceBank, REGISTERBANK, rnode->getBank(), 
		       rnode->row_low - sourceRow, rnode->col_low - sourceCol);

      }
      //If it is not a destination, add it's neighbors to the PQ
      else{
	rnode->create_fanoutListItr();
	for(nb_edge=rnode->getnext_fanoutList(); nb_edge!=0; nb_edge=rnode->getnext_fanoutList()){
	  nb_rnode=nb_edge->destNode;
	  //Only add unexplored neighbors to the PQ
	  if(nb_rnode->isExplored == false){
	    newPathDelay = pathDelay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	    PQ.push(pqNode(nb_rnode, newPathDelay, 0.0 ,rnode, newPathDelay));
	  }
	}
	rnode->remove_fanoutListItr();
      }
    }
    PQ.pop();  
  }
  
  //Reset all of the nodes we hit to unexplored
  while(!cleanupList.empty()){
    cleanupList.front()->isExplored=false;
    cleanupList.pop_front();
  }
}



//Set a particular location in the table
//Notice that for the x separation, we can have values
//ranging from negative (num columns + 1) to positive (num columns + 1)
//so we have an offset built in.
void vprFPGA::addTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
//   //Basically this is an uninitialized place, so replace the really big value with the value we want to add in
//   if(timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0E19){
//     timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
//   }
//   else
//     timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) += value;

  //These changes assume that all banks have the same connectivity
  if(timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0E19){
    timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
  }
  else
    timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) += value;

  //These changes assume that A) All types of blocks have the same connectivity (IOB, LB, REG) and B) All
  // registers at different banks have the same connectivity.
  //This is fine until we introducet interconnect registers
//   if(timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0E19){
//     timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
//   }
//   else
//     timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) += value;
}

void vprFPGA::incNumHitTable(e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
  //timingNumHitTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) ++;
  //These changes assume that all banks have the same connectivity
  timingNumHitTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) ++;
  //These changes assume that A) All types of blocks have the same connectivity (IOB, LB, REG) and B) All
  // registers at different banks have the same connectivity.
  //This is fine until we introducet interconnect registers
  //timingNumHitTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) ++;
}


void vprFPGA::setTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
  cout << "Timing Table[" << sourceType << "][" << sourceBank << "][" << destType <<"]["  << destBank;
  cout << "]["<< colSep << "][" << rowSep << "] = " << value << endl;
  
  //timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
  //These changes assume that all banks have the same connectivity
  timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;

  //These changes assume that A) All types of blocks have the same connectivity (IOB, LB, REG) and B) All
  // registers at different banks have the same connectivity.
  //This is fine until we introduce interconnect registers
  //timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
}

//Try to get the CLB at the given location.
//If there isn't one there, create one.
vprCLB* vprFPGA::getOrMakeCLBAt(int row, int col){
  //Make sure we are asking for a CLB location
  if(row < 1 || row > numRows || col < 1 || col > numCols){
    ERROR("Requested invalid CLB location");
  }

  // We might not have that many rows
  if(array.size() <= row){
    array.resize(row + 1);
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void*) new vprCLB(row, col);
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void *) new vprCLB(row, col);
  }
  //Even if we have enough rows and columns, we might have skipped one in between
  // so we have to fill it in.
  if(!array.at(row).at(col)){
    array.at(row).at(col) = (void *) new vprCLB(row, col);
  }

  return ((vprCLB*) array.at(row).at(col));
}

//Try to get the clb at the given location.
//There might not be one at that location though.
vprCLB* vprFPGA::getCLBAt(int row, int col){
  //Make sure we are asking for a CLB location
  if(row < 1 || row > numRows || col < 1 || col > numCols){
    ERROR("Requested invalid CLB location");
  }
  // We might not have that many rows
  if(array.size() <= row){
    return NULL;
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    return NULL;
  }
  else{
    return ((vprCLB*) array.at(row).at(col));
  }
}

//Try to get the IOB at the given location.
//If there isn't one there, create one.
vprIOB* vprFPGA::getOrMakeIOBAt(int row, int col){
  short int a = 0;
  short int b = 0;
  short int c = 0;
  short int d = 0;

  //Make sure we are asking for a valid IOB location
  //Basically this is an XOR of the possible combinations
  if(row == 0){
    a = 1;
  }
  else if(row == numRows + 1){
    b = 1;
  }
  if(col == 0){
    c = 1;
  }
  else if(col == numCols + 1){
    d = 1;
  }

  if((a ^ b ^ c ^ d) == 0){
    ERROR("Requested invalid IOB location");
  }

  // We might not have that many rows
  if(array.size() <= row){
    array.resize(row + 1);
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void*) new vprIOB(row, col, numIOs_perIOB);
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void *) new vprIOB(row, col, numIOs_perIOB);
  }

  return ((vprIOB*) array.at(row).at(col));
}

//Try to get the IOB at the given location.
//There might not be one at that location though.
vprIOB* vprFPGA::getIOBAt(int row, int col){
  short int a = 0;
  short int b = 0;
  short int c = 0;
  short int d = 0;

  //Make sure we are asking for a valid IOB location
  //Basically this is an XOR of the possible combinations
  if(row == 0){
    a = 1;
  }
  else if(row == numRows + 1){
    b = 1;
  }
  if(col == 0){
    c = 1;
  }
  else if(col == numCols + 1){
    d = 1;
  }

  if((a ^ b ^ c ^ d) == 0){
    ERROR("Requested invalid IOB location");
  }

  // We might not have that many rows
  if(array.size() <= row){
    return NULL;
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    return NULL;
  }
  else{
    return ((vprIOB*) array.at(row).at(col));
  }
}

int vprFPGA::getNumberofBlockType(int blockType){
  switch(blockType){
  case LUT:
    return lutList.size();
  case MULTIPLIER:
    return multList.size();
  case RAM_BLOCK:
    return ramList.size();
  case INPUTPAD:
    ERROR("We shouldn't see an input block here!");
    break;
  case OUTPUTPAD:
    ERROR("We shouldn't see an output block here!");
    break;
  case REGISTER_BANK:
    return registerBankVector.size();
  }
}

registerBank* vprFPGA::getRandomFreeRegBank(){
  list<registerBank *>::iterator iter;
  int i;
  int select;

  if(freeRegisterBankList.size() == 0){
    return NULL;
  }
  else{
    select = lrand48() % freeRegisterBankList.size();
    for(iter = freeRegisterBankList.begin(), i = 0; 
	iter != freeRegisterBankList.end() && i < select; iter++, i ++){
    }
    return (*iter);
  }
}

vprIOB* vprFPGA::getOtherIOB(int row, int col, int rLim){
  int rowRLim = min(rLim, numRows);
  int colRLim = min(rLim, numCols);
  int side;
  int newRow = row;
  int newCol = col;
  int moveAmt;

  //See if the rLim is still the whole chip
  if(numRows <= rowRLim && numCols <= colRLim){
    while(newRow == row && newCol == col){
      //From VPR:
      /*                              *
       *       +-----1----+           *
       *       |          |           *
       *       |          |           *
       *       0          2           *
       *       |          |           *
       *       |          |           *
       *       +-----3----+           *
       *                              */
      side = lrand48() % 4;
      switch(side){
      case 0:
	newCol = 0;
	newRow = (lrand48() % numRows) + 1;
	break;
      case 1:
	newCol = (lrand48() % numCols) + 1;
	newRow = 0;
	break;
      case 2:
	newCol = numCols + 1;
	newRow = (lrand48() % numRows) + 1;
	break;
      case 3:
	newCol = (lrand48() % numCols) + 1;
	newRow = numRows + 1;
	break;
      }
    }
  }
  else{
    //The idea of windowing, according to VPR is a bit different for
    // IO pins.  If we generate a positive/negative move amount
    // and moving would leave the side of the chip we are
    // currently on, we generate a second move amount value
    // and fix the correct row or column.  I think this is done
    // to provide a smoother transistion between an rLim of
    // numRow and numRow-1.
    moveAmt = 0;

    if(col == 0 || col == numCols + 1){
      //Coming from the left side or right side.
      //Try to find a new row, keep the same column.
      while (moveAmt == 0){
	moveAmt = (lrand48() % ((2 * rowRLim) + 1)) - rowRLim;
      }
      newRow += moveAmt;

      //We might have tried to move up or down too far and fell off of
      // the left/right sides to the top or bottom sides.
      if(newRow < 1 || newRow > numRows){
	//Did we want to move to top to bottom sides?
	if(newRow < 1){
	  newRow = 0;
	}

	else{
	  newRow = numRows + 1;
	}

	//Find a new column.
	if(col == 0){
	  //If we came from left side, the new column should be (1, colRLim)
	  newCol = (lrand48() % colRLim) + 1;
	}
	else{
	  //If we came from the right side, the new column should be (numCols - colRLim + 1, numCols)
	  newCol = numCols - (lrand48()% colRLim);
	}
      }
    }
    else if(row == 0 || row == numRows + 1){
      //Coming from the top or bottom.
      //Try to find a new column, keep the same row
      while (moveAmt == 0){
	moveAmt = (lrand48() % ((2 * colRLim) + 1)) - colRLim;
      }
      newCol += moveAmt;
      
      //We might have tried to move left or right too far and fell off of
      // the top/bottom to the left or right sides.
      if(newCol < 1 || newCol > numCols){
	//Did we want to move to the left or right side?
	if(newCol < 1){
	  newCol = 0;
	}
	else{
	  newCol = numCols + 1;
	}

	//Find a new row
	if(row == 0){
	  //If we came from the top, the new row should be (1, rowRLim)
	  newRow = (lrand48() % rowRLim) + 1;
	}
	else{
	  //If we came from the bottom, the new row should be (numRows - rowRLim + 1, numRows)
	  newRow = numRows - (lrand48()% rowRLim);
	}
      }
    }
  }
  return(getIOBAt(newRow, newCol));
}

vprLUT* vprFPGA::getOtherLUT(int row, int col, int rLim){
  int moveLimitRow = min(numRows, rLim);
  int moveLimitCol = min(numCols, rLim);
  int newRow = row;
  int newCol = col;
  int moveAmtRow = 0;
  int moveAmtCol = 0;
  vprLUT* lut = 0;
  
  while(!lut || (newRow == row && newCol == col)){
    moveAmtRow = (lrand48() % ((2 * moveLimitRow) + 1)) - moveLimitRow;
    moveAmtCol = (lrand48() % ((2 * moveLimitCol) + 1)) - moveLimitCol;

    newRow = row + moveAmtRow;
    newCol = col + moveAmtCol;
    
    //VPR does a wrap around, as commented out, but that doesn't seem right.
    //I'm doing mirroring instead.
    if(newRow < 1){
      //newRow = newRow + numRows;
      newRow = (newRow * -1) + 1;
    }
    else if(newRow > numRows){
      //newRow = newRow - numRows;
      newRow = numRows - (newRow - (numRows + 1));
    }
    if(newCol < 1){
      //newCol = newCol + numCols;
      newCol = (newCol * -1) + 1;
    }
    else if(newCol > numCols){
      //newCol = newCol - numCols;
      newCol = numCols - (newCol - (numCols + 1));
    }
    lut = getCLBAt(newRow, newCol)->getLUT(lrand48() % getCLBAt(newRow, newCol)->getNumLUTs());

    if(lut->get_blockType() != LUT){
      lut = 0;
    }
  }
  return(lut);
}

//Since there is likely large gaps in the arrangements of multipliers, I've got to think of something
// to make windowing work - otherwise we'll just do un-windowed.
//vprLUT* vprFPGA::getOtherMult(int row, int col, int rLim);
//vprLUT* vprFPGA::getOtherRam(int row, int col, int rLim);
registerBank * vprFPGA::getOtherRegBank(int row, int col, int rLim){
  int moveLimitRow = min(numRows, rLim);
  int moveLimitCol = min(numCols, rLim);
  int newRow = row;
  int newCol = col;
  int moveAmtRow;
  int moveAmtCol;
  
  while(newRow == row && newCol == col){
    moveAmtRow = (lrand48() % ((2 * moveLimitRow) + 1)) - moveLimitRow;
    moveAmtCol = (lrand48() % ((2 * moveLimitCol) + 1)) - moveLimitCol;

    newRow = row + moveAmtRow;
    newCol = col + moveAmtCol;
    
    //VPR does a wrap around, as commented out, but that doesn't seem right
    if(newRow < 1){
      newRow = (newRow * -1) + 1;
    }
    else if(newRow > numRows){
      newRow = numRows - (newRow - (numRows + 1));
    }
    if(newCol < 1){
      newCol = (newCol * -1) + 1;
    }
    else if(newCol > numCols){
      newCol = numCols - (newCol - (numCols + 1));
    }
  }
  return(getCLBAt(newRow, newCol)->getRegBank(lrand48() % getCLBAt(newRow, newCol)->getNumRegBanks()));
}

//Figure out what drives this register bank.
//If it's a LUT, register bank or pipelining register, return a register bank in the same CLB as it's parent.
//If it's an input pad, return NULL.
registerBank* registerBank::getRegBankInParentHostCLB(){
  int sourceRow, sourceCol;
  int myRow = hostCLB->get_row();
  int myCol = hostCLB->get_col();
  Terminal *sink;
  Terminal *source;

  //Regardless of whether or not this register block is currently mapped to
  // a logic block or a string of individual pipelining registers, we can find out
  // what drives this by looking at the first sink.
  sink = getFirstSink();

  //This sink might be driven by a pipelining net or a regular net
  if(sink->sourcePipeliningNet){
    source = sink->sourcePipeliningNet->source;
  }
  else{
    source = sink->net->source;
  }

  sourceRow = source->getLogicBlockRow();
  sourceCol = source->getLogicBlockCol();
  
  if(source->lBlock){
    if(source->lBlock->blockType == INPUTPAD){
      return NULL;
    }
    else if(source->lBlock->blockType == OUTPUTPAD){
      ERROR("Source is an output pad!");
    }
  }

  //Right now we return a completely random register bank - should we preferably return an empty one?
  return(FPGA->getCLBAt(sourceRow, sourceCol)->getRegBank(lrand48() % FPGA->getCLBAt(sourceRow, sourceCol)->getNumRegBanks()));      
}

void vprFPGA::occupyRegBank(registerBank* rb){
  freeRegisterBankList.remove(rb);
  numFreeRegisters -= rb->getNumPossibleDelay();
}

void vprFPGA::vacateRegBank(registerBank* rb){
  freeRegisterBankList.push_back(rb);
  numFreeRegisters += rb->getNumPossibleDelay();
}

//Go though all of the register banks and make sure that the
// empty (and only empty ones) are in the free register list
void vprFPGA::checkFreeRegisters(){
  int count = 0;
  vector<registerBank *>::iterator vIter;
  list<registerBank *>::iterator lIter;
  list<registerBank *>::iterator lIter2;
  
  for(vIter = registerBankVector.begin(); vIter != registerBankVector.end(); vIter++){
    if((*vIter)->isBankEmpty()){
      //We should have this register in the free register bank list
      for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
	if(*lIter == *vIter){
	  break;
	}
      }
      if(lIter == freeRegisterBankList.end()){
	ERROR("Free register mismatch!");
      }
    }
    else{
      //We should not have this register in the free register bank list
      for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
	if(*lIter == *vIter){
	  break;
	}
      }
      if(lIter != freeRegisterBankList.end()){
	ERROR("Free register mismatch!");
      }
    }
  }

  //Look for duplicate registers in the free register list
  for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
    //Find the first instance of this register in the list
    for(lIter2 = freeRegisterBankList.begin(); lIter2 != freeRegisterBankList.end(); lIter2++){
      if(*lIter == *lIter2){
	lIter2++;
	break;
      }
    }
    //Look for more
    for( ; lIter2 != freeRegisterBankList.end(); lIter2++){
      if(*lIter == *lIter2){
	ERROR("Duplicate entry!");
      }
    }
  }

  //Make sure that the number of empty registers matches the number of registers in the free list
  for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
    count += (*lIter)->getNumPossibleDelay();
  }
  if(count != numFreeRegisters){
    ERROR("Free register mismatch!");
  }
}

//We need to find an empty IOB, just grab the first one
void vprFPGA::placeClockPin(int *clockRow, int *clockCol, int *clockPTC){
  vector<vprIOB *>::iterator iter;

  //Look through all the IOBs
  for(iter = iobList.begin(); iter != iobList.end(); iter++){
    //Look at each pin within the IOB
    for(int i = 0; i < (*iter)->get_numIOs(); i++){
      if(!(*iter)->getfrom_currBlockArray(i)){
	*clockRow = (*iter)->get_row();
	*clockCol = (*iter)->get_col();
	*clockPTC = i;
	return;
      }
    }
  }

  for(iter = iobList.begin(); iter != iobList.end(); iter++){
    //Print out the contents of each iob
    for(int i = 0; i < (*iter)->get_numIOs(); i++){
      if((*iter)->getfrom_currBlockArray(i)){
	printf("IOB %d, %d, %d has block %s\n", (*iter)->get_row(),
	       (*iter)->get_col(), i, (*iter)->getfrom_currBlockArray(i)->name.c_str());
      }
      else{
	printf("IOB %d, %d, %d is empty!\n", (*iter)->get_row(),
	       (*iter)->get_col(), i);
      }
    }
  }
  
  ERROR("We need an empty IOB location, but we can't find one!");
}

//Go through all iobs, register banks and LUTs to make sure that,
// if they have a mapping, that the logic block agrees
void vprFPGA::checkArrayMappings(){
  int i;
  vector<logicBlock*>::iterator lbIter;

  for(vector<vprIOB *>::iterator iter = iobList.begin(); iter != iobList.end(); iter++){
    for(lbIter = (*iter)->currBlockArray->begin(), i = 0; 
	lbIter != (*iter)->currBlockArray->end(); lbIter++, i++){
      if(*lbIter){
	if((*lbIter)->row != (*iter)->row || (*lbIter)->col != (*iter)->col || (*lbIter)->ptc_index != i){
	  ERROR("IOB placement mismatch!");
	}
      }
    }
  }

  for(vector<vprLUT *>::iterator iter = lutList.begin(); iter != lutList.end(); iter++){
    if((*iter)->currBlock){
      //We can't check the ptc index, since the physical LUT doesn't keep track of what
      // ptc index it is itself.  We would have to add that info in.
      if((*iter)->currBlock->row != (*iter)->row || (*iter)->currBlock->col != (*iter)->col)
	ERROR("LUT placement mismatch!");
    }
  }

  for(vector<registerBank *>::iterator iter = registerBankVector.begin(); iter != registerBankVector.end(); iter++){
    if(!(*iter)->isBankEmpty()){
      //This register bank is occupied by a logic block register bank
      if((*iter)->get_currBlock()){
	if((*iter)->get_currBlock()->row != (*iter)->getHostCLB()->row || 
	   (*iter)->get_currBlock()->col != (*iter)->getHostCLB()->col || 
	   (*iter)->get_currBlock()->ptc_index != (*iter)->getHostBankNum() || 
	   (*iter)->get_currBlock()->rb != *iter){
	  ERROR("Reg bank placement mismatch!");
	}
      }
      //This register bank is occupied by pipelining registers
      else{
	for(list <Terminal *>::iterator prIter = (*iter)->currSinkList.begin(); prIter != (*iter)->currSinkList.end(); prIter++){
	  if((*prIter)->getLogicBlockRow() != (*iter)->getHostCLB()->row || 
	     (*prIter)->getLogicBlockCol() != (*iter)->getHostCLB()->col){
	    ERROR("Reg bank placement mismatch!");
	  }
	}
      }
    }
  }
}


// We assume that registers are added in the correct order if that is important
// That is, in a particular bank, register 0 is added before register 1,
// register 1 is added before register 2, etc.if they are in a
// shift-chain orientation.
void registerBank::addRoutingNode(routingNode *r){
  routingNodeVector.push_back(r);
}

//Return the first sink in this bank
Terminal* registerBank::getFirstSink(){
  if(currSinkList.empty()){
    return NULL;
  }
  else{
    return currSinkList.front();
  }
}

Terminal* registerBank::getLastSource(){
  if(currSourceList.empty()){
    return NULL;
  }
  else{
    return currSourceList.back();
  }
}

//Return the first routing node of this register bank
routingNode* registerBank::getFirstRoutingNode(){
  if(routingNodeVector.empty()){
    ERROR("Register bank has no routing nodes assigned!");
  }
  else{
    return routingNodeVector.front();
  }
}

//Are there any registers currently mapped to this bank?
bool registerBank::isBankEmpty(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }
  return(currSourceList.empty());
}

//How many registers are currently mapped to this bank?
int registerBank::getNumMappedRegs(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }
  return(currSourceList.size());
}

//How many registers could be mapped to this bank?
int registerBank::getNumPossibleDelay(){
  if(routingNodeVector.size() == 0){
    ERROR("Bank has no routing nodes assigned!");
  }
  return(routingNodeVector.size());
}

//How many registers are currently vacent in this bank?
int registerBank::getNumFreeRegs(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }
  return(routingNodeVector.size() - currSourceList.size());
}


//Is this bank at capacity?
bool registerBank::bankFull(){
  if(routingNodeVector.size() == 0){
    ERROR("Bank has no routing nodes assigned!");
  }

  if(routingNodeVector.size() == currSourceList.size()){
    return true;
  }
  else{
    return false;
  }
}

//Put this register into this bank
void registerBank::placeTerminal(Terminal *T){
  // See if this bank is already full
  if(bankFull()){
    ERROR("placeTerminal: Attempted to map terminal to full bank!");
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isBankEmpty()){
    FPGA->occupyRegBank(this);
  }

  //Since this is only used by pipelined nets, there is only
  // one terminal - it is both a source and a sink
  currSinkList.push_back(T);
  currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));
  currSourceList.push_back(T);
  currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));

  //Make sure this terminal also knows who it belongs to
  T->setRegisterBank(this);
}

//Put this register into the front of this bank instead of the end
//Notice, this does not set any routing nodes since things get
// screwed up every time we push in a new register.
//We must call updateRoutingNodes after this function is done.
void registerBank::placeTerminalFront(Terminal *T){
  // See if this bank is already full
  if(bankFull()){
    ERROR("placeTerminal: Attempted to map terminal to full bank!");
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isBankEmpty()){
    FPGA->occupyRegBank(this);
  }

  //Since this is only used by pipelined nets, there is only
  // one terminal - it is both a source and a sink
  currSinkList.push_front(T);
  currSourceList.push_front(T);

  //Make sure this terminal also knows who it belongs to
  T->setRegisterBank(this);
}

void registerBank::placeLogicBlock(logicBlock *lb){
  if(!lb){
    ERROR("Cannot place null logic block!");
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(!isBankEmpty()){
    ERROR("Attempted to map logic block to occupied bank!");
  }

  if(get_currBlock()){
    ERROR("Attempted to map logic block to bank already allocated to another logic block!!");
  }

  FPGA->occupyRegBank(this);
  
  //Make sure that the logic block knows who it is mapped to
  lb->set_registerBank(this);
  lb->set_ptc_index(hostBankNum);
  currBlock = lb;
  
  //Map the input net of this lb to the first sink node of the RB
  if(lb->inputList.size() > 1){
    ERROR("Register has more than one input!");
  }
  currSinkList.push_back(lb->inputList.front());
  currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));

  //Map the output net of this lb to the first source node the RB
  if(lb->outputList.size() > 1){
    ERROR("Register has more than one output!");
  }
  currSourceList.push_back(lb->outputList.front());
  currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));
}

void registerBank::unPlaceLogicBlock(logicBlock *lb){
  if(!lb){
    ERROR("Cannot unplace null logic block!");
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isBankEmpty()){
    ERROR("Attempted to unmap empty bank!");
  }

  if(get_currBlock() != lb){
    ERROR("Attempted to unmap logic block that was not previously mapped!");
  }

  FPGA->vacateRegBank(this);
  
  //Make sure that the logic block knows who it is mapped to
  lb->unset_registerBank(this);
  lb->set_ptc_index(-1);
  currBlock = 0;
  
  if(currSinkList.size() != 1){
    ERROR("Sink list should have exactly one terminal!");
  }
  
  currSinkList.front()->clearRNode();
  currSinkList.clear();

  if(currSourceList.size() != 1){
    ERROR("Source list should have exactly one terminal!");
  }
  
  currSourceList.front()->clearRNode();
  currSourceList.clear();
}

//Put this register into the bank after the parent.
//The parent might not be mapped to this register bank, though.
void registerBank::insertRegisterForwards(Terminal *T, Terminal *parent){
  list <Terminal *>::iterator sinkIter;
  list <Terminal *>::iterator sourceIter;
  int i;

  // See if this bank is already full
  if(bankFull()){
    ERROR("Attempted to map terminal to full bank!");
  }
  
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isBankEmpty()){
    FPGA->occupyRegBank(this);
    currSinkList.push_back(T);
    currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));
    currSourceList.push_back(T);
    currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));    
  }
  else{
    //Find the parent in the sink/source list
    for(sinkIter = currSinkList.begin(), sourceIter = currSourceList.begin(), i = 0; 
	sinkIter != currSinkList.end() && sourceIter != currSourceList.end();
	sinkIter ++, sourceIter++, i++){
      if(*sinkIter == parent){
	break;
      }
    }
    
    if(sinkIter == currSinkList.end()){
      //We didn't find the parent in the list, so just put it in the front of the bank
      // and then update the routing node assignments of everything behind it
      currSinkList.push_front(T);
      currSourceList.push_front(T);
      for(sinkIter = currSinkList.begin(), i = 0; sinkIter != currSinkList.end(); sinkIter++, i++){
	(*sinkIter)->set_rNode(routingNodeVector.at(i));
      }
    }
    else{
      //We did find the parent in the list, so insert it behind the parent, then update 
      // the routing node assignment of everyone behind the parent
      sinkIter++;
      sourceIter++;
      i++;
      currSinkList.insert(sinkIter, T);
      currSourceList.insert(sourceIter, T);
      sinkIter --;
      sourceIter --;
      for( ; sinkIter != currSinkList.end(); sinkIter++, i++){
	(*sinkIter)->set_rNode(routingNodeVector.at(i));
      }
    }
  }
}

//Put this register into the bank after the parent.
//The parent might not be mapped to this register bank, though.
void registerBank::insertRegisterBackwards(Terminal *T, Terminal *sink){
  list <Terminal *>::iterator sinkIter;
  list <Terminal *>::iterator sourceIter;
  int i;

  // See if this bank is already full
  if(bankFull()){
    ERROR("Attempted to map terminal to full bank!");
  }
  
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isBankEmpty()){
    FPGA->occupyRegBank(this);
    currSinkList.push_back(T);
    currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));
    currSourceList.push_back(T);
    currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));    
  }
  else{
    //Find the sink in the sink/source list
    for(sinkIter = currSinkList.begin(), sourceIter = currSourceList.begin(), i = 0; 
	sinkIter != currSinkList.end() && sourceIter != currSourceList.end();
	sinkIter ++, sourceIter++, i++){
      if(*sinkIter == sink){
	break;
      }
    }
    
    if(sinkIter == currSinkList.end()){
      //We didn't find the sink in the list, so just put it in the back of the bank
      // and then update the routing node assignment
      currSinkList.push_back(T);
      currSourceList.push_back(T);
      currSinkList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));
    }
    else{
      //We did find the sink in the list, so insert it in front of the sink, then update 
      // the routing node assignment of everyone behind the parent
      currSinkList.insert(sinkIter, T);
      currSourceList.insert(sourceIter, T);
      sinkIter--;
      sourceIter--;
      for( ; 	sinkIter != currSinkList.end(); sinkIter++, i++){
	(*sinkIter)->set_rNode(routingNodeVector.at(i));
      }
    }
  }
}

//Take the entire contents of this register bank and swap it with
// the contests of the other one.
//This is only called from pipelining registers, which means we are not
// doing clb-level placement.
void registerBank::swapBank(registerBank *otherBank){
  list <Terminal*> tempTermList;
  list <Terminal*>::iterator iter;
  int i;
  logicBlock *tempLb;

  //The other block may be empty, so we will make it full and empty this one
  if(otherBank->isBankEmpty()){
    FPGA->occupyRegBank(otherBank);
    FPGA->vacateRegBank(this);
  }

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Copy all of sinks from this bank
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    tempTermList.push_back(*iter);
  }
  //Clear the sink list from this bank
  currSinkList.clear();  

  //Copy the sinks from the other bank and update what register bank they are pointed to,
  //along with the routing node they are mapped to
  for(iter = otherBank->currSinkList.begin(), i = 0;
      iter != otherBank->currSinkList.end(); iter++, i++){
    currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(this);
    (*iter)->set_rNode(routingNodeVector.at(i));
  }
  //Clear the sink list from the other block
  otherBank->currSinkList.clear();

  //Copy the sinks that used to be in this block into the other block
  for(iter = tempTermList.begin(), i = 0; iter != tempTermList.end(); iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the temp sink list
  tempTermList.clear();

  //Copy all of sources from this bank
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    tempTermList.push_back(*iter);
  }
  //Clear the source list from this bank
  currSourceList.clear();  

  //Copy the sources from the other bank
  for(iter = otherBank->currSourceList.begin(), i = 0;
      iter != otherBank->currSourceList.end(); iter++, i++){
    currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(this);
    (*iter)->set_rNode(routingNodeVector.at(i));
  }
  //Clear the source list from the other bank
  otherBank->currSourceList.clear();

  //Copy the sources that used to be in this bank into the other bank
  for(iter = tempTermList.begin(), i = 0; iter != tempTermList.end(); iter++, i ++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  
  //Find out if either bank was associated with a logic block
  //If so, we should update the logic block to point to the new delay block and bank
  //Since this was called from a pipelining register, we are certain that
  // this register bank did not have a currBlock.
  //However, the other one may have been a regular logic block, not a pipelining register.
  tempLb = currBlock;
  currBlock = otherBank->currBlock;
  otherBank->currBlock = tempLb;
  if(currBlock){
    currBlock->unset_registerBank(otherBank);
    currBlock->set_registerBank(this);
  }

  if(otherBank->currBlock){
    //I don't think this should ever happen
    //However, if I call it from the physical CLB
    // swapping it can.
    //ERROR("Should this occur?");
    otherBank->currBlock->unset_registerBank(this);
    otherBank->currBlock->set_registerBank(otherBank);
  }
}

//Copy all of the sources and sinks to the other bank, then clean this bank out
void registerBank::copyContentsToAndEmpty(registerBank *otherBank){
  list <Terminal*>::iterator iter;
  int i;

  //The other block may be empty, so we will make it full and empty this one
  if(!otherBank->isBankEmpty()){
    ERROR("The other bank should be empty!");
  }

  FPGA->occupyRegBank(otherBank);
  FPGA->vacateRegBank(this);

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Copy all of sinks from this bank
  for(iter = currSinkList.begin(), i = 0; iter != currSinkList.end(); iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the sink list from this bank
  currSinkList.clear();  

  //Copy all of sources from this bank
  for(iter = currSourceList.begin(), i = 0; iter != currSourceList.end(); iter++, i++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the source list from this bank
  currSourceList.clear();  
}

void registerBank::updateSinkDelaysAndSlacks(){
  list <Terminal*>::iterator iter;

  //Make sure that the bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Update the sink timing for the nets attached to this block
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the sink timing for this net
    (*iter)->updateSinkDelayAndSlack();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    //This terminal is the source of a net.
    //Update the sink timings for all the sinks of this net
    (*iter)->updateSourceDelaysAndSlacks();
  }
}

void registerBank::updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay){
  list<Terminal*>::iterator iter;
  
  //Make sure that the bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Recalculate the placement cost for the net attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    (*iter)->updateSinkWireAndTimingCost(timing_driven, criticalPathDelay);
  }
  
  // Recalculate the placement cost for the nets attached to the sources
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    (*iter)->updateSourceWireAndTimingCosts(timing_driven, criticalPathDelay);
  }
}

double registerBank::determineDeltaWireCost(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in wirelength.
  list<Terminal*>::iterator iter;
  double deltaWire = 0.0;

  //Make sure that this bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }
  
  // Recalculate the placement cost for the nets attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    deltaWire += (*iter)->sinkDeltaWireCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources 
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    deltaWire += (*iter)->sourceDeltaWireCost();
  }
  return deltaWire;
}

double registerBank::determineDeltaTimingCost(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTiming = 0.0;
  
  //Make sure that this bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Recalculate the placement cost for the nets attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    deltaTiming += (*iter)->sinkDeltaTimingCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    deltaTiming += (*iter)->sourceDeltaTimingCost();
  }
  return deltaTiming;
}

double registerBank::determineDeltaTimingCostPrime(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTimingPrime = 0.0;
  
  //Make sure that this bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Recalculate the placement cost for the nets attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    deltaTimingPrime += (*iter)->sinkDeltaTimingCostPrime();
  }
  
  // Recalculate the placement cost for the nets attached to the sources
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    deltaTimingPrime += (*iter)->sourceDeltaTimingCostPrime();
  }
  return deltaTimingPrime;
}

//Move the first N register in this bank to the end of the other bank
void registerBank::migrateFirstNRegs(registerBank *otherBank, int numRegs){
  list <Terminal*>::iterator iter;
  int i;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Make sure that this many registers are even mapped to this block
  if(getNumMappedRegs() < numRegs){
    ERROR("Trying to migrate more registers than are mapped to this block!");
  }
  
  //Make sure that this many registers can fit in the other block
  if(otherBank->getNumFreeRegs() < numRegs){
    ERROR("Trying to migrate more registers than can fit in this block!");
  }

  //Is the other block empty?  If so, we will fill it.
  if(otherBank->isBankEmpty()){
    FPGA->occupyRegBank(otherBank);
  }

  //Push the first N sinks from this block into the end of the other block and update what delay
  // block/routing node they are pointed to.
  for(iter = currSinkList.begin(), i = 0;
      iter != currSinkList.end() && i < numRegs; iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(otherBank->currSinkList.size() - 1));
  }

  currSinkList.erase(currSinkList.begin(), iter);

  //Copy the first N sources from this block into the end of the other block and update what delay
  // block/routing node they are pointed to.
  for(iter = currSourceList.begin(), i = 0;
      iter != currSourceList.end() && i < numRegs; iter++, i++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(otherBank->currSourceList.size() - 1));
  }
  
  //Remove these sources from this block
  //currSourceList.erase(currSourceList.begin(), iter--);
  currSourceList.erase(currSourceList.begin(), iter);
  //CHECK THE ABOVE STATEMENT!
  
  //Did we merge all of the registers from this bank into the other one?
  if(isBankEmpty()){
    FPGA->vacateRegBank(this);
  }

  //Find out if either bank was associated with a logic block.
  //This is an error because we cannot break up monolithic register banks, only swap them
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
  if(otherBank->currSourceList.size() != 0){
    if(otherBank->currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
}

void registerBank::migrateLastNRegs(registerBank *otherBank, int numRegs){
  list <Terminal*>::iterator iter;
  int i;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Make sure that this many registers are even mapped to this block
  if(getNumMappedRegs() < numRegs){
    ERROR("Trying to migrate more registers than are mapped to this block!");
  }
  
  //Make sure that this many registers can fit in the other block
  if(otherBank->getNumFreeRegs() < numRegs){
    ERROR("Trying to migrate more registers than can fit in this block!");
  }

  //Is the other block empty?  If so, we will fill it.
  if(otherBank->isBankEmpty()){
    FPGA->occupyRegBank(otherBank);
  }

  //Push the last N sinks from this block into the front of the other block
  for(iter = currSinkList.begin(), i = 0; i < currSinkList.size() - numRegs; iter++, i++){
    
  }
  otherBank->currSinkList.insert(otherBank->currSinkList.begin(), 
				      iter, currSinkList.end());

  //Remove these sinks from this block
  currSinkList.erase(iter, currSinkList.end());
  
  //Go through all of the sinks in the other block and update what delay block/routing node they are pointed to.
  //At least the routing node portion is absolutely necessary for everyone since all the blocks slide over
  for(iter = otherBank->currSinkList.begin(), i = 0;
      iter != otherBank->currSinkList.end(); iter++, i++){
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }

  //Push the last N sources from this block into the front of the other block
  for(iter = currSourceList.begin(), i = 0; i < currSourceList.size() - numRegs; iter++, i++){
    
  }
  otherBank->currSourceList.insert(otherBank->currSourceList.begin(), 
					iter, currSourceList.end());

  //Remove these sources from this block
  currSourceList.erase(iter, currSourceList.end());
  
  //Go through all of the sources in the other block and update what delay block/routing node they are pointed to.
  //At least the routing node portion is absolutely necessary for everyone since all the blocks slide over
  for(iter = otherBank->currSourceList.begin(), i = 0;
      iter != otherBank->currSourceList.end(); iter++, i++){
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }

  //Did we merge all of the registers from this bank into the other one?
  if(isBankEmpty()){
    FPGA->vacateRegBank(this);
  }

  //Find out if either bank was associated with a logic block.
  //This is an error because we cannot break up static delay blocks, only swap them
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
  if(otherBank->currSourceList.size() != 0){
    if(otherBank->currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
}

//Take the registers in this bank and duplicate them into the other bank.  Randomly
// distribute the fanout between the two, but make sure that each has at least 1 sink
void registerBank::duplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay){
  Net* parentNet;
  PipeliningNet* currPNet;
  PipeliningNet* nextPNet;
  Terminal* regTerm;
  //char tempStr [MAX_STR_LEN];
  string tempStr;
  int i;
  int numMove;
  int select;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Delay Blocks are out of sync!");
  }

  //We know that the other block used to be empty.
  //FPGA->occupyRegBank(otherBank);

  //Find the net that drives this register bank (the source pnet of the first register)
  currPNet = getFirstSink()->sourcePipeliningNet;
  parentNet = currPNet->parentNet;

  //Until we have duplicated all of the registers in the block, do this
  for(i = 0; i < currSinkList.size(); i++){
    //1) Create a new pipeline net at a latency one greater than the current one
    //   and add it to the main net's list
    //strcpy(tempStr, parentNet->name);
    //sprintf(tempStr, "%sPipelineNet%d", tempStr, parentNet->registerID + 1);
    tempStr = parentNet->name + "PipelineNet" + itoa(parentNet->registerID + 1);
    nextPNet = new PipeliningNet(tempStr, currPNet->latency + 1, parentNet);
    parentNet->pipelineNetList.push_back(nextPNet);
    
    //2) Create a new pipelining register and add it to the main net's list of registers
    //   along with the netlist's list of registers
    //strcpy(tempStr, parentNet->name);
    //sprintf(tempStr, "%sReg%d", tempStr, parentNet->registerID);
    tempStr = parentNet->name + "Reg" + itoa(parentNet->registerID);
    parentNet->registerID ++;
    regTerm = new Terminal(parentNet, currPNet, nextPNet, tempStr);
    parentNet->pipelineRegisterList.push_back(regTerm);
    netlist->placementPipelineRegisterVector.push_back(regTerm);
    regTerm->departureTime = 0.0;
    regTerm->requiredTime = criticalPathDelay;

    //3) Make this register a sink of the current net and the source of the next net
    //The slack should be the sink required time minus the source departure time
    //  minus the delay of the wire between the two.
    //That is, CPD - source DT - 0.
    currPNet->pushSinkPlusData(regTerm, 0.0, regTerm->requiredTime - currPNet->source->departureTime);
    nextPNet->source = regTerm;
    
    //4) Place the newly created register into the other register bank
    otherBank->placeTerminal(regTerm);

    currPNet = nextPNet;
  }

  //Now that we have duplicated the registers, we have to split up the sinks
  // of the original register bank.
  currPNet = getLastSource()->destPipeliningNet;

  //Pick a random number of sinks to remove from the original net from 1 to N-1
  numMove = lrand48() % (currPNet->sinkVector.size() - 1);
  numMove++;

  //Randomly select this number of sinks to move from the original net to the new one
  for(i = 0; i < numMove; i++){
    select = lrand48() % currPNet->sinkVector.size();
    
    //The slack should be the sink required time minus the source departure time
    //  minus the delay of the wire between the two.
    //Note that the departure time of a register should always be 0
    //That is, Sink RT - 0 - 0
    nextPNet->pushSinkPlusData(currPNet->getSink(select), 0.0, 
			       currPNet->getSink(select)->requiredTime - nextPNet->source->departureTime);
    //Update the sink's source net
    currPNet->getSink(select)->sourcePipeliningNet = nextPNet;
    currPNet->removeSinkPlusData(select);
  }

  //**********SOMETHING TO LOOK INTO
  //Find out if this bank was associated with a logic block.
  //If so, we should update the netlist to show that we duplicated a monolithic
  // register bank
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Should not be duplicating grouped register banks!");
    }
  }
}

//The other bank has duplicate registers to this one.  We want to transfer the
// nets that use that one to take from this one.  However, only notify the nets
// and sinks, don't disturb the information sitting in the other bank yet.  We
// might need that information if we want to undo this operation.
void registerBank::zipRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay){
  PipeliningNet* recPNet;
  PipeliningNet* sendPNet;
  Net* parentNet;
  vector<Terminal*>::iterator tvIter;
  list<Terminal*>::iterator tlIter;
  int select;

  //Make sure that neither bank is associated with a logic block
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Should not be merging grouped register banks!");
    }
  }
  if(otherBank->currSourceList.size() != 0){
    if(otherBank->currSourceList.front()->lBlock){
      ERROR("Should not be merging grouped register banks!");
    }
  }

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Delay Blocks are out of sync!");
  }

  //Find the net that drives both of these register banks (the source pnet of the first register)
  recPNet = otherBank->getFirstSink()->sourcePipeliningNet;
  parentNet = recPNet->parentNet;

  //Double check that this bank has the same source net
  if(recPNet != getFirstSink()->sourcePipeliningNet){
    ERROR("Trying to combine two nets that do not have the same source!");
  }

  //Take that drive net and remove the reference to the first sink in the other bank
  recPNet->removeSinkPlusData(otherBank->getFirstSink());

  //Now detach all of the sinks from the other bank and attach them to the end of this one
  //Since we might have to undo this move, we'll leave the registers in the other bank alone
  // for the moment. That is, the registers in the other bank will know who they used to 
  //  connect to, but the sinks themselves will think they are attached to this one
  recPNet = getLastSource()->destPipeliningNet;
  
  sendPNet = otherBank->getLastSource()->destPipeliningNet;
  for(tvIter = sendPNet->sinkVector.begin(); tvIter != sendPNet->sinkVector.end(); tvIter++){
    //The slack should be the sink required time minus the source departure time
    //  minus the delay of the wire between the two.
    //Notice, this isn't perfect since the required time for the source might change,
    // but we can't update this without doing a full STA.
    //Thus, we will use CPD - source DT - 0
    recPNet->pushSinkPlusData(*tvIter, 0.0, (*tvIter)->requiredTime - recPNet->source->departureTime);
    (*tvIter)->sourcePipeliningNet = recPNet;    
  }

  //Go through each register in the other bank and delete it from the 
  // parent conventional net's list of registers / main
  // netlist's list of registers.
  //Also, find the destination pnets for each of these registers
  // and remove it from the parent conventional net's list of pnets.
  //Notice, we are not deleting the register itself yet - that is
  // still hanging around in the block, but the netlist should not
  // know anything about these registers any more
  for(tlIter = otherBank->currSinkList.begin(); 
      tlIter != otherBank->currSinkList.end(); tlIter++){
    parentNet->removePipelineRegFromPRList(*tlIter);
    netlist->removePipelineRegisterFromPlacementPRVector(*tlIter);
    parentNet->removePipelineNetFromPNList((*tlIter)->destPipeliningNet);
  }

  //Although we haven't actually done it yet, assume that we will accept this
  // change and vacate the other duplicate bank in terms of free register counting
  //FPGA->vacateRegBank(otherBank);
}

//We zipped the registers from the other bank into this one, but now we want to undo that.
//The other bank still should have all of its registers and internal connections intact, 
// but we have to reconnect the ends.
//We also have to put the references to these registers and pipelining net back into the
// parent nets.
void registerBank::reDuplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay){
  Net* parentNet;
  PipeliningNet* currPNet;
  PipeliningNet* nextPNet;
  int select;

  vector<Terminal*>::iterator tIter;

  //Make sure that this bank is not associated with a monolithic register bank
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Should not be duplicating grouped register banks!");
    }
  }

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Register banks are out of sync!");
  }

  //We have to undo the zip that we did earlier.  As far as empty register bookkeeping
  // goes, we assumed that we were going to accept the zip from earlier.  Thus, to keep
  // things straight, we have to re-occupy the other bank
  //FPGA->occupyRegBank(otherBank);

  //Find the net that drives this register bank (the source pnet of the first register)
  currPNet = getFirstSink()->sourcePipeliningNet;
  parentNet = currPNet->parentNet;

  //First, the registers and their pNets still exist in block 1.
  //We knew we might have to undo this, so we didn't delete them yet. 
  //Thus, we don't have to create them, but we do have to:
  // 1) hook up the first one
  //  The slack should be the sink required time minus the source departure time
  //   minus the delay of the wire between the two.
  //  That is, CPD - source DT - 0.
  currPNet->pushSinkPlusData(otherBank->getFirstSink(), 0.0, 
			     criticalPathDelay - currPNet->source->departureTime);

  // 2) put all of the registers back into the parent net's list/the netlist's list
  //    along with the pnets back into the parent net's list
  for(list <Terminal*>::iterator iter = otherBank->currSinkList.begin(); 
      iter != otherBank->currSinkList.end(); iter++){
    parentNet->pipelineRegisterList.push_back(*iter);
    netlist->placementPipelineRegisterVector.push_back(*iter);
    parentNet->pipelineNetList.push_back((*iter)->destPipeliningNet);    
  }

  //Now that we have duplicated the registers, we have to split up the sinks
  // of the original register bank.  
  //The last pipelining net in bank 1 knows who it wants to connect to,
  // we just have to:
  //1) remove the sink from the last net in bank 0
  //2) tell the sink itself who it's parent is
  currPNet = getLastSource()->destPipeliningNet;
  nextPNet = otherBank->getLastSource()->destPipeliningNet;

  for(tIter = nextPNet->sinkVector.begin(); tIter != nextPNet->sinkVector.end(); tIter++){
    //Find this sink in the bank 0 net and delete it
    currPNet->removeSinkPlusData(*tIter);

    //Now inform the sink itself who it's new source is
    (*tIter)->sourcePipeliningNet = nextPNet;
  }
}

//This terminal should be the first sink mapped to this register.
//We would like to remove this terminal from this register bank.
//However, note that we are NOT remapping the routing nodes.
//That will have to be taken care of later.
void registerBank::removeStartReference(Terminal *T){
  if(currSinkList.front() != T || currSourceList.front() != T){
    ERROR("Trying to remove non-initial register from bank!");
  }
  else{
    currSinkList.pop_front();
    currSourceList.pop_front();
    //currSinkList.erase(currSinkList.begin());
    //currSourceList.erase(currSourceList.begin());

    //If this removes the last mapped sink from the bank, this bank has
    // been vacated.
    if(currSinkList.size() == 0){
      FPGA->vacateRegBank(this);
#ifdef PLACER_DEBUG
      numRegistersEstimatedFreed -= getNumPossibleDelay();
#endif
    }
  }
}

//This terminal should be the last sink mapped to this register.
//We would like to remove this terminal from this register bank.
void registerBank::removeEndReference(Terminal *T){
  if(currSinkList.back() != T || currSourceList.back() != T){
    ERROR("Trying to remove non-final register from bank!");
  }
  else{
    currSinkList.pop_back();
    currSourceList.pop_back();

    //If this removes the last mapped sink from the bank, this bank has
    // been vacated.
    if(currSinkList.size() == 0){
      FPGA->vacateRegBank(this);
#ifdef PLACER_DEBUG
      numRegistersEstimatedFreed -= getNumPossibleDelay();
#endif
    }
  }
}

// When we zipped the registers from this bank into another, we knew
//  we might have to undo the move.  Thus, we left the registers inside
//  this bank intact.  Clear them now.
void registerBank::freeBank(){
  list<Terminal*>::iterator tIter;

  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    delete ((*tIter)->destPipeliningNet);
    delete *tIter;
  }
  currSinkList.clear();
  currSourceList.clear();

  FPGA->vacateRegBank(this);
}

//Return the total wire cost of the nets driven by the registers in this bank
double registerBank::returnTotalWireCost(){
  list<Terminal*>::iterator tIter;
  double wireCost = 0.0;
  
  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    wireCost += ((*tIter)->destPipeliningNet->wireCost);
  }
  return wireCost;
}

//Return the total timing cost of the nets driven by the registers in this bank
double registerBank::returnTotalTimingCost(){
  list<Terminal*>::iterator tIter;
  double timingCost = 0.0;
  
  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    timingCost += ((*tIter)->destPipeliningNet->timingCost);
  }
  return timingCost;
}

//We are going to get rid of the registers in this bank so update the cut cost appropriately
void registerBank::updateDeletedCutCost(){
  list<Terminal*>::iterator tIter;
  int i;

  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    //Decrement the span of the pipelining net driven by this register
    for(i = (*tIter)->destPipeliningNet->minCol; i < (*tIter)->destPipeliningNet->maxCol; i++){
      FPGA->cutCostVector.at(i) += -1.0;
    }
  }
}

void registerBank::print_info(){
  cout << "Location = " << getHostCLB()->col  << ", " << getHostCLB()->row
       << " bank " << getHostBankNum() << endl;
  cout << "\tBank has " << getNumPossibleDelay() << " registers" << endl;
}

//We moved around some of the register mapped to this register bank, so
// make sure everyone is mapped to the correct location.
void registerBank::updateRoutingNodes(){
  list <Terminal *>::iterator sinkIter, sourceIter;
  int i;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Register bank are out of sync!");
  }  
  
  for(sinkIter = currSinkList.begin(), sourceIter = currSourceList.begin(), i = 0;
      sinkIter != currSinkList.end(), sourceIter != currSourceList.end(); 
      sinkIter ++, sourceIter++, i++){
    if(*sinkIter != *sourceIter){
      ERROR("Sink and source mismatch!");
    }
    (*sinkIter)->set_rNode(routingNodeVector.at(i));
    (*sourceIter)->set_rNode(routingNodeVector.at(i));
  }
}

//Are the LUTs in the clb at the same location empty?
//How about the other register banks in the same delay block?
bool registerBank::emptyCLB(){
  if(!FPGA->getCLBAt(getHostCLB()->row, getHostCLB()->col)->areLUTsEmpty() || 
     !FPGA->getCLBAt(getHostCLB()->row, getHostCLB()->col)->areRegistersEmpty())
    return false;
  return true;
}

//Swap the contents of these two register banks.
void registerBank::swapMapping(logicBlock* myLB, registerBank* otherBank, logicBlock* otherLB){
  //Clear the old mappings of the old logic blocks
  //Since this is only called from logic blocks in placePipeline.cc, we know that this register bank
  // must be occupied and myLB must point to a logic block.
  unPlaceLogicBlock(myLB);

  //However, the other register bank might or might not be occupied.
  //First, check to see if otherLB points to something.
  //If it does, then the other register bank is occupied.
  if(otherLB){
    otherBank->unPlaceLogicBlock(otherLB);
  }
  //However, just because otherLB does not point to anything does not mean that the other bank is empty.
  //It could have pipelining registers mapped to it
  else if(!otherBank->isBankEmpty()){
    //Migrate the contents to this register bank and the update the timing information for it.
    otherBank->copyContentsToAndEmpty(this);
    updateSinkDelaysAndSlacks();
  }

  //Make new mappings
  otherBank->placeLogicBlock(myLB);
  //If the other bank had a logic block mapped to it, place that logic block
  if(otherLB){
    placeLogicBlock(otherLB);
  }

  //Update the sink timings and slack for the nets we just moved
  // The slack calculations are not exact since the critical path might have changed
  myLB->updateSinkDelaysAndSlacks();
  if(otherLB)
    otherLB->updateSinkDelaysAndSlacks();
}

//This will swap the contents of the two array CLBs that hold this register bank
// and the other register bank.
//myLB will point to the logic block currently mapped to this register bank.
//otherLB will point to the logic block currently mapped to the other register bank.
// void registerBank::swapEntireCLBMappings(logicBlock* myLB, registerBank* otherBank, logicBlock* otherLB){
//   c_clb* myCLB;
//   c_clb* otherCLB;

//   //Even though it was used to call this function, this register bank will not always be occupied by a logic block.
//   //The original call to this will, but when we undo the move, since we swapped the whole CLB, there is no guarantee that
//   // this register location will still be occupied.
//   //However, if possible, we should check since this is faster than looking through all of the items in the CLB
//   // to find something that is occupied.
//   if(myLB){
//     myCLB = myLB->parentCLB;
//   }
//   else{
//     myCLB = returnMappedCLB();
//   }
//   if(!myCLB){
//     ERROR("This should be mapped to something!");
//   }

//   //However, the other register bank might or might not be occupied.
//   //First, check to see if otherLB points to something.
//   //If it does, then the other register bank is occupied.
//   if(otherLB){
//     otherCLB = otherLB->parentCLB;
//   }

//   //However, just because otherLB does not point to anything does not mean that the other bank is empty.
//   //It could have pipelining registers mapped to it.
//   //However, we will assume this will never happen since we will not do CLB-level placement on
//   // a unpacked netlist.
//   //Instead, we will check the other physical CLB for any potential logic blocks mapped to
//   // its LUTs and FFs.
//   else{
//     otherCLB = otherBank->returnMappedCLB();
//   }

//   myCLB->unPlaceLogicBlocksFromLUTsAndFF(this);
//   if(otherCLB)
//     otherCLB->unPlaceLogicBlocksFromLUTsAndFF(otherBank);

//   myCLB->placeLogicBlocksToLUTsAndFF(otherBank);
//   if(otherCLB)
//     otherCLB->placeLogicBlocksToLUTsAndFF(this);

//   myCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
//   if(otherCLB)
//     otherCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
// }
void registerBank::swapEntireCLBMappings(registerBank* otherBank){
  c_clb* myCLB;
  c_clb* otherCLB;

  //At least one of the register banks in the CLB this register banks belongs to
  // will have a logic block mapped.
  myCLB = returnMappedCLB();

  if(!myCLB){
    ERROR("This should be mapped to something!");
  }

  //However, the other register bank might or might not be occupied.
  //It could have pipelining registers mapped to it.
  //However, we will assume this will never happen since we will not do CLB-level placement on
  // a unpacked netlist.
  //In either case, we will check the other physical CLB for any potential logic blocks mapped to
  // its LUTs and FFs.
  otherCLB = otherBank->returnMappedCLB();

  myCLB->unPlaceLogicBlocksFromLUTsAndFF(this);
  if(otherCLB)
    otherCLB->unPlaceLogicBlocksFromLUTsAndFF(otherBank);

  myCLB->placeLogicBlocksToLUTsAndFF(otherBank);
  if(otherCLB)
    otherCLB->placeLogicBlocksToLUTsAndFF(this);

  myCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
  if(otherCLB)
    otherCLB->updateSinkDelaysAndSlacksForLUTsAndFF();
}

void registerBank::updateCostAndDeltasForEntireCLB(registerBank* otherRegisterBank, 
						   bool timing_driven, double criticalPathDelay, double *deltaWire, 
						   double *deltaTiming){
  c_clb* otherCLB;
  c_clb* myCLB;

  //Since we swap them, the CLB that the other register bank belongs to MUST have
  // at least one register bank mapped to it.
  otherCLB = otherRegisterBank->returnMappedCLB();

  if(!otherCLB){
    ERROR("This should be mapped to something!");
  }
  
  myCLB = returnMappedCLB();

  otherCLB->updateCostsAndDeltasForLUTsAndFF(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  if(myCLB)
    myCLB->updateCostsAndDeltasForLUTsAndFF(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
}

void registerBank::updateCostAndDeltasForEntireCLBCostDeltaDelay(registerBank* otherRegisterBank, 
						   bool timing_driven, double criticalPathDelay, double *deltaWire, 
						   double *deltaTiming, double *deltaTimingPrime){
  c_clb* otherCLB;
  c_clb* myCLB;

  //Since we swap them, the CLB that the other register bank belongs to MUST have
  // at least one register bank mapped to it.
  otherCLB = otherRegisterBank->returnMappedCLB();

  if(!otherCLB){
    ERROR("This should be mapped to something!");
  }
  
  myCLB = returnMappedCLB();

  otherCLB->updateCostsAndDeltasForLUTsAndFFCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  if(myCLB)
    myCLB->updateCostsAndDeltasForLUTsAndFFCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
}

//Go through all of the register banks and LUTs associated with the vprCLB at this location.
//If there are any logic blocks mapped to these locations, return the CLB associated with them
c_clb* registerBank::returnMappedCLB(){
  c_clb* mappedCLB = NULL;
  vprCLB *myCLB = FPGA->getCLBAt(hostCLB->row, hostCLB->col);

  //Try to extract a CLB from the FFs at this locations
  for(vector <registerBank*>::iterator rbIter = myCLB->registerBankVector.begin(); 
      rbIter != myCLB->registerBankVector.end(); rbIter++){
    if((*rbIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*rbIter)->get_currBlock()->parentCLB){
	  ERROR("Register bank is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*rbIter)->get_currBlock()->parentCLB;
      }
    }
  }
  
  //Try to extract a CLB from the LUTs at this locations
  for(vector <vprLUT *>::iterator lutIter = myCLB->lutVector.begin(); 
      lutIter != myCLB->lutVector.end(); lutIter++){
    if((*lutIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*lutIter)->get_currBlock()->parentCLB){
	  ERROR("LUT is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*lutIter)->get_currBlock()->parentCLB;
      }
    }
  }
  return mappedCLB;
}

vprCLB::vprCLB(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum = 21119;
}

// vprCLB::~vprCLB(){
//   if(derived_magicNum!=21119)
//     ERROR("trying to double delete node here...");
//   else
//     magicNum=0;
// }

void vprCLB::addAllRegBanksToRegisterBankVector(){
  for(vector<registerBank *>::iterator iter = registerBankVector.begin();
      iter != registerBankVector.end(); iter++){
    FPGA->registerBankVector.push_back(*iter);
  }
}

void vprCLB::addAllRegBanksToFreeRegisterBankList(){
  for(vector<registerBank *>::iterator iter = registerBankVector.begin();
      iter != registerBankVector.end(); iter++){
    FPGA->freeRegisterBankList.push_back(*iter);
  }
}

// Correspond a routing node in the RR graph with this block
// We assume that registers are added in the correct order if that is important
// That is, in a particular bank, register 0 is added before register 1,
// register 1 is added before register 2, etc.if they are in a
// shift-chain orientation.
void vprCLB::addRoutingNodeToRegisterBank(routingNode* r, int bank){
  // We might not have that many banks yet
  if(registerBankVector.size() < bank + 1){
    registerBankVector.resize(bank + 1);
  }

  //Alternatively, that bank might not be allocated yet
  if(!registerBankVector.at(bank)){
    registerBankVector.at(bank) = new registerBank();
    registerBankVector.at(bank)->setHostCLB(this, bank);
  }
  
  registerBankVector.at(bank)->addRoutingNode(r);
}

vprLUT* vprCLB::getOrMakeLUT(int i){
  //Our lut vector may not be large enough
  if(lutVector.size() <= i){
    lutVector.resize(i + 1);
  } 
  
  if(!lutVector.at(i)){
    lutVector.at(i) = new vprLUT(row, col);
  }

  return lutVector.at(i);
}

bool vprCLB::areRegistersEmpty(){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->isBankEmpty()){
      return false;
    }
  }
  return true;
}

bool vprCLB::areLUTsEmpty(){
  vector<vprLUT *>::iterator lutIter;

  for(lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if(!(*lutIter)->isEmpty()){
      return false;
    }
  }
  return true;
}

//Try to find an empty register bank in the given CLB
//if there isn't one, return NULL
registerBank* vprCLB::getFirstEmptyRegisterBank(){
  //Go through all of the register banks in this clb and return the first empty one
  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->isBankEmpty()){
      //cout << "Found empty register bank!" << endl;
      return (*rbIter);
    }
  }
  return NULL;
}

//Try to find the first empty register bank in the given CLB.
//Failing that, find the first register bank not related to the LUTs in this CLB
registerBank* vprCLB::getFirstEmptyUnrelatedRegisterBank(){
  registerBank *emptyRegBank;
  Terminal *source;
  int sourceRow, sourceCol;
  
  //See if there is an empty register bank in this CLB
  emptyRegBank = getFirstEmptyRegisterBank();
  
  if(emptyRegBank){
    return emptyRegBank;
  }

  //cout << "All register banks are occupied!  Looking for an unrelated FF" << endl;
  //Failing that, find the first register not related to the LUTs in the host CLB
  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    //This register bank might be driven by a pipelining net or a regular net
    if((*rbIter)->getFirstSink()->sourcePipeliningNet){
      source = (*rbIter)->getFirstSink()->sourcePipeliningNet->source;
    }
    else{
      source = (*rbIter)->getFirstSink()->net->source;
    }

    //See if this register bank is occupied by a self-feeding register bank in the netlist (constant generator)
    if(source->lBlock && (*rbIter)->getFirstSink()->lBlock){
      if(source->lBlock == (*rbIter)->getFirstSink()->lBlock){
	//cout << "Found a constant generator register bank!" << endl;
	return(*rbIter);
      }
    }


    sourceRow = source->getLogicBlockRow();
    sourceCol = source->getLogicBlockCol();
    
    if(sourceCol != (*rbIter)->getHostCLB()->get_col() || sourceRow != (*rbIter)->getHostCLB()->get_row()){
      //cout << "Found a register bank not driven by something in this CLB!" << endl;
      return(*rbIter);
    }
  }

  ERROR("Unable to find a suitable register bank in this CLB");
  return NULL;
}

//Swap all of the FFs in this CLB with the ones in the other CLB.
void vprCLB::swapAllFFs(vprCLB *otherCLB){
  vector<registerBank *>::iterator thisRBIter;
  vector<registerBank *>::iterator otherRBIter;

  //Go though both of these clbs
  for(thisRBIter = registerBankVector.begin(), otherRBIter = otherCLB->registerBankVector.begin(); 
      thisRBIter != registerBankVector.end() && otherRBIter != otherCLB->registerBankVector.end(); thisRBIter++, otherRBIter++){
    if(!(*thisRBIter)->isBankEmpty() || !(*otherRBIter)->isBankEmpty()){
      //One of the two isn't empty so we should do a swap.
      if(!(*thisRBIter)->isBankEmpty()){
	(*thisRBIter)->swapBank(*otherRBIter);
      }
      else{
	(*otherRBIter)->swapBank(*thisRBIter);
      }
    }
  }
}

//Update the sink delays and slacks for all of the FFs in this physical CLB
void vprCLB::updateSinkDelaysAndSlacksForAllFFs(){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->isBankEmpty()){
      (*rbIter)->updateSinkDelaysAndSlacks();
    }
  }
}

void vprCLB::updateCostAndDeltasForAllFFs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->isBankEmpty()){
      (*rbIter)->updateWireAndTimingCosts(timing_driven, criticalPathDelay);

      *deltaWire += (*rbIter)->determineDeltaWireCost();
      
      if(timing_driven){
	*deltaTiming += (*rbIter)->determineDeltaTimingCost();
      }
    }
  }
}

void vprCLB::updateCostAndDeltasForAllFFsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->isBankEmpty()){
      (*rbIter)->updateWireAndTimingCosts(timing_driven, criticalPathDelay);

      *deltaWire += (*rbIter)->determineDeltaWireCost();
      
      if(timing_driven){
	*deltaTiming += (*rbIter)->determineDeltaTimingCost();
	*deltaTimingPrime += (*rbIter)->determineDeltaTimingCostPrime();
      }
    }
  }
}


void vprFPGA::placeClockPin(int *clockRow, int *clockCol, int *clockPTC);
